#ifndef _ARTE_TABLE_HH_
#define _ARTE_TABLE_HH_
#define _ARTE_HH_
// *****************************************************************************
//
// source:
//
// type:      source code
//
// created:   09. Mar 1997 by C. Borgmeier
//
// author:    Thorsten Glebe
//            HERA-B Collaboration
//            Max-Planck-Institut fuer Kernphysik
//            Saupfercheckweg 1
//            69117 Heidelberg
//            Germany
//            E-mail: T.Glebe@mpi-hd.mpg.de
//
// Description: This file is part of the C++ interface for Arte
//              the HERA-B reconstruction and analysis tool
//
//              This is the definition of the ArtePointer class.
//
//              ArtePointer allows access to the Arte data tables in
//              a pointer-like way.
//
// changes:
// 09 Mar 1997 creation by Christoph Borgmeier
// 08 Nov 2000 (TG) added constructor to ArteTableIterator
// 11 Nov 2000 (TG) code cleanup
// 13 Nov 2000 (TG) added print() to class ArteTableIterator
// 13 Nov 2000 (TG) ROOT integration of ArteTableIterator
// 13 Nov 2000 (TG) doc++ comments added
// 16 Nov 2000 (TG) small modification in ArteTable
// 16 Nov 2000 (TG) declared some member functions inline
// 30 Nov 2000 (TG) added operator<< for ArteTables, added print() member
// 27 Dec 2000 (TG) end() optimization in ArteTable::print()
// 10 Jan 2001 (TG) added operator<() and operator>() to ArteTableIterator
//
// *****************************************************************************

#include <string>
#include <iterator>
#include <iosfwd>
#include "arte/TObject.hh"

#ifndef NO_ARTE
extern "C" void mmbrel_( const int& link,
			 const ArtePointerBase& e1,
			 const ArtePointerBase& e2 );
extern "C" void mmdrop_( int &link );
extern "C" void mmbook_( int &link, const int& );
#endif

template <class T> class ArtePointer;
template <class T> class ArteTable;
template <class T> class ArteTableIterator;


/** Arte Table Iterator class 
    @author C. Borgmeier, T. Glebe
*/
//==============================================================================
// ArteTableIterator
//==============================================================================
template <class T>
class ArteTableIterator : public TObject
{
  friend class ArteTable<T>;
public:
  typedef ArteTableIterator<T> self;
  typedef ArtePointer<T> value_type;
  typedef value_type* pointer;
  typedef value_type& reference;
#ifndef __CINT__
  typedef const value_type& const_reference;
#endif
  typedef std::random_access_iterator_tag iterator_category;
  typedef ptrdiff_t difference_type;

  /** @name --- Constructors --- */
  ///
  ArteTableIterator() { ptr.setInt( 1 ); }
  ///
  ArteTableIterator( const ArtePointer<T>& rhs ) : 
    ptr( rhs ) {}
  ///
  ArteTableIterator( const ArteTableIterator<T>& rhs ) :
    ptr( rhs.ptr ) {}

  /** @name --- Operators --- */
  ///
  inline ArtePointer<T>& operator*() { return ptr; }
  //  inline reference operator*() { return ptr; }

#ifdef __CINT__
  ///
  inline const ArtePointer<T>& operator*() const { return ptr; }
#else
  inline const_reference operator*() const { return ptr; }
#endif

  ///
  self& operator++() {
    ptr.increment();
    return *this;
  }

  ///
  inline self operator++(int) {
    self tmp = *this;
    ptr.increment();
    return tmp;
  }

  // 

  ///
  self& operator--() {
    ptr.decrement();
    return *this;
  }

  ///
  inline self operator--(int) {
    self tmp = *this;
    ptr.decrement();
    return tmp;
  }

  ///
  inline const self& operator+=( int offset ) { ptr+=offset; return *this; }
  ///
  inline const self& operator-=( int offset ) { ptr-=offset; return *this; }
  ///
  inline bool operator!=( const ArteTableIterator<T>& rhs ) const
    { return ptr != rhs.ptr; }
  ///
  inline bool operator==( const ArteTableIterator<T>& rhs ) const
    { return ptr == rhs.ptr; }
  ///
  inline bool operator<( const ArteTableIterator<T>& rhs) const
    { return ptr < rhs.ptr; }
  ///
  inline bool operator>( const ArteTableIterator<T>& rhs) const
    { return ptr > rhs.ptr; }
  ///
  inline int operator-( const ArteTableIterator<T>& rhs ) const
    { return ptr -  rhs.ptr; }
  ///
  inline ArteTableIterator<T> operator+( int rhs ) const {
    return ArteTableIterator<T>(*this) += rhs;
  }
  //    {  return operator+(*this,rhs); }
  ///
  inline ArteTableIterator<T> operator-( int rhs ) const
    {  return ArteTableIterator<T>(*this) -= rhs; }
  //    {  return operator+(*this,-rhs); }

  ///
  inline void print( std::ostream& os ) const {
    ptr.print(os);
  }

#if !defined(ARTE_CXX_COMPAT_MODE) || (ARTE_CXX_COMPAT_MODE==1)
  /** @name --- Arte compatibility mode --- */
  ///
  operator ArtePointer<T>() const { return ptr; }
  ///
  inline int valid() const;
  ///
  ArtePointer<T> next();
  ///
  inline static unsigned int nrow();
  ///
  inline T* operator->() const;
#else
protected:
#endif
  /// public only in ARTE compat. mode
  inline unsigned int index() const { return ptr.index(); }
  /// forbidden constructor
  ArteTableIterator( unsigned int i )
    : ptr(*( ArteTable<T>::begin() + (i - 1) )) {}

private:
  ArtePointer<T> ptr;  //! non persistent Pointer to object

#if defined(USE_GATE) && defined(USE_ROOT) && defined(NO_ARTE)
  ClassDefT(ArteTableIterator<T>,1)
#endif
}; // End of class ArteTableIterator


#if defined(USE_GATE) && defined(USE_ROOT) && defined(NO_ARTE)
ClassDefT2(ArteTableIterator,T)
  ; // for doc++
#endif


template <class T>
inline std::random_access_iterator_tag
iterator_category(const ArteTableIterator<T>&) {
  return std::random_access_iterator_tag();
}


template <class T>
inline ptrdiff_t*
distance_type(const ArteTableIterator<T>&) {
  return 0;
}

//==============================================================================
// operator <<
//==============================================================================
template <class T>
inline std::ostream& operator<<( std::ostream& os, const ArteTableIterator<T>& t ) {
  t.print( os );
  return os;
}

//  template <class T>
//  inline int operator-( const ArteTableIterator<T>& lhs,
//  		      const ArteTableIterator<T>& rhs )
//  {
//    return operator-( lhs.ptr, rhs.ptr );
//  }


/** Arte Table class
    @author C. Borgmeier, T. Glebe
*/
//==============================================================================
// ArteTable - STL compatible container for Arte tables
//==============================================================================
template <class T>
class ArteTable: public TObject {
  //  friend class ArtePointer<T>;
public:
  typedef ArtePointer<T>    value_type;
  typedef value_type&       reference;
#ifndef __CINT__
  typedef const value_type& const_reference;
#endif
  typedef size_t            size_type;
  typedef ptrdiff_t         difference_type;

  typedef ArteTableIterator<T> iterator;
  typedef ArteTableIterator<T> const_iterator;

  /** @name --- Constructors --- */
  ///
  ArteTable() {};

  /** @name --- Access methods --- */
#ifndef __CINT__
  /// not available in CINT
  static std::string getName();

  // this can become too long for CINT
  typedef std::reverse_bidirectional_iterator<iterator, value_type, reference,
    difference_type> reverse_iterator;

  //   this thing does not go through egcs 1.0
  //   typedef std::reverse_bidirectional_iterator<const_iterator, value_type,
  //     const_reference, difference_type> const_reverse_iterator;
#endif /* __CINT__ */
public:
  ///
  static inline iterator begin();
  //  static iterator begin();
  ///
  static inline iterator end();
  //  static iterator end();
  ///
  static unsigned int size();
  ///
  static ArtePointer<T> push_back(const T& e);
  ///
  static ArtePointer<T> allocate( unsigned int n );

public:
  ///
  void print(std::ostream& os) const {

    const ArteTableIterator<T> end = ArteTable<T>::end();
    for(ArteTableIterator<T> it = ArteTable<T>::begin();
	it != end; ++it)
      cout << **it << endl;
    return;
  }

#ifdef USE_ROOT
  /// only available in ROOT
  virtual void Print( Option_t* = "" );
#endif
  ///
  inline static bool isArteTable() { return true; }

#if !defined(ARTE_CXX_COMPAT_MODE) || (ARTE_CXX_COMPAT_MODE==1)
  /// only in ARTE compat. mode
  inline static unsigned int nrow() { return size(); }
  // private:
#endif
  /// private in Arte compat. mode
  static void* getLink();

#ifndef NO_ARTE
public:
  /** @name --- ARTE only functions --- */
  ///
  inline static void book( int num = 1 )
    { mmbook_( *static_cast<int*>(getLink()), num ); }
  ///
  inline static bool booked() { return *static_cast<int*>(getLink())!=0; }
  ///
  inline static void free() { mmdrop_( *static_cast<int*>(getLink()) ); }
  ///
  inline static void clear() { mmdrop_( *static_cast<int*>(getLink()) ); }

  // private:
  /// Return the width of the Arte table
  inline static int ncol() { return gcbank_.iq[*static_cast<int*>(getLink())]; }
#endif

#if defined(USE_GATE) && defined(USE_ROOT)
  ClassDefT(ArteTable<T>,1)
#endif
}; // end of class ArteTable


#if defined(USE_GATE) && defined(USE_ROOT)
ClassDefT2(ArteTable,T)
  ; // for doc++
#endif

//==============================================================================
// operator <<
//==============================================================================
template <class T>
inline std::ostream& operator<<( std::ostream& os, const ArteTable<T>& t ) {
  t.print( os );
  return os;
}


#endif /* _ARTE_TABLE_HH_ */
