#ifndef ARTE_POINTER_BASE_HH_
#define ARTE_POINTER_BASE_HH_
// *****************************************************************************
//
// source:
//
// type:      source code
//
// created:   10. Jan 1996 by C. Borgmeier
//
// author:    Thorsten Glebe
//            HERA-B Collaboration
//            Max-Planck-Institut fuer Kernphysik
//            Saupfercheckweg 1
//            69117 Heidelberg
//            Germany
//            E-mail: T.Glebe@mpi-hd.mpg.de
//
// Description:  This file is part of the C++ interface for Arte
//               the HERA-B reconstruction and analysis tool
//
//               This is the definition of the ArtePointerBase class.
//
//               ArtePointerBase is the base class of the ArtePointer
//               template class.
//
// changes:
// 10 Jan 1996 creation by Christoph Borgmeier, Hartwig Albrecht
// 08 Nov 2000 (TG) added operator-=()
// 13 Nov 2000 (TG) added doc++ comments
// 17 Nov 2000 (TG) putted code from PointerBase.icc into class def.
// 15 Mar 2001 (TG) null() member added
//
// *****************************************************************************
#include <iosfwd>
#include "arte/TObject.hh"

#ifdef USE_GATE
class TBuffer;
#endif


/** Base class for ArtePointers
    @author C. Borgmeier, T. Glebe
*/
//==============================================================================
// ArtePointerBase
//==============================================================================
class ArtePointerBase {
#ifdef USE_GATE
  friend inline TBuffer& operator<<( TBuffer&, ArtePointerBase );
  friend inline TBuffer& operator>>( TBuffer&, ArtePointerBase& );
#endif

public:
  /** @name --- Constructors --- */
  ///
  ArtePointerBase() : idx(0) {}
  ///
  ArtePointerBase( UShort_t i ) : idx( i ) {}
  ///
  ArtePointerBase( const ArtePointerBase& rhs ): idx( rhs.idx ) {}
  ///
  const ArtePointerBase& operator=( const ArtePointerBase& rhs ) { 
    idx = rhs.idx;
    return *this;
  }

  /** @name --- Access methods --- */
  ///
  inline UShort_t index() const { return idx; }
  ///
  inline void setInt( const UShort_t index ) { idx = index; }
  ///
  inline bool null() const { return (idx == 0) ? true : false; }
  
  /** @name --- Operators --- */
  ///
  inline bool operator< ( const ArtePointerBase& rhs) const { 
    return idx < rhs.idx;
  }
  ///
  inline bool operator> ( const ArtePointerBase& rhs) const { 
    return idx > rhs.idx;
  }
  ///
  inline int  operator+ ( const ArtePointerBase& rhs) const { 
    return idx + rhs.idx;
  }
  ///
  inline int  operator- ( const ArtePointerBase& rhs) const { 
    return idx - rhs.idx;
  }
  ///
  inline bool operator==( const ArtePointerBase& rhs ) const {
    return idx == rhs.idx;
  }
  ///
  inline bool operator!=( const ArtePointerBase& rhs ) const {
    return idx != rhs.idx;
  }
  ///
  inline ArtePointerBase& operator++() { ++idx; return *this; }
  ///
  inline ArtePointerBase& operator--() { --idx; return *this; }
  ///
  inline ArtePointerBase  operator++(int) {
    ArtePointerBase tmp = *this;
    ++idx;
    return tmp;
  }
  ///
  inline ArtePointerBase  operator--(int) {
    ArtePointerBase tmp = *this;
    --idx;
    return tmp;
  }


  /** @name --- Expert functions --- */
  /// used by operator<<()
  inline void print( std::ostream& os ) const {
    os << '[' << index() << ']';
  }

#if !defined(ARTE_CXX_COMPAT_MODE) || (ARTE_CXX_COMPAT_MODE==1)
public:
#else
protected:
#endif

  // made these operators always public (T.Glebe)
  //  int operator==( const ArtePointerBase & arg ) const;
  //  int operator!=( const ArtePointerBase & arg ) const;

  //protected:
  /// Returns the private data.
  inline UShort_t& indexReference() { return idx; }

  ///
  inline void increment() { ++idx; }
  ///
  inline void decrement() { --idx; }
  ///
  const ArtePointerBase& operator+=( int offset ) { idx+=offset; return *this; }
  ///
  const ArtePointerBase& operator-=( int offset ) { idx-=offset; return *this; }
  
private:
  UShort_t idx;

}; // end of class ArtePointerBase


//==============================================================================
// operator <<
//==============================================================================
inline std::ostream& operator<<( std::ostream& os, const ArtePointerBase& ap ) {
  ap.print( os );
  return os;
}

// #include "PointerBase.icc"

#ifdef XXX
//==============================================================================
// ArteTest
//==============================================================================
template <class T> class ArteTest : public ArtePointerBase {
public:
  ArteTest() : ArtePointerBase( 0 ) {}

  inline void pprint( std::ostream& os ) const {
    os << "HHHHU" << endl;
  }
#ifdef USE_ROOT
  ClassDefT(ArteTest<T>,1);
#endif
}; // end of class ArteTest

#ifdef USE_ROOT
ClassDefT2(ArteTest,T)
  ; // for doc++
#endif
#endif // XXX
#endif
