// ********************************************************************
//
// source:
//
// type:      source code
//
// created:   20. Jul 1997
//
// author:    Thorsten Glebe
//            HERA-B Collaboration
//            Max-Planck-Institut fuer Kernphysik
//            Saupfercheckweg 1
//            69117 Heidelberg
//            Germany
//            E-mail: T.Glebe@mpi-hd.mpg.de
//
// Description: Class representing Topp entries via named objects. Additional
//              functionality compared to Topp: pattern can be created via
//              functions swith(), sswith(), sdecaysTo(), createPattern()
//
// changes:
// 23 Jan 2001 (TG) creation
// 28 Feb 2001 (TG) operator==() added
// 12 Mar 2001 (TG) CTOR added, anychrg added
// 15 Aug 2001 (TG) moved topp() to .icc file
// 22 Aug 2001 (TG) added createPattern(), swith(), sswith()
// 27 Aug 2001 (TG) print() improved, updateTopp() improved
// 28 Aug 2001 (TG) modified operator==(ToppIf)
// 05 Sep 2001 (TG) added lund()
// 14 Sep 2001 (TG) renamed sswith() to with()
//
// ********************************************************************

#include <iostream>
#include <string>
#include <algorithm>
#include <map>

#include "clue/GTopp.hh"
#include "clue/Topp.hh"
#include "pattern/SPattern.hh"
#include "pattern/SPatternMatchContainer.hh"
#include "pattern/SmartPointer.hh"

// map to synchronize global GTopp objects with ArteTable<Topp>
typedef std::map<string, GTopp*>  Topp_m;

// Allocate GToppMap by 1st GTopp CTOR call because order
// of static initalization is undefined. Otherwise it
// can happen that GTopp CTOR is called before GToppMap is
// initialized!
static Topp_m* GToppMap = 0;

//==============================================================================
// InitGToppMap: allocate GToppMap if necessary
//==============================================================================
static void InitGToppMap() {
  if( GToppMap == 0 ) GToppMap = new Topp_m;
  return;
}

//==============================================================================
// updateTOPP
//==============================================================================
// This function sets the ArtePointer<Topp> of the global GTopp objects.
// updateTopp() is called in StreamObject::Streamer()
void updateTOPP() {

  const Topp_m::iterator gend = GToppMap->end();

  const ArteTable<Topp>::iterator end = ArteTable<Topp>::end();
  for( ArteTable<Topp>::iterator tit = ArteTable<Topp>::begin();
       tit != end; ++tit ) {
    const ArtePointer<Topp> topp = *tit;
      
    // find pointer of global GTopp object, see PdtNames.[C,hh]
    Topp_m::iterator gtopp = GToppMap->find(string(topp->name()));
    if(gtopp != gend) {
      (*gtopp).second->set_topp( topp );
      Topp_m::iterator ganti = GToppMap->find(string(topp->anti_ptr()->name()));
      if(ganti != gend) {
	(*gtopp).second->set_anti( *(*ganti).second );
      }
    } // if gtopp != gend
  } // for Topp
}


//==============================================================================
// Constructors
//==============================================================================
GTopp::GTopp( const char* name ): iTopp( 0 ), anti_(0), anychrg(0) {
  InitGToppMap();
  GToppMap->insert( std::make_pair( string( name ), this ) );
}


//==============================================================================
// Constructors
//==============================================================================
GTopp::GTopp( const char* name, int chrg ): iTopp( 0 ), anti_(0), anychrg(chrg) {
  InitGToppMap();
  GToppMap->insert( std::make_pair( string( name ), this ) );
}


//==============================================================================
// ToppIf access methods
//==============================================================================
const char* const GTopp::name()  const { return (iTopp!=0) ? topp().name() : "Any"; }
float             GTopp::mass()  const { return topp().mass(); }
float             GTopp::wid()   const { return topp().wid(); }
float             GTopp::trbw()  const { return topp().trbw(); }
float             GTopp::life()  const { return topp().life(); }
int               GTopp::chrg()  const { return valid() ? topp().chrg() : anychrg; }
int               GTopp::lund()  const { return topp().lund(); }
bool              GTopp::valid() const { return iTopp!=0; }

//==============================================================================
// anti()
//==============================================================================
const ToppIf& GTopp::anti() const {
  return *anti_;
}

//==============================================================================
// set_anti()
//==============================================================================
void GTopp::set_anti(const GTopp& anti) {
  anti_ = &anti;
}

//==============================================================================
// set_topp
//==============================================================================
void GTopp::set_topp(const ArtePointer<Topp>& topp) { iTopp = getIDX<Topp>( topp ); }

//==============================================================================
// createPattern
//==============================================================================
PMC_sp GTopp::createPattern() const {
  // new needed because SPMC pointer is saved in SPMC CTOR!
  return PMC_sp(new SPatternMatchContainer(SPattern(*this)));
}

//==============================================================================
// swith
//==============================================================================
SP_sp GTopp::swith(const SProperty<TrackIf,bool>& rhs) const {
  return SP_sp(SPattern(*this, rhs));
}

//==============================================================================
// with
//==============================================================================
PMC_sp GTopp::with(const SProperty<TrackIf,bool>& rhs) const {
  // new needed because SPMC pointer is saved in SPMC CTOR!
  return PMC_sp(new SPatternMatchContainer(SPattern(*this,rhs)));
}

//==============================================================================
// operator==
//==============================================================================
bool GTopp::operator==(const ToppIf& rhs) const {
  //  return (strncmp(name(), rhs.name(), 13) == 0) ? true : false;
  return topp() == rhs;
}

//==============================================================================
// print
//==============================================================================
std::ostream& GTopp::print( std::ostream& os ) const {
  if( iTopp != 0 ) 
    os << setw(3) << getPTR<Topp>(iTopp) << " " << topp();
  else
    os << "No Topp link available!";
  return os;
}
