// *****************************************************************************
//
// source:
//
// type:      source code
//
// created:   28. Jan 1997
//
// author:    Thorsten Glebe
//            HERA-B Collaboration
//            Max-Planck-Institut fuer Kernphysik
//            Saupfercheckweg 1
//            69117 Heidelberg
//            Germany
//            E-mail: T.Glebe@mpi-hd.mpg.de
//
// Description: Class for Monte-Carlo tracks.
//
// changes:
// 28 Jan 1997 creation and 1st implementation by C. Borgmeier
// 03 Nov 2000 (TG) changed ArtePointers to references
// 10 Nov 2000 (TG) beginVertex(): changed to new ArteTable standard
// 20 Nov 2000 (TG) added namespace CLUE
// 06 Dez 2000 (TG) added code for CLUE relations
// 07 Dez 2000 (TG) copy constructor code added, operator= code added
// 12 Jan 2001 (TG) changed constructors
// 12 Jan 2001 (TG) removed beginVertex(), endVertex(), flight()
// 12 Jan 2001 (TG) removed obsolete include files
// 25 Jul 2001 (TG) changed rtrack from pointer to object, removed const from
//                  RelStreamer()
// 13 Aug 2001 (TG) added Streamer(), added TrackIf member functions
// 14 Aug 2001 (TG) added endVtx, RelStreamer(): added streaming of endVtx,
//                  removed Mother, removed set_bVertex(), set_topp(), bug fix
//                  in copy CTOR
// 15 Aug 2001 (TG) bug fix in Streamer, improved print()
// 17 Aug 2001 (TG) outsize() added
// 05 Sep 2001 (TG) RelStreamer: StreamVersion added, improved print(), removed
//                  length() (because not correct)
// 06 Sep 2001 (TG) added initialization of endTrack and Flag, improved print()
// 11 Sep 2001 (TG) bug fix in outsize(), moved outsize() to .icc file
// 13 Sep 2001 (TG) added Mother initialization, improved print()
//
// *****************************************************************************
#include <iostream>
#include <iomanip>
#include <cmath>

#include "clue/cluearte.hh"
#include "clue/MC_Track.hh"
#include "clue/MC_Vertex.hh"
#include "grelation/ClueRelation.hh"
#include "smatrix/SDistance.hh"

#ifdef USE_ROOT
#include <TPad.h>
#endif

#ifdef NAMESPACE
namespace CLUE {
#endif

  // dummy covariance matrices
  static const SMatrix<double,5> MCCOV;
  static const SMatrix<double,5> MCCINV;

  //==============================================================================
  // Constructors
  //==============================================================================
  MC_Track::MC_Track() :
    rtrack  (),
    endVtx  (),
    endTrack(),
    Px      (0.),
    Py      (0.),
    Pz      (0.),
    iTopp   (0),
    B_Vertex(0),
    Mother  (0),
    Flag    (0)
    {}

  MC_Track::MC_Track(const MC_Track& rhs) :
    rtrack  ( rhs.rtrack ),
    endVtx  ( rhs.endVtx ),
    endTrack( rhs.endTrack ),
    Px      ( rhs.Px ),
    Py      ( rhs.Py ),
    Pz      ( rhs.Pz ),
    iTopp   ( rhs.iTopp ),
    B_Vertex( rhs.B_Vertex ),
    Mother  ( rhs.Mother ),
    Flag    ( rhs.Flag )
    {}

  MC_Track::~MC_Track() {
    // Clue relations cannot unbook!
  }

  //==============================================================================
  // operator=
  //==============================================================================
  MC_Track& MC_Track::operator=(const MC_Track& rhs) {
    if(this == &rhs) return *this;

    rtrack   = rhs.rtrack;
    endVtx   = rhs.endVtx;
    endTrack = rhs.endTrack;
    Px       = rhs.Px;
    Py       = rhs.Py;
    Pz       = rhs.Pz;
    iTopp    = rhs.iTopp;
    B_Vertex = rhs.B_Vertex;
    Mother   = rhs.Mother;
    Flag     = rhs.Flag;

    return *this;
  }


  //==============================================================================
  // TrackIf access functions
  //==============================================================================
  float MC_Track::x()        const { return bVertex()->vx(); }
  float MC_Track::y()        const { return bVertex()->vy(); }
  float MC_Track::x(float Z) const { return x() + tx() * (Z - z()); }
  float MC_Track::y(float Z) const { return y() + ty() * (Z - z()); }
  float MC_Track::z()        const { return bVertex()->vz(); }
  float MC_Track::tx()       const { return Px/Pz; }
  float MC_Track::ty()       const { return Py/Pz; }
  float MC_Track::p()        const { return sqrt(Px*Px + Py*Py + Pz*Pz); }
  float MC_Track::pt()       const { return sqrt(Px*Px + Py*Py); }
  float MC_Track::pz()       const { return Pz; }
  float MC_Track::chi2()     const { return 0.; }
  unsigned short MC_Track::ndf() const { return 0; }
  float MC_Track::phi()      const { return 180. + 57.29577957 * atan2(ty(),tx()); }
  float MC_Track::theta()    const { return acos(evec()[2]); }
  float MC_Track::eta()      const { return -log(tan(theta()/2.)); }
  int   MC_Track::charge()   const { return topp().chrg(); }
  float MC_Track::energy(double ms) const { return sqrt( square(mass()) + Px*Px + Py*Py + Pz*Pz ); }
  float MC_Track::xf(double mass)  const { return (Pz - energy()*0.99898)/0.93827; }
  float MC_Track::rap(double mass) const { return 0.5*log((energy()+Pz)/(energy(mass)-Pz)); }
  
  SVector<double,3> MC_Track::xvec() const { return SVector<double,3>(x(),y(),z()); }

  SVector<double,3> MC_Track::tvec() const { return SVector<double,3>(tx(),ty(),1); }

  SVector<double,3> MC_Track::evec() const {
    return SVector<double,3>(1./mag(tvec()) * tvec());
  }

  SVector<double,3> MC_Track::pvec() const {
    return SVector<double,3>(Px,Py,Pz);
  }

  const SMatrix<double,5>& MC_Track::COV()  const { return MCCOV; }
  const SMatrix<double,5>& MC_Track::CINV() const { return MCCINV; }

  float MC_Track::cov_x(double dz) const { return 0; }
  float MC_Track::cov_y(double dz) const { return 0; }
  float MC_Track::cov_tx()         const { return 0; }
  float MC_Track::cov_ty()         const { return 0; }
  float MC_Track::cov_p()          const { return 0; }
  bool  MC_Track::isValid()        const { return true; }
  void  MC_Track::valid()   { return; }
  void  MC_Track::invalid() { return; }
  void  MC_Track::collect(vector<TrackIf*>& c) const {
    c.push_back(const_cast<MC_Track*>(this));
  }
  void MC_Track::propagate(const double zz) {
    return;
  }

  bool MC_Track::operator==( const TrackIf& rhs ) const {
    return &rhs == this;
  }

  //==============================================================================
  // print
  //==============================================================================
  std::ostream& MC_Track::print( std::ostream& os ) const {
    if(iTopp > 0) {
      string nam(topp().name());
      string tab(12-nam.length(),' ');
      os << nam << tab << " topp: " << setw(3) << getPTR<Topp>( iTopp );
    } else {
      os << "MC_Track:";
    }

    os.setf(ios::right,ios::adjustfield);
    os.setf(ios::scientific,ios::floatfield);
    os << " px:"     << setw(10) << setprecision(2) << Px 
       << " py:"     << setw(10) << setprecision(2) << Py 
       << " pz:"     << setw(10) << setprecision(2) << Pz
       << ", mass: " << setw( 9) << setprecision(2) << mass()
       << ", flag: " << setw(1)  << Flag;

    if( B_Vertex > 0 ) {
      os << " Begin-Vtx: " << setw(4) << bVertex();
    }
    if( Mother > 0 ) {
      os << " Mother: " << setw(4) << mother();
    }
    if( endVtx.size() > 0 ) {
      os << " End-Vtx: ";
      copy(endVtx.begin(), endVtx.end(), ostream_iterator<ArtePointer<MC_Vertex> >(os, " "));
    }
    if( endTrack.size() > 0 ) {
      os << " End-Tracks: ";
      copy(endTrack.begin(), endTrack.end(), ostream_iterator<ArtePointer<MC_Track> >(os, " "));
    }
    if( rtrack.size() > 0 ) {
      os << " RecoTrack rel: ";
      copy(rtrack.begin(), rtrack.end(), ostream_iterator<ArtePointer<RecoTrack> >(os, " "));
    }
    return os;
  }

  //==============================================================================
  // print_opt
  //==============================================================================
  void MC_Track::print_opt(char opt, const ArtePointer<MC_Track> mtr) const {
    if(!mtr.null()) {
      cout << mtr->name();
      switch(opt) {
      case 'f':
	cout << " " << Flag;
	break;
      case 'm':
	cout << mtr;
	break;
      case 'r':
	const ClueRelationIterator<RecoTrack> rend = mtr->rTrack().end();
	ClueRelationIterator<RecoTrack> last = rend; --last;
	for(ClueRelationIterator<RecoTrack> rit = mtr->rTrack().begin();
	    rit != rend; ++rit) {
	  cout << *rit;
	  if(rit != last) cout << ",";
	}
      } // switch
    } else {
      cout << name();
      switch(opt) {
      case 'f':
	cout << " " << Flag;
	break;
      case 'm':
	// nothing to be done if no ArtePointer available
	break;
      case 'r':
	const ConstClueRelationIterator<RecoTrack> rend = rTrack().end();
	ConstClueRelationIterator<RecoTrack> last = rend; --last;
	for(ConstClueRelationIterator<RecoTrack> rit = rTrack().begin();
	    rit != rend; ++rit) {
	  cout << *rit;
	  if(rit != last) cout << ",";
	}
      } // switch
    } // if
    return;
  }


  //==============================================================================
  // print_decay
  //==============================================================================
  void MC_Track::print_decay(unsigned int level, char opt,
			     const ArtePointer<MC_Track> mtra,
			     unsigned int lvl) const {
    if(lvl!=0 && endTrack.size() > 0) {
      cout << "[";
    }
    print_opt(opt,mtra);

    if(endTrack.size() > 0) {
      if(Flag >= 4) {
	// interaction, no decay
	cout << " => " << name() << " + " << endl;
      } else {
	// decay
	cout << " -> ";
      }

      const ConstClueRelationIterator<MC_Track> mend = endTrack.end();
      ConstClueRelationIterator<MC_Track> mlast = mend; --mlast;
      for( ConstClueRelationIterator<MC_Track> mit = endTrack.begin();
	   mit != mend; ++mit ) {
	const ArtePointer<MC_Track> mtr = *mit;
	if(level == lvl) {
	  print_opt(opt,mtr);
	} else {
	  mtr->print_decay(level, opt, mtr, lvl+1);
	}
	//	  cout << " " << mtr;
	if(mit == mlast) {
	  continue;
	}
	cout << " + ";
      } // for MC_Track

      if(lvl!=0)
	cout << "]";

    } else {
      cout << " ";
    } // if endTrack.size
  }

#ifdef USE_ROOT
  //==============================================================================
  // RelStreamer
  //==============================================================================
  /// only availabe in ROOT: call Streamer of relation container
  void MC_Track::RelStreamer(TBuffer& R__b, const Version_t& StreamVersion) {
    rTrack().Streamer(R__b);
    EndVtx().Streamer(R__b);
    if(StreamVersion > 2) {
      EndTrack().Streamer(R__b);
    }
  }

  //==============================================================================
  // Streamer
  //==============================================================================
  void MC_Track::Streamer(TBuffer &R__b) {
    // Stream an object of class MC_Track.
    
    if (R__b.IsReading()) {
      const Version_t R__v = R__b.ReadVersion();
      TObject::Streamer(R__b);
      R__b >> Px;
      R__b >> Py;
      R__b >> Pz;
      if (R__v < 2) {
	static float Mass;
	R__b >> Mass;
      }
      R__b >> iTopp;
      R__b >> B_Vertex;
      if (R__v < 2 || R__v > 4) {
	R__b >> Mother;
      }
      if (R__v > 3) {
	R__b >> Flag;
      }
    } else {
      R__b.WriteVersion(MC_Track::IsA());
      TObject::Streamer(R__b);
      R__b << Px;
      R__b << Py;
      R__b << Pz;
      R__b << iTopp;
      R__b << B_Vertex;
      R__b << Mother;
      R__b << Flag;
    }
  }
  
  //==============================================================================
  // Root::Print
  //==============================================================================
  void MC_Track::Print(Option_t *) {
    print( std::cout );
    std::cout << endl;
  }


  //==============================================================================
  // Root::Paint
  //==============================================================================
  // How to use: a) create canvas 
  //             b) create view via:  TView* view = new TView(1)
  //             c) loop over tracks and call Draw()
  // 
  void MC_Track::Paint(Option_t *option) {

    const ArtePointer<MC_Vertex> bv = bVertex();
    ArtePointer<MC_Vertex> ev;
    if(endVtx.size() > 0) {
      ev = *endVtx.begin();
    }

    if(bv.null() || ev.null()) return;

    Float_t xx[3] = { bv->vx(), bv->vy(), bv->vz() };
    Float_t yy[3] = { ev->vx(), ev->vy(), ev->vz() };
    
    gPad->PaintLine3D( xx, yy );
  }

  ClassImp(MC_Track);
#endif

#ifdef NAMESPACE
} // namespace CLUE
#endif
