// ********************************************************************
//
// source:
//
// type:      source code
//
// created:   27. Jan 1997
//
// author:    Thorsten Glebe
//            HERA-B Collaboration
//            Max-Planck-Institut fuer Kernphysik
//            Saupfercheckweg 1
//            69117 Heidelberg
//            Germany
//            E-mail: T.Glebe@mpi-hd.mpg.de
//
// Description: Class for handling reconstucted Tracks
//
// changes:
// 27 Jan 1997 creation by I. Siccama, C. Borgmeier
// 31 Okt 2000 (TG) removed contstructor using FORTRAN-Vt interface
// 20 Nov 2000 (TG) added namespace CLUE
// 06 Dec 2000 (TG) added code for CLUE relations
// 22 Dec 2000 (TG) improved RecoTrack::RecoTrack() initialization list
// 27 Dec 2000 (TG) end() optimization
// 12 Jan 2001 (TG) removed GetLorentzVector, decoupled class from MC_Track
// 22 Jan 2001 (TG) added ClMember pz, energy, xf
// 08 Feb 2001 (TG) added ClMember px,py,xfabs
// 18 Apr 2001 (TG) mommask, Rls, Rlt, Lse, Lsmu, Lspi, Lsk, Lsp added
// 19 Apr 2001 (TG) bug fix in print()
// 20 Apr 2001 (TG) removed lock(), unlock(), locked(), getRecoTrack()
// 20 Apr 2001 (TG) added Clmember rls, rlt, lse, lsmu, lspi, lsk, lsp, mom
// 26 Apr 2001 (TG) improvement in print()
// 28 Jun 2001 (TG) commented ClMembers
// 11 Jul 2001 (TG) ehit added (EcalHit relation)
// 25 Jul 2001 (TG) changed mtrack from pointer to object
// 27 Aug 2001 (TG) print() improved
// 05 Sep 2001 (TG) RelStreamer: StreamVersion added
//
// ********************************************************************

#include <iostream>
#include <iomanip>

#include "clue/cluearte.hh"
#include "clue/Topp.hh"
#include "clue/RecoTrack.hh"
#include "clue/TrackClones.hh"
#include "clue/MC_Track.hh"
#include "clue/EcalHit.hh"

#ifdef NAMESPACE
namespace CLUE {
#endif

  //==============================================================================
  // Constructors
  //==============================================================================
  RecoTrack::RecoTrack():
    RecoSegment(),
    mtrack (),
    clones (),
    rvertex(),
    ehit   (),
    Mass   (0),
    Ebrems (0),
    Cer    (0),
    iTopp  (0),

    Flr ( false ),
    Flt ( false ),
    Fle ( false ),
    Flm ( false ),
	  
    Lre ( 0. ),
    Lrmu( 0. ),
    Lrpi( 0. ),
    Lrk ( 0. ),
    Lrp ( 0. ),
  	  
    Lte ( 0. ),
    Lth ( 0. ),
	  
    Lee ( 0. ),
    Lemi( 0. ),
    Leh ( 0. ),
	  
    Lmmu( 0. ),
    Lmh ( 0. ),
	  
    Lmu ( 0. ),
    Lpi ( 0. ),
    Lk  ( 0. ),
    Lp  ( 0. ),
	  
    Fit ( 0 ),
	  
    Hsi ( 0 ),
    Hin ( 0 ),
    Hou ( 0 ),
    Hri ( 0 ),
    Hca ( 0 ),
    Htr ( 0 ),
    Hmu ( 0 ),

    Mommask ( 0 ),
    
    Rls ( 0 ),
    Rlt ( 0 ),

    Lse ( 0. ),
    Lsmu( 0. ),
    Lspi( 0. ),
    Lsk ( 0. ),
    Lsp ( 0. ) {
    //cout << "default constructing " << this << endl;
  }

  RecoTrack::~RecoTrack() {
    // Clue relations cannot unbook!
  }
  
  RecoTrack::RecoTrack( const RecoTrack& rhs ) :
    RecoSegment( rhs ),
    mtrack (rhs.mtrack),
    clones (rhs.clones),
    rvertex(rhs.rvertex),
    ehit   (rhs.ehit),
    Mass   (rhs.Mass),
    Ebrems (rhs.Ebrems), 
    Cer    (rhs.Cer),
    iTopp  (rhs.iTopp),

    Flr ( rhs.Flr ),
    Flt ( rhs.Flt ),
    Fle ( rhs.Fle ),
    Flm ( rhs.Flm ),

    Lre ( rhs.Lre ),
    Lrmu( rhs.Lrmu),
    Lrpi( rhs.Lrpi),
    Lrk ( rhs.Lrk ),
    Lrp ( rhs.Lrp ),
  
    Lte ( rhs.Lte ),
    Lth ( rhs.Lth ),

    Lee ( rhs.Lee ),
    Lemi( rhs.Lemi),
    Leh ( rhs.Leh ),

    Lmmu( rhs.Lmmu),
    Lmh ( rhs.Lmh ),

    Lmu ( rhs.Lmu ),
    Lpi ( rhs.Lpi ),
    Lk  ( rhs.Lk  ),
    Lp  ( rhs.Lp  ),

    Fit ( rhs.Fit ),

    Hsi ( rhs.Hsi ),
    Hin ( rhs.Hin ),
    Hou ( rhs.Hou ),
    Hri ( rhs.Hri ),
    Hca ( rhs.Hca ),
    Htr ( rhs.Htr ),
    Hmu ( rhs.Hmu ),

    Mommask ( rhs.Mommask ),

    Rls ( rhs.Rls ),
    Rlt ( rhs.Rlt ),

    Lse ( rhs.Lse ),
    Lsmu( rhs.Lsmu ),
    Lspi( rhs.Lspi ),
    Lsk ( rhs.Lsk ),
    Lsp ( rhs.Lsp ) {
  
    // cout << "copy constructing " << this << " from " << &rhs << endl;
  }


  //==============================================================================
  // operator =
  //==============================================================================
  const RecoTrack& RecoTrack::operator=( const RecoTrack& rhs ) {
    if( this == &rhs ) return *this;

    //cout << "assign from " << &rhs << " to " << this << endl;
    RecoSegment::operator=( rhs );

    mtrack  = rhs.mtrack;
    clones  = rhs.clones;
    rvertex = rhs.rvertex;
    ehit    = rhs.ehit;

    Mass   = rhs.Mass;
    Ebrems = rhs.Ebrems;
    Cer    = rhs.Cer;
    iTopp  = rhs.iTopp;

    Flr    = rhs.Flr ;
    Flt    = rhs.Flt ;
    Fle    = rhs.Fle ;
    Flm    = rhs.Flm ;

    Lre    = rhs.Lre ;
    Lrmu   = rhs.Lrmu;
    Lrpi   = rhs.Lrpi;
    Lrk    = rhs.Lrk ;
    Lrp    = rhs.Lrp ;
    
    Lte    = rhs.Lte ;
    Lth    = rhs.Lth ;

    Lee    = rhs.Lee ;
    Lemi   = rhs.Lemi;
    Leh    = rhs.Leh ;

    Lmmu   = rhs.Lmmu;
    Lmh    = rhs.Lmh ;
  
    Lmu    = rhs.Lmu;
    Lpi    = rhs.Lpi;
    Lk     = rhs.Lk;
    Lp     = rhs.Lp;
    
    Fit    = rhs.Fit;
    
    Hsi    = rhs.Hsi;
    Hin    = rhs.Hin;
    Hou    = rhs.Hou;
    Hri    = rhs.Hri;
    Hca    = rhs.Hca;
    Htr    = rhs.Htr;
    Hmu    = rhs.Hmu;

    Mommask = rhs.Mommask;

    Rls    = rhs.Rls;
    Rlt    = rhs.Rlt;

    Lse    = rhs.Lse;
    Lsmu   = rhs.Lsmu;
    Lspi   = rhs.Lspi;
    Lsk    = rhs.Lsk;
    Lsp    = rhs.Lsp;

    return *this;
  }


  //==============================================================================
  // energy()
  //==============================================================================
  float RecoTrack::energy(double mass) const { 
    return (mass==0.) ? 
      sqrt( Mass*Mass + p()*p() ) :
      sqrt( mass*mass + p()*p() );
  }

  //==============================================================================
  // topp()
  //==============================================================================
  const ArtePointer<Topp> RecoTrack::topp() const {
    return getPTR<Topp>( iTopp );
  }


  //==============================================================================
  // set_topp()
  //==============================================================================
  void RecoTrack::set_topp(const ArtePointer<Topp>& topp) {
    iTopp = getIDX<Topp>( topp );
  }


  //==============================================================================
  // print
  //==============================================================================
  std::ostream& RecoTrack::print( std::ostream& os ) const {

    RecoSegment::print( os );
    os << "  ";

    os.setf(ios::left,ios::adjustfield);
    os.setf(ios::fixed,ios::floatfield);
    if(mass() != 0)
      os << "mass: "   << setw(6) << setprecision(2) << mass() << " ";
    if(ebrems() != 0)
      os << "Ebrem: "  << setprecision(2) << ebrems() << " ";
    if(cer() > 0)
      os << "Cer: " << setprecision(3) << cer() << " ";
    if( iTopp != 0 )
      os << " Topp: " << topp()->name();

    os << " Mom: " << mommask();

    if( !mTrack().empty() ) {
      os << " MC_Track: ";
      const ConstClueRelationIterator<MC_Track> mend = mTrack().end();
      for( ConstClueRelationIterator<MC_Track> mT = mTrack().begin();
	   mT != mend; ++mT ) {
	const ArtePointer<MC_Track> mct = *mT;
	
	os << setw(4) << mct << " " << mct->topp().name();
      }
      os << " ";
    }

    // print the clones
    if( !Clones().empty() ) {
      os << " clones: ";

      const ConstClueRelationIterator<TrackClones> cend = Clones().end();
      for( ConstClueRelationIterator<TrackClones> tc = Clones().begin();
	   tc != cend; ++tc ) {
	const ArtePointer<TrackClones> trcl = *tc;
	os << trcl->type();

	const ClueRelationIterator<RecoTrack> rend = trcl->rtrack.end();
	for( ClueRelationIterator<RecoTrack> rt = trcl->rtrack.begin();
	     rt != rend; ++rt ) {

	  if( &**rt != this ) {
	    os << *rt;
	  }
	}
	os << " ";
      } // for ClueRelationIterator
    } // if !clones.empty()

    return os;
  }

  //==============================================================================
  // printMC
  //==============================================================================
  void RecoTrack::printMC() const {
    const ConstClueRelationIterator<MC_Track> end = mTrack().end();
    for( ConstClueRelationIterator<MC_Track> m = mTrack().begin();
	 m != end; ++m ) {
      std::cout << "MC:" << setw(3) << *m << " " << **m << endl;
    }
  }
  

#ifdef USE_ROOT
  //==============================================================================
  // RelStreamer
  //==============================================================================
  /// only availabe in ROOT: call Streamer of relation container
  void RecoTrack::RelStreamer(TBuffer& R__b, const Version_t& StreamVersion) {
    mTrack().Streamer(R__b);
    Clones().Streamer(R__b);
    rVertex().Streamer(R__b);
    Ehit().Streamer(R__b);
  }

  //==============================================================================
  // Print
  //==============================================================================
  void RecoTrack::Print( Option_t* ) {
    print( std::cout );
    std::cout << endl;
  }


  //==============================================================================
  // Streamer
  //==============================================================================
  void RecoTrack::Streamer(TBuffer &R__b) {
   // Stream an object of class RecoTrack.

    if (R__b.IsReading()) {
      Version_t R__v = R__b.ReadVersion();
      RecoSegment::Streamer(R__b);
      R__b >> Mass;
      R__b >> Ebrems;
      R__b >> Cer;
      R__b >> iTopp;
      R__b >> Flr;
      R__b >> Flt;
      R__b >> Fle;
      R__b >> Flm;
      R__b >> Lre;
      R__b >> Lrmu;
      R__b >> Lrpi;
      R__b >> Lrk;
      R__b >> Lrp;
      R__b >> Lte;
      R__b >> Lth;
      R__b >> Lee;
      R__b >> Lemi;
      R__b >> Leh;
      R__b >> Lmmu;
      R__b >> Lmh;
      R__b >> Lmu;
      R__b >> Lpi;
      R__b >> Lk;
      R__b >> Lp;
      R__b >> Fit;
      R__b >> Hsi;
      R__b >> Hin;
      R__b >> Hou;
      R__b >> Hri;
      R__b >> Hca;
      R__b >> Htr;
      R__b >> Hmu;
      if (R__v > 2) {
	R__b >> Mommask;
	R__b >> Rls;
	R__b >> Rlt;
	R__b >> Lse;
	R__b >> Lsmu;
	R__b >> Lspi;
	R__b >> Lsk;
	R__b >> Lsp;
      }
    } else {
      R__b.WriteVersion(RecoTrack::IsA());
      RecoSegment::Streamer(R__b);
      R__b << Mass;
      R__b << Ebrems;
      R__b << Cer;
      R__b << iTopp;
      R__b << Flr;
      R__b << Flt;
      R__b << Fle;
      R__b << Flm;
      R__b << Lre;
      R__b << Lrmu;
      R__b << Lrpi;
      R__b << Lrk;
      R__b << Lrp;
      R__b << Lte;
      R__b << Lth;
      R__b << Lee;
      R__b << Lemi;
      R__b << Leh;
      R__b << Lmmu;
      R__b << Lmh;
      R__b << Lmu;
      R__b << Lpi;
      R__b << Lk;
      R__b << Lp;
      R__b << Fit;
      R__b << Hsi;
      R__b << Hin;
      R__b << Hou;
      R__b << Hri;
      R__b << Hca;
      R__b << Htr;
      R__b << Hmu;
      R__b << Mommask;
      R__b << Rls;
      R__b << Rlt;
      R__b << Lse;
      R__b << Lsmu;
      R__b << Lspi;
      R__b << Lsk;
      R__b << Lsp;
    }
  }

  ClassImp(RecoTrack)
#endif

#ifdef NAMESPACE
} // namespace CLUE
#endif
