// *********************************************************************
//
// source:
//
// type:      source code
//
// created:   02. Apr 1997
//
// author:    Thorsten Glebe
//            HERA-B Collaboration
//            Max-Planck-Institut fuer Kernphysik
//            Saupfercheckweg 1
//            69117 Heidelberg
//            Germany
//            E-mail: T.Glebe@mpi-hd.mpg.de
//
// Description: Class for a reconstructed vertex
//
// changes:
// 02 Apr 1997 creation by I. Siccama, C. Borgmeier
// 30 Okt 2000 (TG) removed FORTRAN-VT
// 06 Nov 2000 (TG) removed call to removed function addVertex()
// 20 Nov 2000 (TG) added namespace CLUE
// 06 Dec 2000 (TG) added code for CLUE relations, removed begin(), end()
//                  member functions
// 27 Dec 2000 (TG) end() optimization
// 11 Feb 2001 (TG) added members Chi2, Ndf, Mass; removed inheritance
//                  from RecoTrack
// 16 Mar 2001 (TG) ntracks added
// 16 Mar 2001 (TG) print_cov() added, bug fix in copy constructor: added
//                  copying of covariance matrix
// 24 Apr 2001 (TG) isfit replaced by valid_, replaced valid() function
// 25 Apr 2001 (TG) removed lock(), unlock(), locked(), modified prob()
// 09 Jul 2001 (TG) removed ntrack(), changed type of ntracks
// 27 Jul 2001 (TG) added Streamer
// 05 Sep 2001 (TG) RelStreamer: StreamVersion added
//
// **********************************************************************
#include <algorithm>
#include <iostream>
#include <iomanip>

#ifdef USE_ROOT
#include <TPolyMarker3D.h>
#include "geometry/Point3D.hh"   // for Paint()
#endif

#include "clue/cluearte.hh"
#include "clue/RecoVertex.hh"
#include "clue/RecoTrack.hh"
#include "smatrix/Prob.hh"

#ifdef NAMESPACE
namespace CLUE {
#endif

  //==============================================================================
  // Constructors
  //==============================================================================
  RecoVertex::RecoVertex() :
    out(),
    Vx     ( 0.),
    Vy     ( 0.),
    Vz     ( 0.),
    Chi2   ( 0.),
    Ndf    ( 0 ),
    Mass   ( 0.),
    Ntracks( 0 ),
    flag   ( 1 ),
    fitFlag( 0 ),
    momFlag( 0 ),
    valid_ ( true ) {
    //cout << "empty vertex " << this << endl;
    for( unsigned int i=0; i<6; ++i ) covMatrix[i] = 0.;
  }


  RecoVertex::RecoVertex( const RecoVertex& rhs ):
    out    ( rhs.out ),
    Vx     ( rhs.Vx ),
    Vy     ( rhs.Vy ),
    Vz     ( rhs.Vz ),
    Chi2   ( rhs.Chi2 ),
    Ndf    ( rhs.Ndf ),
    Mass   ( rhs.Mass ),
    Ntracks( rhs.Ntracks ),
    flag   ( rhs.flag ),
    fitFlag( rhs.fitFlag ),
    momFlag( rhs.momFlag ),
    valid_ ( rhs.valid_ )
    {
      for( unsigned int i=0; i<6; ++i ) covMatrix[i] = rhs.covMatrix[i];
    }
  
  RecoVertex::~RecoVertex() {
    // Clue relations cannot unbook!
  }


  //==============================================================================
  // operator =
  //==============================================================================
  const RecoVertex& RecoVertex::operator=( const RecoVertex& rhs ) {
    //cout << "copy " << iv << " from " << rhs.iv << endl;
    if( this == &rhs ) return *this;
    
    out       = rhs.out;
    Vx        = rhs.Vx;
    Vy        = rhs.Vy;
    Vz        = rhs.Vz;
    Chi2      = rhs.Chi2;
    Ndf       = rhs.Ndf;
    Mass      = rhs.Mass;

    for( unsigned int i=0; i<6; ++i ) covMatrix[i] = rhs.covMatrix[i];
    Ntracks   = rhs.Ntracks;
    flag      = rhs.flag;
    fitFlag   = rhs.fitFlag;
    momFlag   = rhs.momFlag;
    valid_    = rhs.valid_;
    
    return *this;
  }


  //==============================================================================
  // VertexIf Access methods
  //==============================================================================
  float RecoVertex::vx()   const { return Vx; }
  float RecoVertex::vy()   const { return Vy; }
  float RecoVertex::vz()   const { return Vz; }
  float RecoVertex::chi2() const { return Chi2; }
  unsigned short RecoVertex::ndf() const { return Ndf; }
  unsigned short int RecoVertex::ntracks() const { return Ntracks; }
  float RecoVertex::vtx_cov_x()   const { return covMatrix[0]; }
  float RecoVertex::vtx_cov_y()   const { return covMatrix[2]; }
  float RecoVertex::vtx_cov_z()   const { return covMatrix[5]; }
  
  const TrackIf* RecoVertex::track(unsigned int i) const {
    return (i>static_cast<unsigned int>(out.size())) ? 0 : &(*out[i]);
  }

  const TrackIf*& RecoVertex::track(unsigned int i) {
    static const TrackIf* nullptr = 0;
    cerr << "Modification of a Relation is not allowed!" << endl;
    return nullptr;
  }

  //==============================================================================
  // print
  //==============================================================================
  std::ostream& RecoVertex::print( std::ostream& os ) const {
    
    os.setf(ios::right,ios::adjustfield);
    os.setf(ios::fixed,ios::floatfield);
    os << "Point: " 
       << setw(7) << setprecision(2) << vx() << ","
       << setw(7) << setprecision(2) << vy() << ","
       << setw(7) << setprecision(2) << vz();

    os.setf(ios::left,ios::adjustfield);
    os << " mass: " << setw(5) << setprecision(2) << mass() << " GeV"
       << " chi2: " << setw(5) << setprecision(2) << chi2()
       << " ndf: "  << setw(3) << ndf()
       << " prob: " << setw(4) << setprecision(3) << prob()
       << endl;
    os << "tracks in vertex: " << setw(4) << ntracks() << endl;

    print_cov(os);

#ifdef XXX
    os << "Incoming track: ";
    if( flag != 0 ) os << " *INVALID* " << endl;
    //      else {
    //        RecoTrack::print( os );
    //      }
#endif

    if( out.size() > 0 ) {
      os << " Outgoing tracks: ";
      const ConstClueRelationIterator<RecoTrack> end = out.end();
      for( ConstClueRelationIterator<RecoTrack> rTrack = out.begin();
	   rTrack != end; ++rTrack ) {
	os << *rTrack << ' ';
      }
      os << endl;
    }
    return os;
  }


  //==============================================================================
  // print_cov
  //==============================================================================
  std::ostream& RecoVertex::print_cov( std::ostream& os ) const {
    os << " Cov. Matrix: " << endl;
    os.setf(ios::right,ios::adjustfield);
    os.setf(ios::scientific,ios::floatfield);
    os << setw(12) << setprecision(3) << vtx_cov_x() << endl
       << setw(12) << setprecision(3) << vtx_cov_x_y()
       << setw(12) << setprecision(3) << vtx_cov_y() << endl
       << setw(12) << setprecision(3) << vtx_cov_x_z()
       << setw(12) << setprecision(3) << vtx_cov_y_z()
       << setw(12) << setprecision(3) << vtx_cov_z();

    return os;
  }

  //==============================================================================
  // prob
  //==============================================================================
  float RecoVertex::prob() const {
    return Prob(Chi2, Ndf);
  }

  
#ifdef USE_ROOT
  //==============================================================================
  // RelStreamer
  //==============================================================================
  void RecoVertex::RelStreamer(TBuffer& R__b, const Version_t& StreamVersion) {
    Out().Streamer(R__b);
  }


  //==============================================================================
  // Streamer
  //==============================================================================
  void RecoVertex::Streamer(TBuffer &R__b) {
    // Stream an object of class RecoVertex.
    
    if (R__b.IsReading()) {
      Version_t R__v = R__b.ReadVersion();
      TObject::Streamer(R__b);
      if (R__v < 2) {
	out.Streamer(R__b);
	out.clear(); // clear relations, because they are read at the end
      }
      R__b >> Vx;
      R__b >> Vy;
      R__b >> Vz;
      R__b >> Chi2;
      R__b >> Ndf;
      R__b >> Mass;
      R__b.ReadStaticArray(covMatrix);
      R__b >> Ntracks;
      R__b >> flag;
    } else {
      R__b.WriteVersion(RecoVertex::IsA());
      TObject::Streamer(R__b);
      R__b << Vx;
      R__b << Vy;
      R__b << Vz;
      R__b << Chi2;
      R__b << Ndf;
      R__b << Mass;
      R__b.WriteArray(covMatrix, 6);
      R__b << Ntracks;
      R__b << flag;
    }
  }

  //==============================================================================
  // Clear
  //==============================================================================
  void RecoVertex::Clear(Option_t *) {
    flag = 1;
  }
  

  //==============================================================================
  // Print
  //==============================================================================
  void RecoVertex::Print(Option_t *) {
    print( std::cout );
    std::cout << endl;
  }


  //==============================================================================
  // Root::Paint
  //==============================================================================
  // How to use: a) create canvas 
  //             b) create view via:  TView* view = new TView(1)
  //             c) loop over tracks and call Draw()
  // 
  void RecoVertex::Paint(Option_t *option) {
    //    Float_t xx[1] = { thisPoint.x() };
    //    Float_t yy[1] = { thisPoint.y() };
    
    //    gPad->PaintPolyMarker(1, xx, yy);
    
    cout << "Set A TPolyMarker" << endl;
    
    // !!!!!! This is a memory leak!
    // (I don't know where to delete m...)
    // in a root kumac, gPad->Clear() deletes this pointer.
    //    TPolyMarker3D* m = new TPolyMarker3D();
    
    //    m->SetPoint(0, thisPoint.x(), thisPoint.y(), thisPoint.z());
    //    m->SetMarkerSize(2);
    Point3D thisPoint(vx(), vy(), vz());
    thisPoint.Paint();
    //  m->SetMarkerColor(4);
    //  m->SetMarkerStyle(3);
    
    //  m->Draw();
  }

  ClassImp(RecoVertex)
#endif // USE_ROOT

#ifdef NAMESPACE
} // namespace CLUE
#endif
