#ifndef __AnalysisChain_HH
#define __AnalysisChain_HH
// ********************************************************************
//
// source:
//
// type:      source code
//
// created:   28. Feb 2001
//
// author:    Thorsten Glebe
//            HERA-B Collaboration
//            Max-Planck-Institut fuer Kernphysik
//            Saupfercheckweg 1
//            69117 Heidelberg
//            Germany
//            E-mail: T.Glebe@mpi-hd.mpg.de
//
// Description: Class to manage execution of several AnalysisIf objects
//
// changes:
// 28 Feb 2001 (TG) creation
//
// ********************************************************************
#include <iostream>
#include <vector>
#include <pthread.h>
#include "interfaces/AnalysisIf.hh"

/** Class to manage execution of a series of AnalysisIf objects.
    
    @author T. Glebe
*/
//==============================================================================
// AnalysisChain
//==============================================================================
class AnalysisChain {
  typedef vector<AnalysisIf*>   Ana_l;
  typedef Ana_l::iterator       Ana_it;
  typedef Ana_l::const_iterator Ana_cit;
  typedef vector<pthread_t*>    Pth_l;
  typedef Pth_l::iterator       Pth_it;
  typedef Pth_l::const_iterator Pth_cit;

public:
  /** @name --- Constructors --- */
  ///
  AnalysisChain() {}
  ///
  virtual ~AnalysisChain() {
    const Pth_cit cend = thread.end();
    for(Pth_cit it = thread.begin(); it != cend; ++it)
      delete *it;
    thread.clear();
  }

  /** @name --- Access methods --- */
  ///
  void push_back(AnalysisIf& ana) {
    chain.push_back(&ana);
    pthread_t* thr = new pthread_t;
    thread.push_back(thr);
  }

  ///
  bool EventSelect() const { return true; }

  ///
  bool EventCleanup() {
    const Ana_cit end = chain.end();
    for(Ana_cit it = chain.begin(); it != end; ++it) {
      (*it)->EventCleanup();
    }
    return true;
  }

  ///
  bool End() {
    const Ana_cit end = chain.end();
    for(Ana_cit it = chain.begin(); it != end; ++it) {
      (*it)->End();
    }
    return true;
  }

  ///
  bool DefineHistograms() {
    const Ana_cit end = chain.end();
    for(Ana_cit it = chain.begin(); it != end; ++it) {
      (*it)->DefineHistograms();
    }
    return true;
  }

  ///
  bool DefinePattern() {
    const Ana_cit end = chain.end();
    for(Ana_cit it = chain.begin(); it != end; ++it) {
      (*it)->DefinePattern();
    }
    return true;
  }

  ///
  bool Event() {
    const Ana_cit end = chain.end();
    for(Ana_cit it = chain.begin(); it != end; ++it) {
      AnalysisIf& a = **it;
      if(!a.EventSelect()) continue;
      a.Event();
    }
    return true;
  }

  ///
  bool Event_Thread() {
    //    cout << " Start Thread " << threadno << ": " << chain[threadno] << endl;
    AnalysisIf& a = *chain[threadno];
    ++threadno;

    if(!a.EventSelect()) return false;
    a.Event();
    return true;
  }
  
  ///
  bool ThreadEvent() {
    threadno = 0;

    const unsigned int csize = chain.size();
    for(unsigned int i=0; i < csize; ++i) {
      pthread_create(thread[i], NULL, &AnalysisChain::RunThread, static_cast<void*>(this));
    }

    const Pth_cit cend = thread.end();
    for(Pth_cit it = thread.begin(); it != cend; ++it) {
      pthread_join(**it, 0);
    }

    return true;
  }

private:

  static void* RunThread(void* arg) {
    static_cast<AnalysisChain*>(arg)->Event_Thread();
    return 0;
  }

  Ana_l        chain;
  Pth_l        thread;
  unsigned int threadno;
};
#endif
