#ifndef MC_VERTEX_HH
#define MC_VERTEX_HH
// *****************************************************************************
//
// source:
//
// type:      source code
//
// created:   27. Jan 1997
//
// author:    Thorsten Glebe
//            HERA-B Collaboration
//            Max-Planck-Institut fuer Kernphysik
//            Saupfercheckweg 1
//            69117 Heidelberg
//            Germany
//            E-mail: T.Glebe@mpi-hd.mpg.de
//
// Description: Class for Monte-Carlo tracks.
//
// changes:
// 27 Jan 1997 creation and 1st implementation by I. Siccama and C. Borgmeier
// 03 Nov 2000 (TG) changed ArtePointers to references
// 20 Nov 2000 (TG) added namespace CLUE
// 11 Jan 2001 (TG) removed dependency from Point3D
// 05 Apr 2001 (TG) xvec() added
// 13 Aug 2001 (TG) added VertexIf inheritance
// 14 Aug 2001 (TG) added MC_Track relations, added RelStreamer(), added
//                  Mother, mother(), incremented class version, added copy
//                  CTOR, added operator=()
// 15 Aug 2001 (TG) added Paint()
// 05 Sep 2001 (TG) RelStreamer: StreamVersion added
//
// *****************************************************************************

#include <iosfwd>
#include "arte/TObject.hh"
#include "interfaces/VertexIf.hh"
#include "arte/artefwd.hh"
#include "clue/cluefwd.hh"
#include "grelation/ClueRelation.hh"

#ifndef __CINT__
#include "smatrix/SVector.hh"
#endif

#ifdef NAMESPACE
namespace CLUE {
#endif
  class abc; // to fake doc++

  /**
     Monte Carlo Vertex
     @author C. Borgmeier, T. Glebe
  */
  //==============================================================================
  // MC_Vertex
  //==============================================================================
  class MC_Vertex: public TObject, public VertexIf {
  public:
    ///
    MC_Vertex();
    ///
    MC_Vertex( const MC_Vertex& rhs );
#ifndef NO_ARTE
    /// only available in ARTE
    MC_Vertex( const ArtePointer<MVER>& mv );
#endif
    ///
    ~MC_Vertex();
    ///
    MC_Vertex& operator=( const MC_Vertex& );
  
    /// used by operator<<()
    std::ostream& print( std::ostream& ) const;

    /** @name --- VertexIf access methods --- */
    /// $x$ of vertex
    float vx() const;
    /// $y$ of vertex
    float vy() const;
    /// $z$ of vertex
    float vz() const;
    /// dummy function: return 0
    float chi2() const;
    /// dummy function: return 0
    unsigned short int ndf() const;
    /// dummy function: return 1
    float prob() const;
    /// no of tracks in vertex
    unsigned short int ntracks() const;
    /// returns pointer to track in vertex (slow access via relation!)
    const TrackIf* track(unsigned int i) const;
    /// returns null pointer + prints warning
    const TrackIf*& track(unsigned int i);

#ifndef __CINT__
  /// $\vec{x} = (x,y,z)$
  SVector<double,3> vpos() const;
  /// dummy function: return zero matrix
  const SMatrix<double,3>& VCOV() const;
#endif

    /// dummy function: return 0
    float vtx_cov_x()   const;
    /// dummy function: return 0
    float vtx_cov_y()   const;
    /// dummy function: return 0
    float vtx_cov_z()   const;

    /** @name --- Access methods --- */
    /// return pointer to mother track
    const ArtePointer<MC_Track>  mother()  const;
    /// set pointer to mother track
    void set_mother(const ArtePointer<MC_Track>& p);

    /** @name --- CLUE relations --- */
    /// MC\_Vertex $\leftrightarrow$ MC\_Track (relations to outgoing tracks)
    ClueRelationList<MC_Track> out;    //! non persistent
    ///
    const ClueRelationList<MC_Track>& Out() const;
    ///
    ClueRelationList<MC_Track>& Out();
    ///
    ClueRelationList<MC_Track>&
    container(const ArtePointer<MC_Track>& p);
    
  private:
    Float_t X;
    Float_t Y;
    Float_t Z; 
    UShort_t Mother;     // MC_Track pointer of incoming track

#ifdef USE_ROOT
public:
    /// only availabe in ROOT: call Streamer of relation container
    void RelStreamer(TBuffer& R__b, const Version_t& StreamVersion);
    /// only available in ROOT
    void Print(Option_t *option="");
    /// only available in ROOT
    void Paint( Option_t *option="" );

    ClassDef(MC_Vertex,2)
#endif
  }; // end of class MC_Vertex


  //==============================================================================
  // operator <<
  //==============================================================================
  std::ostream& operator<<( std::ostream& os, const MC_Vertex& mv);

#ifdef NAMESPACE
} // namespace CLUE
#endif

#ifndef __CINT__
#include "MC_Vertex.icc"
#endif
#endif
