#ifndef RECO_SEGMENT_HH
#define RECO_SEGMENT_HH
// *****************************************************************************
//
// source:
//
// type:      source code
//
// created:   27. Jan 1997
//
// author:    Thorsten Glebe
//            HERA-B Collaboration
//            Max-Planck-Institut fuer Kernphysik
//            Saupfercheckweg 1
//            69117 Heidelberg
//            Germany
//            E-mail: T.Glebe@mpi-hd.mpg.de
//
// Description: Reconstructed Track class
//
// changes:
// 27 Jan 1997 created by I. Siccama, C. Borgmeier
// 31 Okt 2000 (TG) removed constructors using FORTRAN-Vt interface calls
// 31 Okt 2000 (TG) removed functions using CLHEP: merge(), getPar(), setPar(), getCov()
// 03 Nov 2000 (TG) changed ArtePointers to references
// 20 Nov 2000 (TG) added namespace CLUE
// 15 Dec 2000 (TG) modified types of Rtra, Ecal, Nf
// 22 Dec 2000 (TG) changed type of Rtra, Ecal, Nf to UShort_t
// 12 Jan 2001 (TG) removed propagate() member function, print() changed
// 12 Jan 2001 (TG) removed GetLorentzVector(), mass() and massDiff() functions
//                  to decouple RecoSegment from other classes
// 08 Feb 2001 (TG) added px(), py()
// 27 Feb 2001 (TG) inherited RecoSegment from TrackIf
// 15 Mar 2001 (TG) EoP(), x(z), y(z) added
// 16 Mar 2001 (TG) print_cov() added
// 05 Apr 2001 (TG) xvec(), tvec(), evec(), pvec() added
// 10 Apr 2001 (TG) CINV_mom(), CINV_nomom(), CINV_, CINV2_, CINV(),
//                  FillCINV(), COV() added
// 17 Apr 2001 (TG) CINV2_, CINV_mom(), CINV_nomom() removed
// 20 Apr 2001 (TG) moved inline functions to RecoSegment.icc
// 27 Apr 2001 (TG) xfabs() added
// 02 Mai 2001 (TG) phi(), theta(), eta() added
// 11 Jun 2001 (TG) added member COV_, changed COV(), COV_ is filled in FillCINV()
// 11 Jun 2001 (TG) added cov_p()
// 29 Jun 2001 (TG) added energy(), xf(), rap(), moved SPatternMatch Relation from
//                  RecoTrack to here
// 05 Jul 2001 (TG) collect() added
// 09 Jul 2001 (TG) lock(), unlock(), locked() replaced by valid(), invalid(), isValid()
// 10 Jul 2001 (TG) operator==() added
// 11 Jul 2001 (TG) removed ecal(), set_ecal(), Ecal
// 16 Jul 2001 (TG) removed EoP()
// 31 Aug 2001 (TG) added PrepareForVertexing()
// 13 Sep 2001 (TG) renamed *_pinv() to *_p()
// 14 Sep 2001 (TG) removed test() member function
//
// *****************************************************************************

#include <iosfwd>
#include "arte/TObject.hh"
#include "interfaces/TrackIf.hh"
#include "arte/artefwd.hh"
#include "clue/cluefwd.hh"
#include "clue/RsegC.hh"

#ifndef __CINT__
#include "grelation/GRelationList.hh"
class SPatternMatch;
#endif

#ifdef NAMESPACE
namespace CLUE {
#endif
  class abc; // to fake doc++

  /** Reconstructed track class.
      In contrast to \texttt{ARTE}, the momentum covariance matrix elements are
      computed with the momentum $p$ and not with $Q/p$. The reason is that
      they enter in this way into the vertexing (Vt based vertexing codes).
      The transformation $cov_{ij}(Q/p) \to cov_{ij}(p)$ is done before writing
      the \texttt{BEE} files in the \texttt{ARTE} project \texttt{cluearte}.
      The \texttt{RecoSegment} memberfunctions \texttt{sPinvToP()} and
      \texttt{sPtoPinv()} do the appropriate transformation in the respective
      direction.

      @author C. Borgmeier, T. Glebe
  */
  //==============================================================================
  // RecoSegment
  //==============================================================================
  class RecoSegment : public TObject, public TrackIf {
  public:
    /** @name --- Constructors ---*/
    ///
    RecoSegment();
    ///
    RecoSegment( const RecoSegment& rhs );
    ///
    ~RecoSegment();
    ///
    const RecoSegment& operator=( const RecoSegment& rhs );

#ifndef NO_ARTE
    /** @name --- ARTE only functions --- */
    /// fill RecoSegment from RSEG or RTRA
    template<class T> void convert( T rtra );

    /// create object from RTRA table
    RecoSegment( const ArtePointer<RTRA>& r );

    /// create object from RSEG table
    RecoSegment( const ArtePointer<RSEG>& r );

    /// return RTRA pointer
    ArtePointer<RTRA> rtra() const;

    /// set RTRA pointer
    void set_rtra(const ArtePointer<RTRA>& rtra);
#else
    /// not available in ARTE
    unsigned int rtra() const;
#endif

    /** @name --- Operators --- */
    /// compare TrackIf pointers
    bool operator==(const TrackIf& rhs) const;
    /// compare Rtra value
    bool operator==( const RecoSegment& rhs ) const;
    /// compare Rtra value
    bool operator<( const RecoSegment& rhs ) const;

    /** @name --- TrackIf methods --- */
    /// $x$ of track
    float x()     const;
    /// $y$ of track
    float y()     const;
    /// linear extrapolation of $x$ to z
    float x(float z) const;
    /// linear extrapolation of $y$ to z
    float y(float z) const;
    /// $z$ of track
    float z()     const;
    /// in case of propagation: $z$ before propagation
    float zorig() const;
    /// $dx/dz$ slope of track ($t_x$)
    float tx()    const;
    /// $dy/dz$ slope of track ($t_y$)
    float ty()    const;
    /// momentum of track ($p$)
    float p()     const;
    /// transverse Track momentum $p_t = p\cdot\sqrt{(t_x^2+t_y^2)/(t_x^2+t_y^2+1)}$
    float pt()    const;
    /// momentum along $z$: $p / \sqrt{t_x^2+t_y^2+1}$ 
    float pz()    const;
    /// $\chi^2$ of track fit
    float chi2()  const;
    /// degrees of freedom
    unsigned short int ndf() const;
    /// azimuthal angle $\phi$ [deg]
    float phi() const;
    /// polar angle $\theta = \cos^{-1}(e_z)$ [deg]
    float theta() const;
    /// rapidity $\eta = -\log\tan(\theta/2.)$ rapidity
    float eta() const;
    /// particle charge: -1 for neg. +1 for pos. \& neutrals
    int   charge() const;

    /// $E = \sqrt{m^2 + p^2}$
    virtual float energy(double mass = 0.) const;
    /** $x_f = (p_z - E(m)\cdot\sqrt{1 - \frac{2m_p}{920 GeV/c + m_p}})/m_p$,
	$m_p$ = Proton mass, $m$ = Particle mass */
    float xf(double mass = 0.) const;
    /// Rapidity $y = \frac{1}{2}\ln(\frac{E+p_z}{E-p_z})$
    float rap(double mass = 0.) const;

    /** @name --- Access methods ---*/
    /// upper tail prob. of track $\chi^2$
    float trprob() const;
    /// set $\chi^2$
    void set_chi2(const float chi2);
    /// set degrees of freedom
    void set_nf(const UShort_t nf);

    /// inverse momentum of track
    float pinv() const;
    /// $t_x \cdot p_z$ 
    float px()   const;
    /// $t_y \cdot p_z$ 
    float py()   const;
    /// Component id
    RsegC cmp()  const;
    /// is RICH component id set?
    int isRICH() const;
    /// is OTR component id set?
    int isOTR()  const;
    /// is VDS component id set?
    int isVDS()  const;
    /// is ECAL component id set?
    int isECAL() const;

    /// set $x$ parameter
    void set_x( const float x);
    /// set $y$ parameter
    void set_y( const float y);
    /// set $z$ parameter
    void set_z( const float z);
    /// set $t_x$ parameter
    void set_tx(const float tx);
    /// set $t_y$ parameter
    void set_ty(const float ty);
    /// set $p$ and $p_{inv}$ parameter
    void set_p( const float p);

    /// get $\sigma_x^2|_{z+dz}$
    float cov_x(double dz=0.)    const;
    /// get $\sigma_x \times \sigma_y|_{z+dz}$
    float cov_x_y(double dz=0.)  const;
    /// get $\sigma_y^2|_{z+dz}$
    float cov_y(double dz=0.)    const;
    /// get $\sigma_x \times \sigma_{t_x}|_{z+dz}$
    float cov_x_tx(double dz=0.) const;
    /// get $\sigma_y \times \sigma_{t_x}|_{z+dz}$
    float cov_y_tx(double dz=0.) const;
    /// get $\sigma_{t_x}^2$
    float cov_tx()               const;
    /// get $\sigma_x \times \sigma_{t_x}|_{z+dz}$
    float cov_x_ty(double dz=0.) const;
    /// get $\sigma_y \times \sigma_{t_y}|_{z+dz}$
    float cov_y_ty(double dz=0.) const;
    /// get $\sigma_{t_x} \times \sigma_{t_y}$
    float cov_tx_ty()            const;
    /// get $\sigma_{t_y}^2$
    float cov_ty()               const;
    /// get $\sigma_x \times \sigma_p|_{z+dz}$
    float cov_x_p(double dz=0.) const;
    /// get $\sigma_y \times \sigma_p|_{z+dz}$
    float cov_y_p(double dz=0.) const;
    /// get $\sigma_{t_x} \times \sigma_p$
    float cov_tx_p()  const;
    /// get $\sigma_{t_y} \times \sigma_p$
    float cov_ty_p()  const;
    /// get $\sigma_p^2$
    float cov_p()     const;

    /// set $\sigma_x^2$
    void set_cov_x      (const float c);
    /// set $\sigma_x \times \sigma_y$
    void set_cov_x_y    (const float c);
    /// set $\sigma_y^2$
    void set_cov_y      (const float c);
    /// set $\sigma_x \times \sigma_{t_x}$
    void set_cov_x_tx   (const float c);
    /// set $\sigma_y \times \sigma_{t_x}$
    void set_cov_y_tx   (const float c);
    /// set $\sigma_{t_x}^2$
    void set_cov_tx     (const float c);
    /// set $\sigma_x \times \sigma_{t_x}$
    void set_cov_x_ty   (const float c);
    /// set $\sigma_y \times \sigma_{t_y}$
    void set_cov_y_ty   (const float c);
    /// set $\sigma_{t_x} \times \sigma_{t_y}$
    void set_cov_tx_ty  (const float c);
    /// set $\sigma_{t_y}^2$
    void set_cov_ty     (const float c);
    /// set $\sigma_x \times \sigma_p$
    void set_cov_x_p (const float c);
    /// set $\sigma_y \times \sigma_p$
    void set_cov_y_p (const float c);
    /// set $\sigma_{t_x} \times \sigma_p$
    void set_cov_tx_p(const float c);
    /// set $\sigma_{t_y} \times \sigma_p$
    void set_cov_ty_p(const float c);
    /// set $\sigma_p^2$
    void set_cov_p   (const float c);

    /** @name --- Expert functions ---*/
    /// collect pointers
    void collect(vector<TrackIf*>& c) const;
    /// propagate track to $z$
    void propagate(const double zz);
    /// $\sigma(Q/p) \to \sigma(p)$
    void sPinvToP();
    /// $\sigma(p) \to \sigma(Q/p)$
    void sPToPinv();

#ifndef __CINT__
    /// $\vec{x} = (x,y,z)$
    SVector<double,3> xvec() const;
    /// $\vec{t} = (t_x,t_y,1)$
    SVector<double,3> tvec() const;
    /// $\vec{e} = (e_x,e_y,ez)$ unit vector along $\vec{p}$
    SVector<double,3> evec() const;
    /// $\vec{p} = (p_x,p_y,p_z)$
    SVector<double,3> pvec() const;
    /// fill COV\_ matrix and compute inverse cov. matrix
    bool FillCINV();
    /// return covariance matrix $\textbf{\textrm{Cov}}$
    const SMatrix<double,5>& COV() const;
    /// return inverse covariance matrix $\textbf{\textrm{Cov}}^{-1}$
    const SMatrix<double,5>& CINV() const;
#endif

    /// is track valid?
    virtual bool isValid() const;

    /** Prepare reconstructed track for vertexing with SVertex package:
	propagate track parameters and errors to $z=0$, fill inverse covariance matrix.
    */
    bool PrepareForVertexing();
    /// mark track as valid
    virtual void valid();
    /// mark track as invalid
    virtual void invalid();
    /// used by operator<<()
    std::ostream& print( std::ostream& ) const;
    /// print covariance matrix
    std::ostream& print_cov( std::ostream& ) const;

    /** @name --- GRelation relations --- */
#ifndef __CINT__
    /// RecoTrack $\rightarrow$ PatternMatch
    GRelationList<SPatternMatch> pmatch; //! not persistent
    ///
    const GRelationList<SPatternMatch>& Pmatch() const;
    ///
    GRelationList<SPatternMatch>& Pmatch();
    ///
    GRelationList<SPatternMatch>& container(const SPatternMatch& p);
#endif

  private:
    Float_t X;              // x of track
    Float_t Y;              // y of track
    Float_t Z;              // z of track
    Float_t Zorig;          //! in case of propagation: Z before propagation
    Float_t Tx;             // dx/dz slope
    Float_t Ty;             // dy/dz slope
    Float_t P;              // Momentum of particle
    Float_t Chi2;           // chi square
    Float_t Cov[15];        // lower triangle of cov-matrix
    Short_t Charge;         // charge: -1 for neg. +1 for pos. & neutrals
    RsegC   Cmp;            // Component id Flag
    
    UShort_t Rtra;          // RTRA pointer
    UShort_t Nf;            // degrees of freedom
    Bool_t   valid_;        //! validity (not persistent)

#ifndef __CINT__
    SMatrix<double,5> COV_;   //! cov. matrix (not persistent)
    SMatrix<double,5> CINV_;  //! inverse cov. matrix (not persistent)
#endif

#ifdef USE_ROOT
  public:
    /// only available in ROOT
    void Print(Option_t *option="");
    
    ClassDef(RecoSegment,3)
#endif
  }; // end of class RecoSegment


  //==============================================================================
  // operator <<
  //==============================================================================
  std::ostream& operator<<(std::ostream& os, const RecoSegment& t);

#ifndef __CINT__
#include "RecoSegment.icc"
#endif

#ifdef NAMESPACE
} // namespace CLUE
#endif
#endif
