#ifndef RECO_TRACK_HH
#define RECO_TRACK_HH
// *****************************************************************************
//
// source:
//
// type:      source code
//
// created:   27. Jan 1997
//
// author:    Thorsten Glebe
//            HERA-B Collaboration
//            Max-Planck-Institut fuer Kernphysik
//            Saupfercheckweg 1
//            69117 Heidelberg
//            Germany
//            E-mail: T.Glebe@mpi-hd.mpg.de
//
// Description: Class for handling reconstructed Tracks
//
// changes:
// 27 Jan 1997 creation by I. Siccama, C. Borgmeier
// 31 Okt 2000 (TG) removed contstructor using FORTRAN-Vt interface
// 20 Nov 2000 (TG) added namespace CLUE
// 06 Dec 2000 (TG) replaced ARTE relations by CLUE relations
// 14 Dec 2000 (TG) RelStreamer method added
// 21 Dec 2000 (TG) declared ClueRelationList container non-persistent (//!),
//                  added non-const RelationList access functions
// 22 Dec 2000 (TG) changed type of iTopp to UShort_t
// 12 Jan 2001 (TG) changed print() return type, added energy()
// 12 Jan 2001 (TG) removed GetLorentzVector, decoupled class from MC_Track
// 22 Jan 2001 (TG) added member function xf(), added ClMember pz, energy, xf
// 08 Feb 2001 (TG) added ClMember px,py,xfabs
// 04 Apr 2001 (TG) Cer and cer() member added
// 18 Apr 2001 (TG) mommask, Rls, Rlt, Lse, Lsmu, Lspi, Lsk, Lsp added,
//                  changed type of Fit, Hsi, Hin, Hou, Hri, Hca, Htr, Hmu to
//                  UChar_t, moved inline functions to RecoTrack.icc
// 20 Apr 2001 (TG) removed lock(), unlock(), locked(), getRecoTrack()
// 20 Apr 2001 (TG) added Clmember rls, rlt, lse, lsmu, lspi, lsk, lsp, mom
// 24 Apr 2001 (TG) added Relation to PatternMatchS
// 28 Jun 2001 (TG) changed PatternMatchS to SPatternMatch, commented ClMembers
// 29 Jun 2001 (TG) moved energy(mass), xf() to RecoSegment, removed xfabs,
//                  moved SPatternMatch Relation to RecoSegment class
// 11 Jul 2001 (TG) added EcalHit relation
// 25 Jul 2001 (TG) changed mtrack from pointer to object
// 05 Sep 2001 (TG) RelStreamer: StreamVersion added
//
// *****************************************************************************

#include <iosfwd>
#include "clue/RecoSegment.hh"
#include "grelation/ClueRelation.hh"

#ifdef NAMESPACE
namespace CLUE {
#endif
  class abc; // to fake doc++

  /** reconstructed track class
      @author C. Borgmeier, T. Glebe
  */
  //==============================================================================
  // RecoSegment
  //==============================================================================
  class RecoTrack: public RecoSegment {
  public:
    /** @name --- Constructors ---*/
    //RecoTrack( const RecoSegment& rt );
    ///
    RecoTrack();
    ///
    RecoTrack( const RecoTrack& rt );
    ///
    ~RecoTrack();
    
#ifndef NO_ARTE
    /// not available in ROOT
    RecoTrack( const ArtePointer<RTRA>& rtra, double mass = 0 );
    
    /// not available in ROOT
    RecoTrack( const ArtePointer<RSEG>& rseg );
#endif
    ///
    const RecoTrack& operator=( const RecoTrack& rhs );

    /** @name --- Access methods ---*/
    /// Rich: particle identification flag
    int  flr() const;
    /// TRD: particle identification flag
    int  flt() const;
    /// ECAL: particle identification flag
    int  fle() const;
    /// MUON: particle identification flag
    int  flm() const;
    
    /// Rich: Riter particle identification flag
    int rls() const;
    /// Rich: Rise particle identification flag 
    int rlt() const;

    /// Rich: Rise likelihood of electron hypothesis 
    float lre()  const;
    /// Rich: Rise likelihood of muon hypothesis 
    float lrmu() const;
    /// Rich: Rise likelihood of pion hypothesis 
    float lrpi() const;
    /// Rich: Rise likelihood of kaon hypothesis 
    float lrk()  const;
    /// Rich: Rise likelihood of proton hypothesis 
    float lrp()  const;
    
    /// Rich: Riter likelihood of electron hypothesis 
    float lse()  const;
    /// Rich: Riter likelihood of muon hypothesis 
    float lsmu() const;
    /// Rich: Riter likelihood of pion hypothesis 
    float lspi() const;
    /// Rich: Riter likelihood of kaon hypothesis 
    float lsk()  const;
    /// Rich: Riter likelihood of proton hypothesis 
    float lsp()  const;
    
    /// TRD: likelihood of electron hypothesis 
    float lte()  const;
    /// TRD: likelihood of hadron hypothesis
    float lth()  const;

    /// ECAL: likelihood of electron hypothesis 
    float lee()  const;
    /// ECAL: likelihood of MIP hypothesis 
    float lemi() const;
    /// ECAL: likelihood of hadron hypothesis 
    float leh()  const;

    /// MUON: likelihood of muon hypothesis 
    float lmmu() const;
    /// MUON: likelihood of hadron hypothesis 
    float lmh()  const;

    /// global likelihood of muon hypothesis
    float lmu()  const;
    /// global likelihood of pion hypothesis
    float lpi()  const;
    /// global likelihood of kaon hypothesis
    float lk()   const;
    /// global likelihood of proton hypothesis
    float lp()   const;
    
    /// number of hits in the SVD 
    int hsi() const;
    /// number of hits in the inner tracker 
    int hin() const;
    /// number of hits in the outer tracker
    int hou() const;
    /// number of hits in the RICH 
    int hri() const;
    /// number of hits in the calorimeter 
    int hca() const;
    /// number of hits in the TRD 
    int htr() const;
    /// number of hits in the muon chambers 
    int hmu() const;
    
    /// fit flag
    int fit() const;
    
    /// particle mass $m$
    float mass() const;
    /// set particle mass $m$
    void  set_mass(const float newMass);
    /// $E = \sqrt{m^2 + p^2}$
    float energy(double mass = 0.) const;
    /// possible Bremsstrahlung correction
    float ebrems() const;
    /// Cerenkov angle of associated RICH ring
    float cer() const;

    /**
       Momentum determination flag:\\
	1: estimated by REFIT on all hits assigned,\\
	2: estimated by REFIT with rejection of hits in magnet,\\
	3: estimated in VDS-ECAL match,\\
	4: obtained during magnet tracking,\\
	5: target constrain of OTR-PC segment,\\
	6: PC-kick in the magnet is used for estimation,\\
	7: estimated in VDS-RICH match,\\
	8: estimated in VDS-PC match,\\
	9: assigned by CATS for VDS segment
     */
    int mommask() const;
    /// link to Topp table
    const ArtePointer<Topp> topp() const;
    /// set link to Topp table
    void set_topp(const ArtePointer<Topp>& topp);

    /** @name --- Expert functions ---*/
    /// used by operator<<()
    std::ostream& print( std::ostream& os ) const;
    
    /// print related MC tracks
    void printMC() const;

    /** @name --- CLUE relations --- */
    /// RecoTrack $\leftrightarrow$ MC\_Track relation
    ClueRelationList<MC_Track> mtrack;  //! not persistent
    ///
    const ClueRelationList<MC_Track>& mTrack() const;
    ///
    ClueRelationList<MC_Track>& mTrack();
    ///
    ClueRelationList<MC_Track>& container(const ArtePointer<MC_Track>& p);

    /// RecoTrack $\leftrightarrow$ TrackClones
    ClueRelationList<TrackClones> clones; //! not persistent
    ///
    const ClueRelationList<TrackClones>& Clones() const;
    ///
    ClueRelationList<TrackClones>& Clones();
    ///
    ClueRelationList<TrackClones>& container(const ArtePointer<TrackClones>& p);

    /// RecoTrack $\leftrightarrow$ RecoVertex
    ClueRelationList<RecoVertex> rvertex; //! not persistent
    ///
    const ClueRelationList<RecoVertex>& rVertex() const;
    ///
    ClueRelationList<RecoVertex>& rVertex();
    ///
    ClueRelationList<RecoVertex>& 
    container(const ArtePointer<RecoVertex>& p);

    /// RecoTrack $\leftrightarrow$ EcalHit
    ClueRelationList<EcalHit> ehit; //! not persistent
    ///
    const ClueRelationList<EcalHit>& Ehit() const;
    ///
    ClueRelationList<EcalHit>& Ehit();
    ///
    ClueRelationList<EcalHit>& 
    container(const ArtePointer<EcalHit>& p);

#ifdef USE_ROOT
    /// only availabe in ROOT: call Streamer of relation container
    void RelStreamer(TBuffer& R__b, const Version_t& StreamVersion);
#endif

  private:
    Float_t  Mass;      // particle mass
    Float_t  Ebrems;    // Bremsstrahlung - correction
    Float_t  Cer;       // Cerenkov angle of associated RICH ring
    UShort_t iTopp;     // link to topp table
    
    Bool_t  Flr;        // Rich: particle identification flag
    Bool_t  Flt;        // TRD: particle identification flag
    Bool_t  Fle;        // ECAL: particle identification flag
    Bool_t  Flm;        // MUON: particle identification flag
    
    Float_t Lre;        // Rich: Rise likelihood of electron hypothesis
    Float_t Lrmu;       // Rich: Rise likelihood of muon hypothesis
    Float_t Lrpi;       // Rich: Rise likelihood of pion hypothesis
    Float_t Lrk;        // Rich: Rise likelihood of kaon hypothesis
    Float_t Lrp;        // Rich: Rise likelihood of proton hypothesis
    
    Float_t Lte;        // TRD: likelihood of electron hypothesis
    Float_t Lth;        // TRD: likelihood of hadron hypothesis
    
    Float_t Lee;        // ECAL: likelihood of electron hypothesis
    Float_t Lemi;       // ECAL: likelihood of MIP hypothesis
    Float_t Leh;        // ECAL: likelihood of hadron hypothesis
    
    Float_t Lmmu;       // MUON: likelihood of muon hypothesis
    Float_t Lmh;        // MUON: likelihood of hadron hypothesis
    
    Float_t Lmu;        // global likelihood of muon hypothesis
    Float_t Lpi;        // global likelihood of pion hypothesis
    Float_t Lk;         // global likelihood of kaon hypothesis
    Float_t Lp;         // global likelihood of proton hypothesis

    UChar_t Fit;        // ARTE fit flag
    
    UChar_t Hsi;        // number of hits in the SVD
    UChar_t Hin;        // number of hits in the inner tracker
    UChar_t Hou;        // number of hits in the outer tracker
    UChar_t Hri;        // number of hits in the RICH
    UChar_t Hca;        // number of hits in the calorimeter
    UChar_t Htr;        // number of hits in the TRD
    UChar_t Hmu;        // number of hits in the muon chambers
    
    UChar_t Mommask;    // momentum determination flag

    UChar_t Rls;        // Riter: particle identification flag
    UChar_t Rlt;        // Rise: particle identification flag 

    Float_t Lse;        // Rich: Riter likelihood of electron hypothesis
    Float_t Lsmu;       // Rich: Riter likelihood of muon hypothesis
    Float_t Lspi;       // Rich: Riter likelihood of pion hypothesis
    Float_t Lsk;        // Rich: Riter likelihood of kaon hypothesis
    Float_t Lsp;        // Rich: Riter likelihood of proton hypothesis
    
#ifdef USE_ROOT
  public:
    /// only available in ROOT
    void Print( Option_t* o="" );  
    
    ClassDef(RecoTrack,3)
#endif
  }; // end of class RecoTrack

  
  //==============================================================================
  // operator <<
  //==============================================================================
  std::ostream& operator<< ( std::ostream& os, const RecoTrack& t );

#ifndef __CINT__
#include "RecoTrack.icc"
#endif // __CINT__

#ifdef NAMESPACE
} // namespace CLUE
#endif
#endif
