#ifndef RTEVENT_HH
#define RTEVENT_HH
// **********************************************************************
//
// source:
//
// type:      source code
//
// created:   20. Jul 1997
//
// author:    Thorsten Glebe
//            HERA-B Collaboration
//            Max-Planck-Institut fuer Kernphysik
//            Saupfercheckweg 1
//            69117 Heidelberg
//            Germany
//            E-mail: T.Glebe@mpi-hd.mpg.de
//
// Description: Class to contain the ROOT Event
//
// changes:
// 20 Jul 1997 creation by C. Borgmeier
// 30 Okt 2000 (TG) removed fillWireBuffer(), fillTrackBuffer(),
//             findPrimVertices(), findSecVertices()
// 20 Nov 2000 (TG) added namespace CLUE
// 22 Dec 2000 (TG) code cleanup
// 02 Jan 2001 (TG) changed type of Bx to UShort_t, code cleanup
// 03 Jan 2001 (TG) removed init(), moved storeRSEG_Flag to cluearte/cluearte.C
// 03 Jan 2001 (TG) added FillTopp(), FillHeader()
// 04 Jan 2001 (TG) added Delete()
// 12 Jan 2001 (TG) removed removeMC_Track()
// 16 Mar 2001 (TG) added rad0 and hits in detector components
// 22 Mar 2001 (TG) added user, StreamUser()
// 30 Mar 2001 (TG) added Tcode, Ir, Bg, tcode(), ir(), bg()
// 31 Aug 2001 (TG) added prepareRecoTracksforVertexing()
//
// **********************************************************************

#include <iosfwd>
#include <vector>
#include "arte/artefwd.hh"
#include "gate/gatefwd.hh"
#include "clue/cluefwd.hh"

#undef getLink

#ifdef NAMESPACE
namespace CLUE {
#else
#endif
  class abc; // to fake doc++

  /** Holds information of ARTE in C++ objects. The Track lists can
      also be accessed via the ARTE table interface.
      @memo event class
      @author C. Borgmeier, T. Glebe
  */
  //==============================================================================
  // RtEvent
  //==============================================================================
  class RtEvent : public TObject {
  public:
    /** @name --- Constructors --- */
    ///
    RtEvent();
    ///
    virtual ~RtEvent();

    // static members are not streamed by ROOT
    //    static bool storeRSEG_Flag;

    /** @name --- Access methods --- */
    /// seconds since 1.1.70 0:00 somewhere in the U.S.
    long int unixtime() const;
    /// event number
    int evt() const;
    /// run number
    int run() const;
    /** @memo experiment number
	experiment number
	exp > 0: real data
	exp < 0: Monte Carlo; specifies the lab:
	-1: DESY HH
	-2: DESY Zeuthen
	...: not yet specified */
    int exp() const;
    /// version of the mag field MC and reco
    float vfld() const;
    /// event classification flag from/to DST
    int flagdst() const;
    /// major key for constants on data base
    int dbmajor() const;
    /// minor key for constants on data base
    int dbminor() const;
    /// event classification flag from/to EDIR if no EDIR: same as flagdst
    int flag()    const;
    /// bx number from DIHD
    unsigned int bx() const;
    /** Trigger code:
	0: FLT, 2: VME standalone, 4: LEMO1, 5: LEMO2, 6: LEMO3, 7: LEMO4,
	8: FLT + random, 9: random, 10: VME + random, 12: LEMO1 + random,
	13: LEMO2 + random, 14: LEMO3 + random, 15: LEMO4 + random
    */
    int tcode() const;
    /// Interaction rate
    float ir()  const;
    /// Background rate
    float bg()  const;
    /// Cherenkov radius for $\beta = 1$
    float rad0() const;
    /// Hits in VDS
    int hvds()  const;
    /// Hits in wire chambers
    int hwch()  const;
    /// Hits in MSG microstrip chambers
    int hmsg()  const;
    /// Hits in PTC chambers (pi-pi trigger)
    int hptc()  const;
    /// Hits in Cerenkov counter
    int hrich() const;
    /// Hits in forward TRD
    int htrd()  const;
    /// Hits in electromagnetic calorimeter
    int hcal()  const;
    /// Hits in muon outer chambers
    int hmuo()  const;
    /// Hits in muon central chambers
    int hmux()  const;

    /// get Topp table pointer
    GClonesArray<Topp*>*           topp()       const;
    /// get RecoTrack table pointer
    GClonesArray<RecoTrack*>*      tracks()     const;
    ///
    GClonesArray<VDS_Track*>*      vds()        const;
    ///						     
    GClonesArray<OTR_SpacePoint*>* otr()        const;
    ///						     
    GClonesArray<RICH_Track*>*     rich()       const;
    ///						     
    GClonesArray<EcalHit*>*        ecal()       const;
    ///						     
    GClonesArray<MuonTrack*>*      muon()       const;
    ///						     
    GClonesArray<RecoVertex*>*     vertices()   const;
    ///						     
    GClonesArray<Wire*>*           wires()      const;
    ///						     
    GClonesArray<TrackClones*>*    clones()     const;
    ///						     
    GClonesArray<MC_Track*>*       mcTracks()   const;
    ///						     
    GClonesArray<MC_Vertex*>*      mcVertices() const;

    static GClonesArray<Topp*>*             toppLink; // always a constant table
#if defined(USE_GATE)
    static GClonesArray<Wire*>*             wireLink;
    static GClonesArray<RecoTrack*>*        recoTrackLink;
    static GClonesArray<VDS_Track*>*        vdsTrackLink;
    static GClonesArray<OTR_SpacePoint*>*   otrSpacePointLink;
    static GClonesArray<RICH_Track*>*       richTrackLink;
    static GClonesArray<EcalHit*>*          EcalHitLink;
    static GClonesArray<MuonTrack*>*        muonTrackLink;
    static GClonesArray<RecoVertex*>*       recoVertexLink;
    static GClonesArray<TrackClones*>*      trackClonesLink;
    static GClonesArray<MC_Vertex*>*        mcVertexLink;
    static GClonesArray<MC_Track*>*         mcTrackLink;
#endif

    /** @name --- Event algorithms --- */
    /** Prepare reconstructed tracks for vertexing with SVertex package:
	transform $\sigma^2(1/p)\rightarrow\sigma^2(p)$, propagate
	RecoTracks to $z=0$, fill inverse covariance matrix.
    */
    bool prepareRecoTracksforVertexing();

    /** @name --- Expert functions --- */
    /// used in operator<<()
    std::ostream& print( std::ostream& os ) const;
#ifdef USE_ARTE
    /// fill Topp table from ARTE TOPP
    bool FillTopp();
    /// fill event header from ARTE EVHD, DIHD
    bool FillHeader();
#endif
    /// initialize objects contained in GClonesArrays
    void unlock();
    
#ifdef USE_ROOT
    /** @name --- ROOT only functions --- */
    /// 
    void Clear( Option_t * = "" );
    /// 
    void Delete( Option_t * = "" );
    /// 
    void Print( Option_t *options="" );
#else
    /// replacement of ROOT Clear()
    void Clear();
    /// replacement of ROOT Delete()
    void Delete();
#endif
  

#ifdef USE_ROOT
    /// user space for ROOT file
    vector<float> user;  //! not persistent, stream on our own
    /// only availabe in ROOT
    void StreamUser(TBuffer &R__b);
#endif

  private:
    /// do not allow copying
    RtEvent( const RtEvent& rhs );
    /// do not allow copying
    RtEvent& operator=( const RtEvent& rhs );

    UInt_t   Unixtime;    // Long_t not supported in split mode
    Int_t    Evt;
    Int_t    Run;
    Int_t    Exp;
    Float_t  Vfld;
    Int_t    Flagdst;
    Int_t    Dbmajor;
    Int_t    Dbminor;
    Int_t    Flag;
    UShort_t Bx;          // Bx in range [0,220]
    Short_t  Tcode;       // Trigger code
    Float_t  Ir;          // Interaction rate
    Float_t  Bg;          // Background rate
    Float_t  Rad0;
    UShort_t Hvds;
    UShort_t Hwch;
    UShort_t Hmsg;
    UShort_t Hptc;
    UShort_t Hric;
    UShort_t Htrd;
    UShort_t Hcal;
    UShort_t Hmuo;
    UShort_t Hmux;

    GClonesArray<RecoTrack*>*      Tracks;
    GClonesArray<VDS_Track*>*      Vds;
    GClonesArray<OTR_SpacePoint*>* Otr;
    GClonesArray<RICH_Track*>*     Rich;
    GClonesArray<EcalHit*>*        Ecal;
    GClonesArray<MuonTrack*>*      Muon;
    GClonesArray<RecoVertex*>*     Vertices;
    GClonesArray<Wire*>*           Wires;
    GClonesArray<TrackClones*>*    Clones;
    GClonesArray<MC_Track*>*       McTracks;
    GClonesArray<MC_Vertex*>*      McVertices;

#if defined(USE_GATE) && defined(USE_ROOT)
    StreamObject* streamer;  // must be last member!
#endif
    
#ifdef USE_ROOT
    ClassDef(RtEvent,1)
#endif
  }; // end of class RtEvent


  //==============================================================================
  // operator <<
  //==============================================================================
  std::ostream& operator<< (std::ostream& os, const RtEvent& t);

#if defined(USE_GATE)
  //==============================================================================
  // Template specialization 
  //==============================================================================
  template<> inline void* ArteTable<RecoTrack>::getLink() {
    return CLUE::RtEvent::recoTrackLink;
  }
  
  template<> inline void* ArteTable<VDS_Track>::getLink() {
    return CLUE::RtEvent::vdsTrackLink;
  }
  
  template<> inline void* ArteTable<OTR_SpacePoint>::getLink() {
    return CLUE::RtEvent::otrSpacePointLink;
  }
  
  template<> inline void* ArteTable<RICH_Track>::getLink() {
    return CLUE::RtEvent::richTrackLink;
  }
  
  template<> inline void* ArteTable<EcalHit>::getLink() {
    return CLUE::RtEvent::EcalHitLink;
  }
  
  template<> inline void* ArteTable<MuonTrack>::getLink() {
    return CLUE::RtEvent::muonTrackLink;
  }
  
  template<> inline void* ArteTable<RecoVertex>::getLink() {
    return CLUE::RtEvent::recoVertexLink;
  }
  
  template<> inline void* ArteTable<TrackClones>::getLink() {
    return CLUE::RtEvent::trackClonesLink;
  }
  
  template<> inline void* ArteTable<MC_Vertex>::getLink() {
    return CLUE::RtEvent::mcVertexLink;
  }

  template<> inline void* ArteTable<MC_Track>::getLink() {
    return CLUE::RtEvent::mcTrackLink;
  }

  template<> inline void* ArteTable<Topp>::getLink() {
    return CLUE::RtEvent::toppLink;
  }

  template<> inline void* ArteTable<Wire>::getLink() {
    return CLUE::RtEvent::wireLink;
  }
#endif // USE_GATE

#ifdef NAMESPACE
} // namespace CLUE
#endif
#ifndef __CINT__
#include "RtEvent.icc"
#endif
#endif
