#ifndef COMBINER_HH
#define COMBINER_HH
// *****************************************************************************
//
// source:
//
// type:      source code
//
// created:   14. Jul 1997
//
// author:    Thorsten Glebe
//            HERA-B Collaboration
//            Max-Planck-Institut fuer Kernphysik
//            Saupfercheckweg 1
//            69117 Heidelberg
//            Germany
//            E-mail: T.Glebe@mpi-hd.mpg.de
//
// Description: Combines particles
//
// changes:
// 14 Jul 1997 creation and implementation by C. Borgmeier
// 03 Nov 2000 (TG) removed #include "TrackBase.hh", added #include "TObject.hh"
// 16 Nov 2000 (TG) added cluefwd.hh, gatefwd.hh
// 20 Nov 2000 (TG) added namespace CLUE
//
// *****************************************************************************

#include <iosfwd>
#include <vector>
#include "arte/TObject.hh"
#include "clue/cluefwd.hh"
#include "gate/gatefwd.hh"

#ifndef USE_ROOT
#define TObjArray std::vector<TrackBase*>
#endif


#ifdef NAMESPACE
namespace CLUE {
#endif

  class vecplusit;
  typedef std::vector<vecplusit>           vec_v;
  typedef std::vector<vecplusit>::iterator vec_vit;

  typedef std::vector<std::TObjArray*>     TObjArray_v;

  //=============================================================================
  // vectplusit
  //==============================================================================
  class vecplusit {
  public:
    vecplusit( GObjArray<TrackBase*>* v )
      : list(v), iter( list->begin() ), startList( 0 )
      { }
    
    vecplusit() { }
    
    GObjArray<TrackBase*>* list;
    GObjArray<TrackBase*>::iterator iter;
    
    vec_vit startList;
  };

  
  //=============================================================================
  // operator <<
  //=============================================================================
  inline std::ostream& operator<<(std::ostream& os, const vecplusit& v) {
    return os << "|i=" << v.iter << " start=" << v.startList
	      << "list=" << v.list <<"|";
  }



  /** Particle combination. Taken from CLHEP Chooser
      @memo Particle combination
      @author C. Borgmeier, T. Glebe
  */
  //==============================================================================
  // Combiner
  //==============================================================================
  class Combiner: public TObject {
  public:
    /** @name --- Constructors --- */
    /// Constructor taking up to six pointers to lists as argument
    Combiner( TObjArray* l1,
	      TObjArray* l2 = 0,
	      TObjArray* l3 = 0,
	      TObjArray* l4 = 0,
	      TObjArray* l5 = 0,
	      TObjArray* l6 = 0 );
    ///
    Combiner( TObjArray&, int mask = 0 );
    ///
    Combiner( TObjArray&, TObjArray&, int mask = 0 );
    ///
    Combiner( TObjArray&, TObjArray&, TObjArray&, int mask = 0 );
    ///
    Combiner( TObjArray&, TObjArray&, TObjArray&, TObjArray&, int mask = 0 );
    ///
    Combiner( TObjArray&, TObjArray&, TObjArray&,
	      TObjArray&, TObjArray&, int mask = 0 );
    /// default constructor
    Combiner();
    ///
    Combiner( const Combiner& rhs );
    
#ifndef __CINT__
    /// not available in CINT
    Combiner( std::vector<GObjArray<PatternMatch*>*>& l, int mask = 0 );
#endif
    ///
    virtual ~Combiner();
    ///
    Combiner& operator=( const Combiner& rhs );
    
    /** @name --- Operators --- */
    /// postfix increment
    Combiner& operator++();
    
    /// prefix increment
    Combiner operator++(int);

    /** Returns a pointer to a list with the next combination of objects or
	zero if all combinations are done.
    */
    inline const GObjArray<TrackBase*>& operator*() { return currList; };
    /** Returns a pointer to a list with the next combination of objects or
	zero if all combinations are done. */
    inline const GObjArray<TrackBase*>& current()   { return currList; };
    
    /** @name --- Access methods --- */
    /// returns true, if all combinations done
    inline int end() const { return done; }
    ///
    void orderMatters();
    ///
    float mass() const;

    /// used by operator<<()
    void print( std::ostream& os ) const;
    
  protected:

    ///
    void init( TObjArray_v &ll );
    ///
    void nextComb();
    ///
    bool nextComb( vec_vit begin, vec_vit end);
    ///
    bool incrIndex(vecplusit& i);

  private:  
    vec_v* listData;
    int done;
    GObjArray<TrackBase*> currList; //!
    
    int allowCollisionPattern;

#ifdef USE_ROOT
  public:
    /// only available in ROOT
    void Print( Option_t* ="" );

    ClassDef(Combiner,1)
#endif
  }; // end of class Combiner

  
  //==============================================================================
  // operator <<
  //==============================================================================
  inline std::ostream& operator<<( std::ostream& os, const Combiner& l ) {
    l.print(os);
    return os;
  }

#ifdef NAMESPACE
}// end of namespace CLUE
#endif
#endif
