// **********************************************************************
//
// source:
//
// type:      source code
//
// created:   22. Jul 1997
//
// author:    Thorsten Glebe
//            HERA-B Collaboration
//            Max-Planck-Institut fuer Kernphysik
//            Saupfercheckweg 1
//            69117 Heidelberg
//            Germany
//            E-mail: T.Glebe@mpi-hd.mpg.de
//
// Description: construct a tree out of Pdt
//
// changes:
// 22 Jul 1997 created by C. Borgmeier
// 30 Okt 2000 (TG) removed call to RecoVertex::ivFree()
// 20 Nov 2000 (TG) added namespace CLUE
// 27 Dec 2000 (TG) end() optimization
// 02 Jan 2001 (TG) end() bug fixes
// 09 Jan 2001 (TG) handle mcMatches as pointer
// 12 Jan 2001 (TG) removed obsolete include files
//
// **********************************************************************
#include <iostream>
#include <algorithm>

#include "clue/cluearte.hh"
// #include "geometry/Point3D.hh"  // needed by RecoSegment.hh
#include "clue/Topp.hh"
// #include "geometry/ClLorentzVector.hh" // needed by MC_Track.hh
#include "gate/ObjArray.hh"     // needed for Pattern.hh
#include "gate/ClonesArray.hh"  // needed for Pattern.hh
#include "clue/SmartPointer.hh" // needed for Pattern.hh
#include "clue/Property.hh"
#include "clue/Pattern.hh"
#include "clue/PatternMatch.hh"
// #include "clue/RecoVertex.hh"
// #include "clue/MC_Vertex.hh"    // needed by MC_Track
#include "clue/MC_Track.hh"
#include "clue/PdtEntry.hh"
// #include "clue/Combiner.hh"

#ifdef NAMESPACE
namespace CLUE {
#endif

  // NULL pointer
  static const ArtePointer<MC_Track> MC_Track0;

  //==============================================================================
  // delete_it
  //==============================================================================
  static void delete_it( TObject* obj ) {
    delete obj;
  }

  //==============================================================================
  // Constructors
  //==============================================================================
  Pattern::Pattern( ClPdtEntry const& leaf,
		    const ClProperty<RecoTrack,int>* pat ) :
    connection( NO ),
    children( new Mset ),
    entry( &leaf ),
    pattern( pat!=0? pat->propertyClone(): 0 ),
    mcMatches( new GObjArray<PatternMatch*> ) {
    //std::cerr << "created: " << pattern << ',' << pat << endl;
  }


  Pattern::Pattern( ConnectionType c,
		    Mset::iterator first,
		    Mset::iterator last ) :
    connection( c ),
    children( new Mset ),
    entry( 0 ),
    pattern( 0 ),
    mcMatches( new GObjArray<PatternMatch*> ) {


    if( first!=last ) {
      Mset::iterator testIt( first );
      if( ++testIt==last ) {
	// if it is only one node, use assignment.
	Pattern const& f = **first;
	*this = f;
      } else {
	// else use connection.
	std::copy( first, last, inserter( *children, (*children).begin() ) );
      }
    }
  }


  Pattern::Pattern( ClPdtEntry const& decay, Pattern const& products) :
    connection( products.connection ),
    children( new Mset( *products.children ) ),
    entry( &decay ),
    pattern( 0 ),
    mcMatches( new GObjArray<PatternMatch*> )
    {}


  Pattern::Pattern( ConnectionType c,
		    Pattern const& e1,
		    Pattern const& e2 ) :
    connection( c ),
    children( new Mset ),
    entry( 0 ),
    pattern( 0 ),
    mcMatches( new GObjArray<PatternMatch*> )
    {

    // std::cerr << "Constructing " << c << ' ' << e1 << ',' << e2 << endl;
    (*children).insert( SmartPointer<Pattern>( new Pattern( e1 ) ) );
    (*children).insert( SmartPointer<Pattern>( new Pattern( e2 ) ) );
  }


  Pattern::Pattern( Pattern const& n ) :
    connection( n.connection ),
    children( new Mset( *n.children ) ),
    entry( n.entry ),
    pattern( n.pattern!=0? n.pattern->propertyClone(): 0 ),
    mcMatches( new GObjArray<PatternMatch*> )
    {
    // std::cerr << "Copying " << n << endl;
  }


  Pattern::~Pattern() {
    //std::cerr << "deleting Node " << this << endl;
    delete children;
    const GObjArray<PatternMatch*>::iterator end = mcMatches->end();
    for( GObjArray<PatternMatch*>::iterator i = mcMatches->begin();
	 i != end; ++i ) {
      //std::cerr << "removing pointer in " << *i << endl;
      if( (*i)->pattern!=this ) {
	std::cerr << "inconstistency in PdtNode " << this
		  << ',' << (*i)->pattern << endl; 
      }
      (*i)->pattern = 0;
    }
    if( pattern!=0 )   delete pattern;
    if( mcMatches!=0 ) delete mcMatches;
  }


  //==============================================================================
  // operator=
  //==============================================================================
  Pattern& Pattern::operator=( Pattern const& n ) {
    //std::cerr << "assigning " << n << endl;
    if( this == &n ) return *this;

    connection = n.connection;
    entry = n.entry;
#ifdef USE_ROOT
    mcMatches->Delete();  // This is not very good.
#endif
    (*children).erase( (*children).begin(), (*children).end() );
    std::copy( (*n.children).begin(), (*n.children).end(),
	       inserter( *children, (*children).begin() ) );
    if( pattern!=0 ) delete pattern;
    pattern = (n.pattern!=0? n.pattern->propertyClone(): 0);
    
    return *this;
  }


  //==============================================================================
  // operator &
  //==============================================================================
  Pattern Pattern::operator&( Pattern const& e ) const {
    //std::cerr << "anding to " << *this << " node " << e << endl; 
    if( connection==AND && entry==0 ) {
      Pattern n( *this );
      if( e.connection==AND ) {
	//std::cerr << "i.e. inserting in " << n << ": ";
	const Mset::const_iterator end = (*e.children).end();
	for( Mset::const_iterator i =(*e.children).begin();
	     i != end; ++i ) {
	  //std::cerr << " node " << **i << endl;
	  (*n.children).insert( *i );
	}
	//std::cerr << endl;
      } else {
	//std::cerr << "i.e. inserting in " << n << " node "<< e << endl;
	(*n.children).insert( e );
      }
      return n;
    }
    else if( e.connection==AND && e.entry==0 ) {
      Pattern n( e );
      //std::cerr << "i.e. inserting in " << n << " node "<< *this << endl;
      (*n.children).insert( *this );
      return n;
    } else {
      //std::cerr << "???" << endl;
      return Pattern( AND, *this, e );
    }
  }


  //==============================================================================
  // operator |
  //==============================================================================
  Pattern Pattern::operator|( Pattern const& e ) const {
    //std::cerr << "oring to " << *this << " node " << e << endl; 
    if( connection==OR && entry==0 ) {
      Pattern n( *this );
      if( e.connection==OR ) {
	//std::cerr << "i.e. inserting in " << n << ": ";
	const Mset::const_iterator end = (*e.children).end();
	for( Mset::const_iterator i = (*e.children).begin();
	     i != end; ++i ) {
	  //std::cerr << " node " << **i << endl;
	  (*n.children).insert( **i );
	}
	//std::cerr << endl;
      } else {
	//std::cerr << "i.e. inserting in " << n << " node "<< e << endl;
	(*n.children).insert( e );
      }
      return n;
    }
    else if( e.connection==OR && e.entry==0 ) {
      Pattern n( e );
      //std::cerr << "i.e. inserting in " << n << " node "<< *this << endl;
      (*n.children).insert( *this );
      return n;
    } else { 
      return Pattern( OR, *this, e );
    }
  }


  //==============================================================================
  // operator <
  //==============================================================================
  bool Pattern::operator<( Pattern const& e ) const {
    if( connection==e.connection ) {
      if( connection==Pattern::NO ) {
	if( entry==0 || e.entry==0 )
	  return (entry==0) > (e.entry==0);
	else
	  return *entry < *e.entry;
      } else {
	return *(*children).begin() < *(*e.children).begin();
      }
    } else return connection < e.connection;
  }


  //==============================================================================
  // operator ==
  //==============================================================================
  bool Pattern::operator==( const Pattern& e ) const {
    if( connection != e.connection || 
	( (entry != e.entry) && 
	  ( entry == 0 || e.entry == 0 || *entry != *e.entry)) ) {
	    return false;
	  } else {
	    Pattern::Mset::iterator i1 = (*children).begin();
	    Pattern::Mset::iterator i2 = (*e.children).begin();
	    for(;;) {
	      if( (i1==(*children).end()) != (i2==(*e.children).end()) ) break;
	      if( i1==(*children).end() ) return true;
	      if( **i1 != **i2 ) break;
	      ++i1;
	      ++i2;
	    }
	    return false;
	  }
  }  


  //==============================================================================
  // print
  //==============================================================================
  void Pattern::print( std::ostream& os ) const {
    if( entry!=0 && *entry!=ANY) {
      os << *entry;
      if( (*children).size()!=0 ) os << "=>";
    }
    if( pattern!=0 ) {
      pattern->print( os );
    }
    //os << '(' << (*children).size() << ')';
    if( (*children).size()!=0 ) {
      os << '(';
      const Mset::const_iterator end = (*children).end();
      for( Mset::const_iterator i = (*children).begin();
	   i != end; ) {
	os << *(*i);
	if( ++i != end ) {
	  if( connection==AND ) os << " & ";
	  else os << " | ";
	}
      }
      os << ')';
    }
  }
  

  //==============================================================================
  // countMC_Children
  //==============================================================================
  static size_t 
    countMC_Children( const GObjArray<MC_Track*>::const_iterator begin,
		      const GObjArray<MC_Track*>::const_iterator end,
		      const MC_Track* mother ) {
    size_t counter = 0;
    for( GObjArray<MC_Track*>::const_iterator i = begin; i != end; ++i ) {
      const MC_Track* const t = *i;
      if( t->mother() != MC_Track0 && mother == &*t->mother() ) {
	++counter;
      }
    }
    return counter;
  }


  //==============================================================================
  // select
  //==============================================================================
  int Pattern::select( const GObjArray<PatternMatch*>& tracks,
		       GObjArray<PatternMatch*>& result,
		       FindFlag f ) const {
    return select( reinterpret_cast<const GObjArray<TrackBase*>&>( tracks ),
		   result, f );
  }
  
  int Pattern::select( const GObjArray<TrackBase*>& tracks,
		       GObjArray<PatternMatch*>& result,
		       FindFlag f ) const {
    std::for_each( result.begin(), result.end(), delete_it );
    result.clear();

    if( (*children).size()==0 ) {
      // single track => direct search
      const GObjArray<TrackBase*>::const_iterator tend = tracks.end();
      for( GObjArray<TrackBase*>::const_iterator track = tracks.begin();
	   track != tend; ++track ) {

	//if( entry!=0 && compatible( *entry, **track ) )
	if( pattern==0 || (*pattern)( **track ) ) {
	  // ownership? no?
	  result.push_back( new PatternMatch( **track, *this, false ) ); 
	}
      }
    } else if( (*children).size()==1 ) {
      GObjArray<PatternMatch*> child;
      (**(*children).begin()).select( tracks, child, f );
      if( child.size()==0 ) {
	//continue;
      } else {
	const GObjArray<PatternMatch*>::iterator pend = child.end();
	for( GObjArray<PatternMatch*>::iterator i = child.begin();
	     i != pend; ++i ) {
	  
	  result.push_back( new PatternMatch( **i, *this ) );
	}
      }
    } else if( connection==AND ) {
      std::cerr << "WARNING: DISABLED CODE (T. Glebe, 11.01.2001)" << endl;
#ifdef XXX
      std::vector<GObjArray<PatternMatch*>*> matchedChildren;
      const Mset::iterator mend = (*children).end();
      for( Mset::iterator child = (*children).begin();
	   child != mend; ++child ) {
	
	GObjArray<PatternMatch*>* match = new GObjArray<PatternMatch*>;
	(**child).select( tracks, *match, f );
	matchedChildren.push_back( match );
      }
      //std::cerr << "start combining" << endl;
      for( Combiner comb( matchedChildren ); !comb.end(); ++comb ) {
	
	RecoVertex* v = new RecoVertex( *comb );
	if( v->valid() && entry!=0 && fabs( v->mass() - entry->mass())<.2 ) {
	  //	v->ivFree();
	  //std::cerr << "pushing " << *v << endl;
	  //++numResult;
	  PatternMatch* match = new PatternMatch( *v, *this );
	  match->children
	    = reinterpret_cast<const GObjArray<PatternMatch*>&>(*comb); 
	  result.push_back( match );
	} else delete v;
      }
#endif
    } else {
      //
      // OR
      //
      const Mset::iterator cend = (*children).end();
      for( Mset::iterator j = (*children).begin();
	   j != cend; ++j ) {
	
	// Avoid deletion of already found children.
#ifdef USE_ROOT
	(**j).mcMatches->Clear();
#endif
	GObjArray<PatternMatch*> result;
      
	(**j).select( tracks, result, f );
      }
    }

    return result.size();
  }

  //
  // Search range for this pattern. MC_Track mother is a second condition
  //
  int Pattern::select( const ConstGObjArrayIterator<MC_Track*>& begin,
		       const ConstGObjArrayIterator<MC_Track*>& end,
		       const MC_Track* mother,
		       GObjArray<PatternMatch*>& result,
		       FindFlag f ) const {
    std::for_each( result.begin(), result.end(), delete_it );
    result.clear();
    
    //std::for_each( mcMatches->begin(), mcMatches->end(), delete_it );
    //mcMatches->clear();
    
    for( GObjArray<MC_Track*>::const_iterator i = begin; i != end; ++i ) {
      const MC_Track* const t = *i;
      if( mother==0 ||
	  ( t->mother() != ArtePointer<MC_Track>() && mother==&*t->mother() )) {
	if( entry==0 ||
	    ( (f==CC || t->topp() != entry->topp() ) &&
	      ( f==NONE || t->topp()!=entry->topp()->anti() ))) continue;
      
	//GObjArray<PatternMatch*> tryMatch;
	//= new PatternMatch( *i, this );
	FindFlag fType = (t->topp()==entry->topp()) ? NONE : CC;
	
	if( (*children).size() == 0 ) {
	  
	  PatternMatch* temp = new PatternMatch( *t, *this );
	  if( fType==CC ) temp->ccFlag = true;
	  result.push_back( temp );
	  
	} else if( (*children).size()==1 ) {
	  GObjArray<PatternMatch*> child;
	  (**(*children).begin()).select( begin, end, t, child, fType );
	  if( child.size()==0 ) {
	    continue;
	  } else {
	    const GObjArray<PatternMatch*>::iterator cend = child.end();
	    for( GObjArray<PatternMatch*>::iterator i = child.begin();
		 i != cend;  ++i ) {

	      PatternMatch* temp = new PatternMatch( *t, *this );
	      if( fType==CC ) temp->ccFlag = true;
	      result.push_back( temp );
	    }
	  }
	} else if( connection==AND ) {
	  //
	  // AND
	  //
	  std::vector<GObjArray<PatternMatch*>*> matches;
	  
	  size_t numChildren = countMC_Children( begin, end, t );
	  
	  if( numChildren==(*children).size() ) {
	    
	    Mset::iterator j;
	    const Mset::iterator Mend = (*children).end();
	    const Pattern* lastNode = 0;
	    GObjArray<PatternMatch*>* lastMatch = 0;

	    for( j = (*children).begin(); j != Mend; ++j ) {
	      
	      // Do not make seperate lists for identical children.
	      // -> Combiner avoids double counting
	      if( lastNode!=0 && *lastNode == **j ) {
		assert( lastMatch!=0 );
		matches.push_back( lastMatch );
	      } else {
	      
		// Avoid deletion of already found children.
		//(**j).mcMatches->Clear();
		GObjArray<PatternMatch*> child;
		
		(**j).select( begin+1, end, t, child, fType );
		
		if( child.size()==0 ) {
		  break;
		} else {
		  lastMatch = new GObjArray<PatternMatch*>( child );
		  matches.push_back( lastMatch );
		}
		lastNode = &**j;
	      }
	    }
	    if( j != Mend ) {
	      
	      std::vector<GObjArray<PatternMatch*>*>::iterator mend = matches.end();
	      for( std::vector<GObjArray<PatternMatch*>*>::iterator j
		     = matches.begin(); j != mend; ++j ) {
#ifdef USE_ROOT
		(*j)->Delete();
#endif
	      }
	    } else {
	      std::cerr << " 2.WARNING: missing code (T.Glebe, 11.01.2001)" << endl;
#ifdef XXX
	      for( Combiner comb( matches ); !comb.end(); ++comb ) {
		
		// lock replacement for MC tracks
		std::set<const TrackBase*> checker;
		GObjArray<TrackBase*>::const_iterator tb;
		for( tb = (*comb).begin(); tb != (*comb).end(); ++tb ) {
#ifdef USE_ROOT
		  assert( !tb.InheritsFrom( PatternMatch::Class() ) );
#endif
		  const TrackBase* tmp
		    = reinterpret_cast<PatternMatch*>( *tb )->track;
		  if( std::find( checker.begin(), checker.end(), tmp )
		      !=checker.end() ) break;
		  checker.insert( tmp );
		}
		if( tb!=(*comb).end() ) continue;

		PatternMatch* temp = new PatternMatch( *t, *this );
		if( fType==CC ) temp->ccFlag = true;
		temp->children
		  = reinterpret_cast<const GObjArray<PatternMatch*>&>( *comb );
		result.push_back( temp );
	      }
#endif
	    }
	  }
	} else {
	  //
	  // OR
	  //
	  const Mset::iterator Mend = (*children).end();
	  for( Mset::iterator j = (*children).begin(); j != Mend; ++j ) {
	    
	    // Avoid deletion of already found children.
#ifdef USE_ROOT
	    (**j).mcMatches->Clear();
#endif
	    GObjArray<PatternMatch*> result;
	    
	    (**j).select( begin+1, end, t, result, fType );
	  } // for
	}
      }
    }
    return result.size();
  }


  int Pattern::select( const GObjArray<MC_Track*>& tracks,
		       GObjArray<PatternMatch*>& result,
		       FindFlag f ) const {
    return select( tracks.begin(), tracks.end(), 0, result, f );
  }


#ifdef USE_ROOT
  int Pattern::select( const GClonesArray<MC_Track*>& tracks,
		       GObjArray<PatternMatch*>& result,
		       FindFlag f ) const {
    return select( *reinterpret_cast<const GObjArray<MC_Track*>*>(&tracks),
		   result, f );
  }
  
  
  int Pattern::select( const GClonesArray<TrackBase*>& tracks,
		       GObjArray<PatternMatch*>& result,
		       FindFlag f ) const {
    return select( reinterpret_cast<const GObjArray<TrackBase*>&>( tracks ),
		   result, f );
  }
  
  
  int Pattern::select( const GClonesArray<RecoTrack*>& tracks,
		       GObjArray<PatternMatch*>& result,
		       FindFlag f ) const {
    return select( reinterpret_cast<const GObjArray<TrackBase*>&>( tracks ),
		   result, f );
  }
#endif


  //==============================================================================
  // GetParticles
  //==============================================================================
  const GObjArray<PatternMatch*>& Pattern::GetParticles() const {
    return *mcMatches;
  }
  
  
  //==============================================================================
  // findPattern
  //==============================================================================
  const Pattern* Pattern::findPattern( const Pattern& pat ) const {
    if( isPartialMatch( pat ) ) return this;
    if( children!=0 ) {
      const Mset::iterator end = (*children).end();
      for( Mset::iterator child = (*children).begin();
	   child != end; ++child ) {
	
	const Pattern* check = (**child).findPattern( pat );
	if( check!=0 ) return check;
      }
    }
    return 0;
  }


  //==============================================================================
  // isPartialMatch
  //==============================================================================
  bool Pattern::isPartialMatch( const Pattern& pat ) const {
    if( pat.children!=0 && children!=0 ) {
      Mset::iterator theirChild = (*pat.children).begin();

      const Mset::iterator end = (*children).end();
      for( Mset::iterator myChild = (*children).begin();
	   myChild != end; ++myChild, ++theirChild ) {
	
	if( theirChild == (*pat.children).end() ) return false;
	if( !(**myChild).isPartialMatch( **theirChild ) ) return false;
      }
      if( theirChild != (*pat.children).end() ) return false;
    }
    
    if( pat.entry!=0 && entry!=0 && entry!=pat.entry ) return false;
    //if( pat.pattern!=0 && pattern!=0 &&
    //  (*pat.pattern.isEqualProperty( *pattern ) ) ) return false;
    else return true;
  }


  //==============================================================================
  // addMatch
  //==============================================================================
  int Pattern::addMatch( const PatternMatch& match ) {
    //std::cerr << "push node " << this << " match=" << &match << endl;
    //std::cerr << "matches=";
    //mcMatches->Print();
    mcMatches->push_back( const_cast<PatternMatch*>( &match ) );
    //std::cerr << "afterwards:";
    //mcMatches.Print();
    return 0;
  }


  //==============================================================================
  // removeMatch
  //==============================================================================
  int Pattern::removeMatch( const PatternMatch& match ) {
#ifdef USE_ROOT
    //std::cerr << "remove" << this << ',' << &match << endl;
    mcMatches->Remove( const_cast<PatternMatch*>( &match ) );
    return 0;
#else
    return 0;
#endif
  }

  
#ifdef USE_ROOT
  //==============================================================================
  // Streamer
  //==============================================================================
  void Pattern::Streamer(TBuffer &) {
    std::cerr << "Pattern::Streamer not implemented!" << endl;
  }

  //==============================================================================
  // Print
  //==============================================================================
  void Pattern::Print( Option_t* ) {
    print( std::cout );
    std::cout << endl;
  }

  ClassImp(Pattern)
#endif // USE_ROOT

#ifdef NAMESPACE
} // namespace CLUE
#endif
