/**********************************************************************/
//
// source:
//
// type:      source code
//
// created:   22. Jul 1997
//
// author:    Thorsten Glebe
//            HERA-B Collaboration
//            Max-Planck-Institut fuer Kernphysik
//            Saupfercheckweg 1
//            69117 Heidelberg
//            Germany
//            E-mail: T.Glebe@mpi-hd.mpg.de
//
// Description: construct a tree out of Pdt
//
// changes:
// 22 Jul 1997 created by C. Borgmeier
// 20 Nov 2000 (TG) added namespace CLUE
// 09 Jan 2001 (TG) changed mcMatches to a pointer to decouple header files
// 09 Jan 2001 (TG) changed GObjArray<MC_Track*>::const_iterator to
//                  ConstGObjArrayIterator<MC_Track*> to decouple header files
//
/**********************************************************************/
#ifndef CLUE_PATTERN_HH_
#define CLUE_PATTERN_HH_

#include <iosfwd>
#include <set>
#include "arte/TObject.hh"
#include "gate/gatefwd.hh"
#include "clue/cluefwd.hh"

#ifdef NAMESPACE
namespace CLUE {
#endif

  enum FindFlag { NONE, OR_CC, CC };

  /** Pattern is the central class for pattern-based physics analysis.
      It is the building block for a tree of PdtEntry.
      
      @memo pattern selection
      @author C. Borgmeier
  */
  //==============================================================================
  // Pattern
  //==============================================================================
  class Pattern: public TObject {
  public:
#if !CINTDICT && !CINTOBJECT
    typedef std::multiset<SmartPointer<Pattern> > Mset;
#else
    typedef Int_t Mset;
#endif

    enum ConnectionType { NO, AND, OR } connection;
    
    /** @name --- Constructors --- */
    ///
    Pattern()
      : connection( NO ),
	children( new Mset ),
	entry( 0 ),
	pattern( 0 ) {}
    ///
    Pattern( ClPdtEntry const& leaf,
	     const ClProperty<RecoTrack,int>* pat = 0 );
    ///
    Pattern( ClPdtEntry const& decay, Pattern const& products );
#if !CINTDICT && !CINTOBJECT
    /// not available in CINT
    Pattern( ConnectionType c, Mset::iterator first, Mset::iterator last );
#endif
    ///
    Pattern( ConnectionType c,
	     Pattern const& e1,
	     Pattern const& e2 );
    ///
    Pattern( Pattern const& n );
    ///
    virtual ~Pattern();
    ///
    Pattern& operator=( Pattern const& n );
    
    /** @name --- Operators --- */
    ///
    Pattern  operator&( Pattern const& e ) const;
    ///
    Pattern  operator|( Pattern const& e ) const;
    ///
    bool operator<( const Pattern& e) const;
    ///
    bool operator==( const Pattern& e) const;

    /** @name --- Access functions --- */
    ///
    int select( const GObjArray<TrackBase*>& tracks,
		GObjArray<PatternMatch*>& result,
		FindFlag = NONE ) const;
    
    ///
    int select( const GObjArray<MC_Track*>& tracks,
		GObjArray<PatternMatch*>& result,
		FindFlag = NONE ) const;
    
    ///
    int select( const ConstGObjArrayIterator<MC_Track*>& begin,
		const ConstGObjArrayIterator<MC_Track*>& end,
		const MC_Track* mother,
		GObjArray<PatternMatch*>& result,
		FindFlag = NONE ) const;
    
    ///
    int select( const GObjArray<PatternMatch*>& tracks,
		GObjArray<PatternMatch*>& result,
		FindFlag = NONE ) const;

#ifdef USE_ROOT     // because of typedef to vector
    ///
    int select( const GClonesArray<TrackBase*>& tracks,
		GObjArray<PatternMatch*>& result,
		FindFlag = NONE ) const;
    
    ///
    int select( const GClonesArray<RecoTrack*>& tracks,
		GObjArray<PatternMatch*>& result,
		FindFlag = NONE ) const;
    
    ///
    int select( const GClonesArray<MC_Track*>& tracks,
		GObjArray<PatternMatch*>& result,
		FindFlag = NONE ) const;
#endif

    /// Return the list of particles, which have been matched to this Pattern.
    const GObjArray<PatternMatch*>& GetParticles() const;
    
    /** @name --- Expert functions --- */
    /// Find pattern among the children of this.
    const Pattern* findPattern( const Pattern& pat ) const;
  
    /// Check if pattern is partially equal to this.
    bool isPartialMatch( const Pattern& pat ) const;

    ///
    int removeMatch( const PatternMatch& );

    ///
    int addMatch( const PatternMatch& );

    /// used by operator<<()
    void print( std::ostream& os ) const;
  
  public:
    ///
    Mset* children;
    
    ///
    const ClPdtEntry* entry;
    
    ///
    const ClProperty<RecoTrack,int>* pattern;
    
    /// to decouple GObjArray from Pattern
    mutable GObjArray<PatternMatch*>* mcMatches;
    
#ifdef USE_ROOT
    /// only available in ROOT
    virtual void Print( Option_t* = 0 );
    
    ClassDef(Pattern,1)
#endif
  }; // end of class Pattern


  //==============================================================================
  // operator <<
  //==============================================================================
  inline std::ostream& operator<<( std::ostream& os, Pattern const& t ) {
    t.print( os );
    return os;
  }

#ifdef NAMESPACE
} // namespace CLUE
#endif
#endif
