// *****************************************************************************
//
// source:
//
// type:      source code
//
// created:   22. Jul 1997
//
// author:    Thorsten Glebe
//            HERA-B Collaboration
//            Max-Planck-Institut fuer Kernphysik
//            Saupfercheckweg 1
//            69117 Heidelberg
//            Germany
//            E-mail: T.Glebe@mpi-hd.mpg.de
//
// Description: describe a track pattern
//
// changes:
// 22 Jul 1997 creation and 1st implementation by C. Borgmeier
// 03 Nov 2000 (TG) removed #include statemens
// 20 Nov 2000 (TG) added namespace CLUE
// 12 Jan 2001 (TG) removed obsolete include files
//
// *****************************************************************************
#include <iostream>

#include "clue/cluearte.hh"
// #include "geometry/ClLorentzVector.hh" // needed by MC_Track.hh
// #include "geometry/Point3D.hh"  // needed by RecoSegment.hh
#include "clue/TrackBase.hh"
#include "gate/ObjArray.hh"     // needed for Pattern.hh
#include "gate/ClonesArray.hh"  // needed for Pattern.hh
#include "clue/SmartPointer.hh" // needed for Pattern.hh
#include "clue/Pattern.hh"
#include "clue/PatternMatch.hh"
//  #include "clue/RecoVertex.hh"
//#include "clue/MC_Vertex.hh"    // needed by MC_Track
// #include "clue/MC_Track.hh"
#include "clue/PdtEntry.hh"     // needed by pattern->mass()

#ifdef NAMESPACE
namespace CLUE {
#endif

  //==============================================================================
  // Constructors
  //==============================================================================
  PatternMatch::PatternMatch( const TrackBase& match,
			      const Pattern& pat,
			      bool owns ) :
    track( owns ? static_cast<TrackBase*>(const_cast<TrackBase&>(match).Clone())
	   : &const_cast<TrackBase&>(match) ),
    pattern( const_cast<Pattern*>( &pat ) ),
    ccFlag( 0 ),
    ownsTrack( owns ) {
    const_cast<Pattern&>(pat).addMatch( *this );
  }


  PatternMatch::PatternMatch() :
    pattern(0),
    ccFlag( 0 ),
    ownsTrack( false ) {
    //cerr << "create B " << this << endl;
  }


  PatternMatch::PatternMatch( const PatternMatch& rhs ) {
    //cerr << "create C " << this << endl;
    *this = rhs;
  }


  PatternMatch::~PatternMatch() {
    //cerr << "deleting " << this << ',' << pattern << endl;
    if( pattern!=0 ) const_cast<Pattern*>(pattern)->removeMatch( *this );
    if( ownsTrack ) delete track;
#ifdef USE_ROOT
    children.Delete();
#else
#warning missing statement ( C.B. 8/17/99 )
#endif
  }


  //==============================================================================
  // operator =
  //==============================================================================
  const PatternMatch& PatternMatch::operator=( const PatternMatch& rhs ) {
    //cerr << "assigning from " << &rhs << " to " << this << endl;
    //cerr << "from pattern= " <<rhs.pattern << " to " << pattern << endl;
    if( this == &rhs ) return *this;

    if( rhs.ownsTrack ) {
      track     = static_cast<TrackBase*>( rhs.track->Clone() );
    } else {
      track     = rhs.track;
    }
    ownsTrack   = rhs.ownsTrack;
    if( pattern!=0 ) const_cast<Pattern*>(pattern)->removeMatch( *this );
    pattern   = rhs.pattern;
    if( pattern!=0 ) const_cast<Pattern*>(pattern)->addMatch( *this );
    
    ccFlag = rhs.ccFlag;

#ifdef USE_ROOT
    children.Delete();
#else
#warning missing statement ( C.B. 8/17/99 )
#endif
    std::copy( rhs.children.begin(), rhs.children.end(),
	       back_inserter( children ) );
    
    //cout << "deleting " << nextMatch << endl;
    //delete nextMatch;
    //nextMatch = rhs.nextMatch? new PatternMatch( *rhs.nextMatch ): 0;
    
    return *this;
  }


  //==============================================================================
  // transfer
  //==============================================================================
  void PatternMatch::transfer( PatternMatch& target ) {
    //cerr << "transfer" << endl;
    
    target.ccFlag   = ccFlag;
    target.track    = track;
#ifdef USE_ROOT
    target.children.Delete();
#else
#warning missing statement ( C.B. 8/17/99 )
#endif
    target.children = children;
    
    target.pattern  = pattern;
    if( pattern!=0 ) const_cast<Pattern*>(pattern)->addMatch( target );

  //delete target.nextMatch;
  //target.nextMatch = nextMatch;

#ifdef USE_ROOT
    children.Clear();
#else
#warning missing statement ( C.B. 8/17/99 )
#endif
  //nextMatch = 0;
  }


  //==============================================================================
  // print
  //==============================================================================
  std::ostream& PatternMatch::print( std::ostream& os ) const {
    os << '[';
    if( pattern != 0 ) {
      if( ccFlag ) os << '~';
      os << *pattern << ',';
    }
    else os << "(no pattern)";
    os << "(m=" << mass() << ')';
    if( track != 0 ) track->print( os );
#ifdef USE_ROOT
    if( children.size()!=0 ) os << endl << children;
    os << ']';
#else
#warning missing statement ( C.B. 8/17/99 )
#endif
    return os;
  }


  //==============================================================================
  // lock
  //==============================================================================
  void PatternMatch::lock() {
    if( track ) const_cast<TrackBase*>(track)->lock();
  }


  //==============================================================================
  // unlock
  //==============================================================================
  void PatternMatch::unlock() {
    if( track ) const_cast<TrackBase*>( track )->unlock();
  }


  //==============================================================================
  // locked
  //==============================================================================
  int PatternMatch::locked( int bit ) const {
    return track == 0 || track->locked( bit );
  }


  //==============================================================================
  // isCC
  //==============================================================================
  bool PatternMatch::isCC() const {
    return ccFlag;
  }


  //==============================================================================
  // mass
  //==============================================================================
  float PatternMatch::mass() const {
    //std::cout << "pm::getmass " << pattern;
    //if( pattern!=0 ) {
    //  std::cout << "pat.entry=" << pattern->entry
    //      << " mass=" << pattern->entry->mass();
    //}
    //std::cout << std::endl;
    if( pattern!=0 && pattern->entry!=0 ) return pattern->entry->mass();
    else if( track!=0 ) return track->mass();
    else return 0;
  }
  

  //==============================================================================
  // getRecoTrack
  //==============================================================================
  const RecoTrack* PatternMatch::getRecoTrack() const {
    return track!=0? track->getRecoTrack(): 0;
  }


#ifdef XXX
  //==============================================================================
  // GetLorentzVector
  //==============================================================================
  LorentzVector PatternMatch::GetLorentzVector() const {
    if( track!=0 ) {
      ThreeVector p = track->GetLorentzVector().vect();
      const float m = mass();
      return LorentzVector( p, sqrt( p.mag2() + m*m ) );
    }
    else return LorentzVector();
  }
#endif

#ifdef USE_ROOT
  //==============================================================================
  // Print
  //==============================================================================
  void PatternMatch::Print( Option_t* ) {
    print( std::cout );
    cout << endl;
  }

  ClassImp(PatternMatch)
#endif
#ifdef NAMESPACE
} // namespace CLUE
#endif
