#ifndef PDT_ENTRY_HH
#define PDT_ENTRY_HH
// *****************************************************************************
//
// source:
//
// type:      source code
//
// created:   20. Jul 1997
//
// author:    Thorsten Glebe
//            HERA-B Collaboration
//            Max-Planck-Institut fuer Kernphysik
//            Saupfercheckweg 1
//            69117 Heidelberg
//            Germany
//            E-mail: T.Glebe@mpi-hd.mpg.de
//
// Description: Class to describe a particle decay pattern
//
// changes:
// 20 Jul 1997 creation by C. Borgmeier
// 03 Nov 2000 (TG) removed #include "arte/arte.hh", added a constructor
// 20 Nov 2000 (TG) added namespace CLUE
// 02 Jan 2001 (TG) made set_topp argument const
// 02 Jan 2001 (TG) changed type of iTopp to UShort_t
// 12 Jan 2001 (TG) made most operator.. function non-inline to avoid class couplings
//
// *****************************************************************************
#include <iosfwd>
#include "arte/TObject.hh"
#include "arte/artefwd.hh"
#include "gate/gatefwd.hh"
#include "clue/cluefwd.hh"

#ifdef NAMESPACE
namespace CLUE {
#endif
  class abc; // to fake doc++

  /** Particle Data Table entry class
      @memo Particle Data
      @author C. Borgmeier
  */
  //==============================================================================
  // ClPdtEntry
  //==============================================================================
  class ClPdtEntry: public TObject {
  public:
    /** @name --- Constructors --- */
    ///
    ClPdtEntry();
    ///
    ClPdtEntry( const char* name );
    ///
    ClPdtEntry( const char* name, const ArtePointer<Topp>& t );
    //  ClPdtEntry( const char* name, ArtePointer<Topp> t = 0 );
    /// constructor
    ClPdtEntry( const char* name, unsigned int t );
    ///
    ~ClPdtEntry();
  
    /** @name --- Operators --- */
    ///
    inline bool operator==( const ClPdtEntry& rhs ) const
      { return iTopp == rhs.iTopp; }
    ///
    inline bool operator!=( const ClPdtEntry& rhs ) const 
      { return iTopp != rhs.iTopp; }
    ///
    bool operator==( const ArtePointer<Topp>& rhs ) const;
    ///
    bool operator!=( const ArtePointer<Topp>& rhs ) const;
    ///
    bool operator==( const RecoTrack& rhs ) const;
    ///
    bool operator!=( const RecoTrack& rhs ) const;
    ///
    bool operator==( const RecoVertex& rhs ) const;
    ///
    bool operator!=( const RecoVertex& rhs ) const;
    ///
    bool operator==( const MC_Track& rhs ) const;
    ///
    bool operator!=( const MC_Track& rhs ) const;
    ///
    Pattern operator&( ClPdtEntry const& e ) const;
    ///
    Pattern operator|( ClPdtEntry const& e ) const;
    ///
    Pattern operator&( Pattern const& e ) const;
    ///
    Pattern operator|( Pattern const& e ) const;
    ///
    Pattern operator()( const ClProperty<RecoTrack,int>& prop ) const;
    ///
    bool operator<( ClPdtEntry const & el ) const;
    
    /** @name --- Access functions --- */
    ///
    Pattern decaysTo( Pattern const& l ) const;
    ///
    Pattern andp( ClPdtEntry const& e ) const;
    ///
    Pattern orp( ClPdtEntry const& e ) const;
    ///
    Pattern andp( Pattern const& e ) const;
    ///
    Pattern orp( Pattern const& e ) const;  
    ///
    Pattern with( const ClProperty<RecoTrack,int>& prop ) const;

    ///
    int select( const GObjArray<RecoTrack*>& tracks,
		GObjArray<PatternMatch*>& result ) const;
  
    ///
    int select( const GObjArray<MC_Track*>& tracks,
		GObjArray<PatternMatch*>& result ) const;
    
#ifdef USE_ROOT
    ///
    int select( const GClonesArray<RecoTrack*>& tracks,
		GObjArray<PatternMatch*>& result ) const;
    
    ///
    int select( const GClonesArray<MC_Track*>& tracks,
		GObjArray<PatternMatch*>& result ) const;
#endif

    /// mass of pdt entry
    double mass() const;

    /// topp table pointer
    inline ArtePointer<Topp> topp() const { return getPTR<Topp>( iTopp ); }
    
    /** @name --- Expert functions --- */
    /// set topp table pointer
    inline void set_topp(const ArtePointer<Topp>& topp) { iTopp = getIDX<Topp>( topp ); }
    
    /// used by operator<<()
    void print( std::ostream& os ) const;
    
    /// used to fill Topp container
    //    static void updateTOPP();
    
    char name[16];
    
  private:
    UShort_t iTopp;      // Topp table id
    
    ///
    const ClPdtEntry& operator=( const ClPdtEntry& );
    
  public:
#ifdef USE_ROOT
    /// only available in ROOT
    virtual void Print( Option_t* = "" );
    
    ClassDef(ClPdtEntry,1)
#endif
  }; // end of class ClPdtEntry


//    //==============================================================================
//    // operator ==
//    //==============================================================================

//    inline int operator==( const ArtePointer<Topp>& e1, const ClPdtEntry& e2 )
//    { return e1==e2.topp(); }

//    int operator==( const ClPdtEntry& e1, const RecoTrack& e2 );
//    int operator==( const RecoTrack& e1, const ClPdtEntry& e2 );
//    int operator==( const ClPdtEntry& e1, const RecoVertex& e2 );
//    int operator==( const RecoVertex& e1, const ClPdtEntry& e2 );
//    int operator==( const ClPdtEntry& e1, const MC_Track& e2 );
//    int operator==( const MC_Track& e1, const ClPdtEntry& e2 );

//    //==============================================================================
//    // operator ==
//    //==============================================================================
  
//    inline int operator!=( const ArtePointer<Topp>& e1, const ClPdtEntry& e2 )
//    { return e1!=e2.topp(); }
  
//    int operator!=( const ClPdtEntry& e1, const RecoTrack&  e2 );
//    int operator!=( const RecoTrack&  e1, const ClPdtEntry& e2 );
//    int operator!=( const ClPdtEntry& e1, const RecoVertex& e2 );
//    int operator!=( const RecoVertex& e1, const ClPdtEntry& e2 );
//    int operator!=( const ClPdtEntry& e1, const MC_Track&   e2 );
//    int operator!=( const MC_Track&   e1, const ClPdtEntry& e2 );
  

  //==============================================================================
  // operator <<
  //==============================================================================
  inline std::ostream& operator<<( std::ostream& os, const ClPdtEntry& node ) {
    node.print( os );
    return os;
  }

  extern ClPdtEntry ANY;
  extern ClPdtEntry X;

#ifdef NAMESPACE
} // namespace CLUE
#endif
#endif
