// ********************************************************************
//
// source:
//
// type:      source code
//
// created:   Mon Jul 16 18:17:47 MEST 2001
//
// author:    glebe
//
// Description: 
//
// changes:
// 14 Sep 2001 (TG) renamed sdecaysTo() to decaysTo()
//
// ********************************************************************

#include "TH1.h"
#include "TH2.h"
#include "TNtuple.h"
#include "TPaveText.h"
#include "BEE.hh"
#include "include/LambdaAnalysis.hh"
#include "smatrix/SDistance.hh" 

int checkPID(const RecoTrack& rtra);


//==============================================================================
// Constructor
//==============================================================================
LambdaAnalysis::LambdaAnalysis(RtEvent* evt) :
 event(evt),
 Piplus(false),
 Piminus(false),
 lambda(false),
 lambdabar(false)
{}

//==============================================================================
// EventSelect
//==============================================================================
bool LambdaAnalysis::EventSelect() {
  if(ArteTable<RecoVertex>::size() == 0)
    return false;
  return true;
}


//==============================================================================
// Dmain - impact parameter to closest main vertex
//==============================================================================
double LambdaAnalysis::Dmain(const TrackIf& track) {

  double dmin = 100000.;

  const ArteTableIterator<RecoVertex> RVERend = ArteTable<RecoVertex>::end();
  for(ArteTableIterator<RecoVertex> RVERit = ArteTable<RecoVertex>::begin();
      RVERit != RVERend; ++RVERit) {
    if((*RVERit)->valid()) {
      const double dist = distance(track,**RVERit);
      // cout << "dist: " << dist;
      if(dist < dmin) dmin = dist;
    }
  }

  // cout << " dmin: " << dmin << endl;
  return dmin;
}       


//==============================================================================
// Dist  - distance to closest primary vertex
//==============================================================================
double LambdaAnalysis::Dist(SVertex<2>& vtx) {

  double dmin = 100.;
  ArtePointer<RecoVertex> closest;

  const ArteTableIterator<RecoVertex> RVERend = ArteTable<RecoVertex>::end();
  for(ArteTableIterator<RecoVertex> RVERit = ArteTable<RecoVertex>::begin();
      RVERit != RVERend; ++RVERit) {
    const ArtePointer<RecoVertex>& rver = *RVERit;
    if(rver->valid()) {
      const double dist = distance(vtx,*rver);
      if(dist < dmin) {
        dmin    = dist;
        closest = rver;
      }
    }
  }

  return sqrt(pow(closest->vx() - vtx.vx(),2)
	      + pow(closest->vy() - vtx.vy(),2)
	      + pow(closest->vz() - vtx.vz(),2));
}

//==============================================================================
// DefinePattern
//==============================================================================
bool LambdaAnalysis::DefinePattern() {

  Piminus    = pi_minus.createPattern();
  Piplus     = pi_plus.createPattern();
  Proton     = proton.createPattern();
  AntiProton = p_bar.createPattern();

  lambda = Lam_zero.decaysTo(Piminus & Proton);
  lambda->with(prob > 0.001 & vz > -5.);

  lambdabar = Lam_bar_zero.decaysTo(Piplus & AntiProton);
  lambdabar->with(prob > 0.001 & vz > -5.);
  
  return true;
}

//==============================================================================
// DefineHistograms
//==============================================================================
bool LambdaAnalysis::DefineHistograms() {

  Lambda    = new TNtuple("Lambda","Lambdas","m");
  Lambdabar = new TNtuple("Lambdabar","Lambdabars","m");

  return true;
}


//==============================================================================
// Vertexing (with cuts)
//==============================================================================
bool LambdaAnalysis::Vertexing(DP_trait<2>::DPMC_sp& cand, TNtuple* ntuple) {

  typedef DP_trait<2>::SDPM_l   SDPM_l;
  typedef DP_trait<2>::SDPM_it  SDPM_it;
  typedef DP_trait<2>::SDPM_cit SDPM_cit;


  const SDPM_it DPM_end = cand->matches().end();
  for(SDPM_it DPM_it = cand->matches().begin(); DPM_it != DPM_end;
      ++DPM_it) {
    SDecayPatternMatch<2>& dpm = *DPM_it;
    if(!dpm.isValid()) continue;    
    dpm.calc_mother_tr();
    
    SVertex<2>& vtx = dpm.vertex();
    SVector<double,3> PtotUnit(vtx.pvec().unit());
    SVector<double,3> Proton_(vtx.kalman(1).pvec());
    
    double Proton_l = dot(Proton_, PtotUnit);
    // double frac     = (Pplus_l - Pminus_l) / (Pplus_l + Pminus_l);
    double pt = mag(Proton_ - PtotUnit * Proton_l);     

    float k_mass = vtx.mass(SVector<double,2>(pi_plus.mass(),pi_plus.mass()));
    if((k_mass > 0.485) && (k_mass < 0.515)) continue;

    float x[1];
    x[0] = dpm.mass();
    
    double ctau = x[0]*Dist(vtx)/vtx.p();
    if(ctau*pt < 0.05) continue;

    if(Dmain(vtx) > 0.06) continue;

    ntuple->Fill(x);
  } // for SDPM


  return true;
}

//==============================================================================
// Event
//==============================================================================
bool LambdaAnalysis::Event() {

  // clear Particle container
  Piplus->clear();
  Piminus->clear();
  Proton->clear();
  AntiProton->clear();

  //----------------- select Particles -------------------
  const ArteTableIterator<RecoTrack> RTend = ArteTable<RecoTrack>::end();
  for(ArteTableIterator<RecoTrack> RTit = ArteTable<RecoTrack>::begin();
      RTit != RTend; ++RTit) {
    RecoTrack& rtra = **RTit;

    if(rtra.hsi() < 6 ||
       (rtra.hou() + rtra.hin() < 16)) continue;
    if(rtra.pt()<0.1 || rtra.p() < 2. || fabs(rtra.xf(pi_plus.mass())) > 0.2)
      continue;

    // prepare track for vertexing
    rtra.sPinvToP();
    rtra.propagate(0);
    rtra.FillCINV();

    // inclusive distributions and PID assignment:
    const int PID  = checkPID(rtra);
    const int chrg = rtra.charge();

    // fill the particle candidates
    if(PID == 0 || PID == PID::pion) {
      if(chrg > 0) {
	Piplus->push_back(rtra);
	//      cout << " pi+ found " << endl;
      }
      if(chrg < 0) {
	Piminus->push_back(rtra);
	//      cout << " pi- found " << endl;
      }
    } // if(PID==0

    if(PID == PID::proton) {
      if(chrg > 0) {
	Proton->push_back(rtra);
	//      cout << " pi+ found " << endl;
      }
      if(chrg < 0) {
	AntiProton->push_back(rtra);
	//      cout << " pi- found " << endl;
      }
    }


  } // for(ArteTable<RecoTrack>


  //------------------ Pattern ----------------------
  lambda->combine();
  lambdabar->combine();
  lambda->prob_collision();
  lambdabar->prob_collision();

  Vertexing(lambda,Lambda);
  Vertexing(lambdabar,Lambdabar);

  return true;
}

//==============================================================================
// EventCleanup
//==============================================================================
bool LambdaAnalysis::EventCleanup() {
  return true;
}

//==============================================================================
// End processing
//==============================================================================
bool LambdaAnalysis::End() {
  return true;
}

