// ********************************************************************
//
// source:
//
// type:      source code
//
// created:   Die Aug 14 10:33:50 MEST 2001
//
// author:    glebe
//
// Description: D0 decaying to pi+ K- are selected from MC truth and only those
//              decays are saved in the containers, where the pi and the K was
//              reconstructed (MC_Track <-> RecoTrack relation is set). In case
//              the mother track is not reconstructed (always the case with
//              D0), the vertex is computed automatically from the
//              reconstructed outgoing tracks. The D0 mother track is not
//              computed automatically, but this can be done later by the user.
//              Selecting reconstructed decays in MC using the
//              selectRECOfromMCtruth() function works for decay pattern of
//              arbitrary(!) complexity.
//              Limitation: decays where the same particle type occurs with
//              different cuts is not treated. For example:
//              D0 -> K-pi+pi+pi- and the two pi+ have different cuts
//
// changes:
// 14 Sep 2001 (TG) renamed sdecaysTo() to decaysTo()
//
// ********************************************************************

#include "TH1.h"
#include "TH2.h"
#include "TNtuple.h"
#include "TPaveText.h"
#include "BEE.hh"
#include "include/MCAna2.hh"

//==============================================================================
// Constructor
//==============================================================================
MCAna2::MCAna(RtEvent* evt) :
 event(evt)
{}

//==============================================================================
// EventSelect
//==============================================================================
bool MCAna2::EventSelect() {
  return true;
}

//==============================================================================
// DefinePattern
//==============================================================================
bool MCAna2::DefinePattern() {
  
  PiPlus = pi_plus.createPattern();
  KMinus = K_minus.createPattern();

  Dzero  = D_zero.decaysTo(PiPlus & KMinus);

  KMinus->with(pt > 0.2);
  PiPlus->with(pt > 0.1);
  Dzero->with(pt > 0.15);

  cout << "D_zero: " << endl << *Dzero->pattern() << endl;

  return true;
}

//==============================================================================
// DefineHistograms
//==============================================================================
bool MCAna2::DefineHistograms() {

  D0mass = new TH1F("D0mass","D0 mass, reconstructed",20,1.75,1.95);
  return true;
}

//==============================================================================
// Event
//==============================================================================
bool MCAna2::Event() {

  PiPlus->clear();
  KMinus->clear();

  // Prepare RecoTracks for vertexing
  const ArteTableIterator<RecoTrack> RTend = ArteTable<RecoTrack>::end();
  for(ArteTableIterator<RecoTrack> RTit = ArteTable<RecoTrack>::begin();
      RTit != RTend; ++RTit) {
    const ArtePointer<RecoTrack> rtra = *RTit;

    rtra->sPinvToP();
    rtra->propagate(0);
    rtra->FillCINV();
  }

  /// select all reconstructed decay chains, compute D0 vertex for those
  Dzero->selectRECOfromMCtruth();

  cout << "Dzero size: " << Dzero->matches().size() << endl;
#ifdef XXX
  cout << "Dzero size: " << Dzero->matches().size() 
       << " PiPlus size: " << PiPlus->matches().size()
       << " KMinus size: " << KMinus->matches().size() 
       << endl;

  cout << "PiPlus container: " << endl;
  cout << PiPlus;
  cout << "KMinus container: " << endl;
  cout << KMinus;

  cout << "Dzero container: " << endl;
  cout << Dzero;
#endif


  // fill histogram
  typedef DP_trait<2>::SDPM_it  SDPM_it;

  const SDPM_it PM_end = Dzero->matches().end();
  for(SDPM_it PM_it = Dzero->matches().begin();
      PM_it != PM_end; ++PM_it) {
    SDecayPatternMatch<2>& Meson_sdpm = *PM_it;

    // fill inv. mass of the reconstructed D0
    D0mass->Fill(Meson_sdpm.mass());
  }

  return true;
}

//==============================================================================
// EventCleanup
//==============================================================================
bool MCAna2::EventCleanup() {
  return true;
}

//==============================================================================
// End processing
//==============================================================================
bool MCAna2::End() {
  return true;
}

