// ********************************************************************
//
// source:
//
// type:      source code
//
// created:   Die Aug 14 10:33:50 MEST 2001
//
// author:    glebe
//
// Description: This job selects from MC truth those D0 -> K- pi+ decays, which
//              were not fully reconstructed (either the K or the pi or both
//              are not reconstructed). In case of the reconstructed track, the
//              RecoTrack is stored via a SPatternMatch object in the
//              container. In case of the not reconstructed track the MC truth
//              track (MC_Track) object is stored via a SPatternMatch object in
//              the SPatternMatchContainer. When looping over the tracks of the
//              SDecayPatternMatch object one can get via the PatternMatch()
//              member function the pointers to the PatternMatch objects. The
//              isMCmatch() member functions tells you whether the track was
//              reconstructed (=false) or not (=true). In this way (see below)
//              you can look at both reconstructed and not reconstructed (=MC
//              truth) tracks.
//
// changes:
// 14 Sep 2001 (TG) renamed sdecaysTo() to decaysTo()
//
// ********************************************************************

#include "TH1.h"
#include "TH2.h"
#include "TNtuple.h"
#include "TPaveText.h"
#include "BEE.hh"
#include "include/MCAna3.hh"

//==============================================================================
// Constructor
//==============================================================================
MCAna3::MCAna3(RtEvent* evt) :
 event(evt)
{}

//==============================================================================
// EventSelect
//==============================================================================
bool MCAna3::EventSelect() {
  return true;
}

//==============================================================================
// DefinePattern
//==============================================================================
bool MCAna3::DefinePattern() {
  
  PiPlus = pi_plus.createPattern();
  KMinus = K_minus.createPattern();

  Dzero  = D_zero.decaysTo(PiPlus & KMinus);

  KMinus->with(pt > 0.2);
  PiPlus->with(pt > 0.1);
  Dzero->with(pt > 0.15);

  cout << "D_zero: " << endl << *Dzero->pattern() << endl;
  cout << "---------------------------------" << endl;
  return true;
}

//==============================================================================
// DefineHistograms
//==============================================================================
bool MCAna3::DefineHistograms() {

  PtMC   = new TH1F("PtMC","pt of not reconstructed MC tracks", 100, 0., 1.);
  PtRECO = new TH1F("PtRECO","pt of reconstructed tracks", 100, 0., 1.);
  return true;
}

//==============================================================================
// Event
//==============================================================================
bool MCAna3::Event() {

  PiPlus->clear();
  KMinus->clear();

  unsigned int nk0s = 0;

  // select partially or not reconstructed D0 -> K- pi+ decays
  Dzero->selectNORECOfromMCtruth();

#ifdef XXX
  cout << "Dzero size: " << Dzero->matches().size() 
       << " PiPlus size: " << PiPlus->matches().size()
       << " KMinus size: " << KMinus->matches().size() 
       << endl;

  cout << "PiPlus container: " << endl;
  cout << PiPlus;
  cout << "KMinus container: " << endl;
  cout << KMinus;

  cout << "Dzero container: " << endl;
  cout << Dzero;
#endif

  // look at partially reconstructed D0
  typedef DP_trait<2>::SDPM_it  SDPM_it;

  const SDPM_it PM_end = Dzero->matches().end();
  for(SDPM_it PM_it = Dzero->matches().begin();
      PM_it != PM_end; ++PM_it) {
    SDecayPatternMatch<2>& dpm = *PM_it;

    for(unsigned int i=0; i<2; ++i) {
      const PatternMatchIf& t = *dpm.PatternMatch(i);

      if(t == K_minus)
	cout << "K_minus found!" << endl;
      if(t == pi_plus)
	cout << "Pi_plus found!" << endl;

      if(t.isMCmatch() == true) {
	const MC_Track& m = t.getMC_Track();
	PtMC->Fill(m.pt());
      } else {
	const RecoTrack& r = t.getRecoTrack();
	PtRECO->Fill(r.pt());
      }
    }
  }

  return true;
}

//==============================================================================
// EventCleanup
//==============================================================================
bool MCAna3::EventCleanup() {
  return true;
}

//==============================================================================
// End processing
//==============================================================================
bool MCAna3::End() {
  return true;
}

