// **********************************************************************
//
// source:
//
// type:      source code
//
// created:   16.07.2001
//
// author:    Thorsten Glebe
//            HERA-B Collaboration
//            Max-Planck-Institut fuer Kernphysik
//            Saupfercheckweg 1
//            69117 Heidelberg
//            Germany
//            E-mail: T.Glebe@mpi-hd.mpg.de
//
// Description: Particle ID routine for RecoTracks. The routine takes
//              an ArtePointer<RecoTrack> as an argument.
//
// changes:
// 16 Jul 2001 (TG) creation from Michael Schmellings code
// 14 Sep 2001 (TG) bug fix: concerning cov_pinv (renamed to cov_p)
//
// **********************************************************************
#include "BEE.hh"

/** @name Particle Identification algorithm
 */
//==============================================================================
// checkPID
//==============================================================================
int checkPID(const RecoTrack& rtra) {
  //  cout << "checkPID: check Track " << *rtra << endl;

  static const float THO2    = square(0.0521);  // cerenkov angle squared for beta=1 particles
  static const float DTHCER2 = square(0.00035); // assumed uncertainty of THO

  int PID = 0;

  if(rtra.rlt() > 0) {
    if(rtra.lre()  > 0.99) PID = PID::electron;
    if(rtra.lrmu() > 0.99) PID = PID::muon;
    if(rtra.lrpi() > 0.30) PID = PID::pion;
    if(rtra.lrk()  > 0.60) PID = PID::kaon;
    if(rtra.lrp()  > 0.30) PID = PID::proton;
  } // if RISE is present
  else {
    if(rtra.rls() > 0) {
      if(rtra.lse()  > 0.99) PID = PID::electron;
      if(rtra.lsmu() > 0.99) PID = PID::muon;
      if(rtra.lspi() > 0.30) PID = PID::pion;
      if(rtra.lsk()  > 0.60) PID = PID::kaon;
      if(rtra.lsp()  > 0.30) PID = PID::proton;
    } // if RITER is present
  } // if rlt > 0

  //  cout << "PID from RISE/RITER: " << PID << endl;

  const float Cer2 = square(rtra.cer());
  const float P2   = square(rtra.p());
  const float Msq  = P2*(THO2 - Cer2);

  // if neither RISE nor RITER present:

  if(Cer2 > 0 && PID==0) {
    float dpsq = rtra.cov_p();
    float vMsq = 4.*P2*(Msq*Msq/dpsq + P2*Cer2*DTHCER2);
    float Lelectron = square(Msq - square(e_minus.mass()))/vMsq;
    float Lmuon     = square(Msq - square(mu_minus.mass()))/vMsq;
    float Lpion     = square(Msq - square(pi_minus.mass()))/vMsq;
    float Lkaon     = square(Msq - square(K_minus.mass()))/vMsq;
    float Lproton   = square(Msq - square(proton.mass()))/vMsq;
    if(Lelectron < 3) Lelectron = exp(-0.5*Lelectron); else Lelectron = 0;
    if(Lmuon     < 3) Lmuon     = exp(-0.5*Lmuon);     else Lmuon     = 0;
    if(Lpion     < 3) Lpion     = exp(-0.5*Lpion);     else Lpion     = 0;
    if(Lkaon     < 3) Lkaon     = exp(-0.5*Lkaon);     else Lkaon     = 0;
    if(Lproton   < 3) Lproton   = exp(-0.5*Lproton);   else Lproton   = 0;
    const float Lsum = Lelectron + Lmuon + Lpion + Lkaon + Lproton;
    if(Lsum > 0) {
      Lelectron /= Lsum;
      Lmuon     /= Lsum;
      Lpion     /= Lsum;
      Lkaon     /= Lsum;
      Lproton   /= Lsum;
    }
    if(Lelectron > 0.99) PID = PID::electron;
    if(Lmuon     > 0.99) PID = PID::muon;
    if(Lpion     > 0.30) PID = PID::pion;
    if(Lkaon     > 0.60) PID = PID::kaon;
    if(Lproton   > 0.30) PID = PID::proton;
  } // Cer2>0 && PID==0

  // if no RICH information is available - believe the muon system
  if(PID != PID::kaon && PID != PID::proton && rtra.hmu()>0) PID = PID::muon;

  //  cout << "FINAL PID: " << PID << endl;

  return PID;
}
