#ifndef G_CLONES_ARRAY_HH
#define G_CLONES_ARRAY_HH
// ********************************************************************
//
// source:
//
// type:      source code
//
// created:   27. Feb 1999
//
// author:    Thorsten Glebe
//            HERA-B Collaboration
//            Max-Planck-Institut fuer Kernphysik
//            Saupfercheckweg 1
//            69117 Heidelberg
//            Germany
//            E-mail: T.Glebe@mpi-hd.mpg.de
//
// Description: GClonesArray is a template wrapper 
//              for the ROOT TClonesArray class.
//
// changes:
// 27 Feb 1999 creation by C. Borgmeier
// 13 Nov 2000 (TG) added doc++ comments
// 14 Nov 2000 (TG) declared some member functions inline
// 20 Nov 2000 (TG) added namespace CLUE
//
// ********************************************************************

#ifdef USE_ROOT

#include <iterator.h>
#include <TClonesArray.h>

#ifdef NAMESPACE
#include "gate/RtypesT.h"

namespace CLUE {
#endif

  template <class T> class GClonesArray;
  template <class T> class GObjArrayIterator;
  template <class T> class ConstGObjArrayIterator;
  template <class T> class GObjArray;

  
  /** GClonesArray is a template wrapper for the ROOT TClonesArray class.
      This class is only available in ROOT. Outside ROOT, GClonesArray is
      typedef'd to a STL vector.
      @memo Array of similar objects.
      @author C. Borgmeier
  */
  //==============================================================================
  // GClonesArray
  //==============================================================================
  template<class T>
    class GClonesArray: public TClonesArray {
    public:
      typedef T                 value_type;
      typedef value_type&       reference;
      typedef value_type*       pointer;
      typedef size_t            size_type;
#ifndef __CINT__
      typedef const value_type& const_reference;
#endif

      typedef GObjArrayIterator<T>      iterator;
      typedef ConstGObjArrayIterator<T> const_iterator;
  
      //typedef reverse_bidirectional_iterator<const_iterator, value_type,
      //                                       const_reference, difference_type>
      //    const_reverse_iterator;
      
      //typedef reverse_bidirectional_iterator<iterator, value_type, reference,
      //                                       difference_type>
      //    reverse_iterator;
      
    public:
      /** @name --- Constructors --- */
      ///
      GClonesArray();
      ///
      GClonesArray( unsigned int newsize );
      ///
      GClonesArray( const GClonesArray<T>& rhs );
      ///
      ~GClonesArray();
      ///
      GClonesArray<T>& operator=( const GClonesArray<T>& rhs );
      
      /* THIS IS EVIL */
      /* remove it as soon as possible! */
      //operator GClonesArray<RecoTrack*>&()
      //{ return *(GClonesArray<RecoTrack*>*)this; }
      
      /** @name --- Access methods --- */
      ///
      iterator begin() {
	int i=0; while( i<GetLast()+1 && UncheckedAt(i)==0 ) ++i; 
	return iterator( *this, i );
      }
      ///
      iterator end()  {
	iterator i( *this, GetLast() );
	return ++i;
      }
      ///
      const_iterator begin() const {
	int i=0; while( i<GetLast()+1 && UncheckedAt(i)==0 ) ++i; 
	return const_iterator( *this, i );
      }
      ///
      const_iterator end() const {
	const_iterator i( *this, GetLast() );
	return ++i;
      }
      
      ///
      inline size_type size() const { return GetEntries(); }
      ///
      inline bool empty() const { return size()==0; }
      
      //void insert( value_type obj ) { Add( obj ); }
      ///
      T push_back( T obj );
      ///
      inline void pop_back() { RemoveAt( GetLast() ); }
      ///
      inline void erase( iterator position ) { RemoveAt( position.fCursor ); }
      ///
      void erase( iterator p1, iterator p2 ) { 
	iterator i=p2; while( i!=p1 ) --i; RemoveAt( i.fCursor );
      }

      //#if !defined(__CINT__) && !defined(CINTOBJECT)
      //T& operator[]( int idx )
      //{ return reinterpret_cast<T&>( TClonesArray::operator[]( idx ) ); }
      //#endif
      
      ///
      inline T& GetAt( int idx )
	{ return reinterpret_cast<T&>( TClonesArray::operator[](idx)); }
      
      //   reference GetAt( int idx )
      //     { return static_cast<reference>( *TClonesArray::operator[](idx)); }
      
      ///
      inline const value_type operator[]( int idx ) const
	{ return static_cast<value_type>
	    ( const_cast<GClonesArray*>(this)->TClonesArray::operator[](idx)); }
      
      //   reference operator[]( int idx )
      //     { return static_cast<reference>(*TClonesArray::operator[](idx)); }
      
      ///
      void set( int idx, T v );// { *((*this)[idx]) = *v; }
      
#if 0
      reference operator[]( int idx )
	{ return GetAt(idx); }
#endif
      
      /// used by operator<<()
      void print( std::ostream& os ) const;
      
      ///
      void saveObjects( GClonesArray<T>& newArray ) const;
      ///
      void copyObjectsFrom( const GObjArray<T>& oldArray );
      
      ///
      inline void Print( Option_t* ="" ) { print( std::cout ); }
      
      ClassDefT(GClonesArray<T>,1)
  }; // end of class GClonesArray


  //==============================================================================
  // operator <<
  //==============================================================================
  template<class T>
    inline std::ostream& operator<<( std::ostream& os, const GClonesArray<T>& ar ) {
    ar.print( os );
    return os;
  }

#ifdef NAMESPACE
} // namespace CLUE

MyClassDefT2(CLUE,GClonesArray,T)
  ; // for doc++
#else
ClassDefT2(GClonesArray,T)
  ; // for doc++
#endif

#else

#include <vector>
#define GClonesArray std::vector

#endif  /* USE_ROOT */

#endif  /* G_CLONES_ARRAY_HH */
