#ifndef _OBJ_ARRAY_HH_
#define _OBJ_ARRAY_HH_
// ********************************************************************
//
// source:
//
// type:      source code
//
// created:   03. Sep 1997
//
// author:    Thorsten Glebe
//            HERA-B Collaboration
//            Max-Planck-Institut fuer Kernphysik
//            Saupfercheckweg 1
//            69117 Heidelberg
//            Germany
//            E-mail: T.Glebe@mpi-hd.mpg.de
//
// Description: Template wrapper for TObjArray class
//
// changes:
// 03 Sep 1997 creation by C. Borgmeier
// 13 Nov 2000 (TG) added doc++ comments
// 14 Nov 2000 (TG) declared some member functions inline
// 20 Nov 2000 (TG) added namespace CLUE
// 05 Dez 2000 (TG) added operator<< for ConstGObjArrayIterator
// 05 Jan 2001 (TG) added operator<() in [Const]GObjArrayIterator
//
// ********************************************************************

#ifdef USE_ROOT

#include <iterator>
#include <iosfwd>
#include <TObjArray.h>


#ifdef NAMESPACE
#include "gate/RtypesT.h"

namespace CLUE {
#endif

  template <class T> class GObjArray;
  template <class T> class GObjArrayIterator;
  template <class T> class ConstGObjArrayIterator;
  template <class T> class GClonesArray;


  // template<class T> TBuffer& operator>>( TBuffer& buf, GObjArray<T>*& obj );
  
  /** GObjArray is a template wrapper for the ROOT TObjArray class.
      This class is only available in ROOT. Outside ROOT, GObjarray is typedef'd
      to a STL vector.
      @memo Array of different objects.
      @author C. Borgmeier
  */
  //==============================================================================
  // GObjArray
  //==============================================================================
  template<class T>
    class GObjArray: public TObjArray {
    public:
      typedef T                 value_type;
      typedef value_type&       reference;
      typedef const value_type& const_reference;
      typedef size_t            size_type;
      
    public:
      typedef GObjArrayIterator<T>      iterator;
      typedef ConstGObjArrayIterator<T> const_iterator;
      
      //typedef reverse_bidirectional_iterator<const_iterator, value_type,
      //                                       const_reference, difference_type>
      //    const_reverse_iterator;
      
      //typedef reverse_bidirectional_iterator<iterator, value_type, reference,
      //                                       difference_type>
      //    reverse_iterator;

    public:
      /** @name --- Constructors --- */
      ///
      GObjArray();
      ///
      GObjArray( const GObjArray<T>& rhs );
      ///
      ~GObjArray();
      ///
      GObjArray<T>& operator=( const GObjArray<T>& rhs );

      /** @name --- Access methods --- */
      ///
      iterator begin() {
	int i=0; while( i<GetLast()+1 && UncheckedAt(i)==0 ) ++i; 
	return iterator( *this, i );
      }
      ///
      inline iterator end() { return iterator( *this, GetLast()+1 ); }
      
      // #ifndef __CINT__
      // CINT seems not to be able to resolve this.
      ///
      const_iterator begin() const {
	int i=0; while( i<GetLast()+1 && UncheckedAt(i)==0 ) ++i; 
	return const_iterator( *this, i );
      }
      ///
      inline const_iterator end() const { return const_iterator( *this, GetLast()+1 ); }
      // #endif

      ///
      inline size_type size() const { return GetEntries(); }
      ///
      inline bool empty() const { return size()==0; }
      
      //void insert( value_type obj ) { Add( obj ); }
      ///
      inline void push_back( value_type obj ) { Add( (TObject*)obj ); }
      ///
      inline void pop_back() { RemoveAt( GetLast() ); }
      ///
      inline void erase( iterator position ) { RemoveAt( position.fCursor ); }
      ///
      void erase( iterator p1, iterator p2 )
	{ iterator i=p2; while( i!=p1 ) --i; RemoveAt( i.fCursor ); }
      ///
      inline void clear() { Clear(); }
      
      //value_type operator[]( int idx )
      //{ return static_cast<value_type>( At(idx) ); }
      
      ///
      inline const value_type operator[]( int idx ) const
	{ return static_cast<value_type>( At(idx) ); }
      ///
      inline value_type GetAt( int idx )
	{ return static_cast<value_type>( At(idx) ); }
      ///
      inline const value_type GetAt( int idx ) const
	{ return static_cast<value_type>
	    ( const_cast<GObjArray<T>*>(this)->At(idx) ); }

      // for ROOT compatibility
      //TObject operator[]( int idx ) const
      //{ return const_cast<GObjArray<T>*>(this)->GetAt(idx); }
      
      ///
      inline void set( int idx, value_type v ) { AddAt( (TObject*)v, idx ); }
      
#if 0
      reference operator[]( int idx )
	{ return GetAt(idx); }
#endif

      /// used by operator<<()
      void print( std::ostream& os ) const;

      ///
      void Print( Option_t* ="" );
      ///
      void saveObjects( GObjArray<T>& newArray ) const;
  
      ClassDefT(GObjArray<T>,1)
  }; // end of class GObjArray


  //==============================================================================
  // operator <<
  //==============================================================================
  template<class T>
    inline std::ostream& operator<<( std::ostream& os, const GObjArray<T>& ar ) {
    ar.print( os );
    return os;
  }

  
  /** @memo GObjectArray Iterator class 
      @author C. Borgmeier, T. Glebe
  */
  //==============================================================================
  // GObjArrayIterator
  //==============================================================================
  template<class T>
    class GObjArrayIterator: public TObject {
      //      friend class vecplusit;
    public:
#ifdef __CINT__
      typedef int ptrdiff_t
#endif
      //  typedef GObjArrayIterator<T> self;
#ifndef __CINT__
      typedef T                    value_type;
      typedef value_type*          pointer;
      typedef value_type&          reference;
      typedef const value_type&    const_reference;
      typedef forward_iterator_tag iterator_category;
#endif
      typedef ptrdiff_t            difference_type;

      /** @name --- Constructors --- */
      ///
      GObjArrayIterator() {}
      ///
      GObjArrayIterator( TObjArray& array, Int_t idx )
	: fArray( &array ), fCursor( idx ) {}
      ///
      GObjArrayIterator( const GObjArrayIterator<T>& rhs )
	: fArray( rhs.fArray ), fCursor( rhs.fCursor ) {}
      ///
      const GObjArrayIterator<T>& operator=( const GObjArrayIterator<T>& rhs )
	{ fArray = rhs.fArray; fCursor = rhs.fCursor; return *this; }

      /** @name --- Operators --- */
      ///
      T operator*()
	{ return static_cast<T>(fArray->UncheckedAt( fCursor )); }
      ///
      const T operator*() const
	{ return //const_cast<const T>
	    (static_cast<T>( fArray->UncheckedAt( fCursor ))); }
      ///
      GObjArrayIterator<T>& operator++() {
	increment();
	return *this;
      }
      ///
      GObjArrayIterator<T> operator++(int) {
	GObjArrayIterator<T> tmp = *this;
	increment();
	return tmp;
      }
      ///
      GObjArrayIterator<T>& operator--() {
	decrement();
	return *this;
      }
      ///
      GObjArrayIterator<T> operator--(int) {
	GObjArrayIterator<T> tmp = *this;
	decrement();
	return tmp;
      }
      ///
      GObjArrayIterator<T> operator+(int i) const {
	GObjArrayIterator<T> tmp = *this;
	while( i-->0 ) ++tmp;
	return tmp;
      }

      ///
      inline bool operator!=( const ConstGObjArrayIterator<T>& rhs ) const
	{ return fCursor!=rhs.fCursor || fArray!=rhs.fArray; }
      ///
      inline bool operator==( const GObjArrayIterator<T>& rhs ) const
	{ return fCursor==rhs.fCursor && fArray==rhs.fArray; }
      ///
      inline bool operator<( const GObjArrayIterator<T>& rhs ) const
	{ return fCursor<rhs.fCursor && fArray==rhs.fArray; }

      /** @name --- Expert functions --- */
      ///
      void Print( Option_t* o="" );

      ///
      void print( std::ostream& os ) const
	{ os << "[" << fArray << "," << fCursor << "]"; }

    protected:
      ///
      void increment() {
	if( fCursor==fArray->GetLast() ) fCursor=fArray->GetLast()+1;
	else while( fCursor < fArray->GetLast()+1
		    && fArray->UncheckedAt(++fCursor) == 0 )
	  { }
      }
      ///
      inline void decrement() {
	while( fCursor > 0
	       && fArray->UncheckedAt(--fCursor) == 0 )
	  { }
      }
      
    public:
      TObjArray* fArray;
      Int_t fCursor;
      //protected:
      
      ClassDefT(GObjArrayIterator<T>,1)
  }; // end of class GObjArrayIterator


  //==============================================================================
  // operator <<
  //==============================================================================
  template<class T>
    inline std::ostream& operator<< (std::ostream& os,
				     const GObjArrayIterator<T>& it ) {
    it.print(os);
    return os;
  }
  
  // template<class T> TBuffer& operator>>(TBuffer&, GObjArrayIterator<T>* &);


  /** @memo ConstGObjectArray Iterator class
      @author C. Borgmeier, T. Glebe
  */
  //==============================================================================
  // ConstGObjArrayIterator
  //==============================================================================
  template<class T>
    class ConstGObjArrayIterator : public TObject {
      friend class GObjArrayIterator<T>;
    public:
#ifdef __CINT__
      typedef int ptrdiff_t
#endif
      //  typedef ConstGObjArrayIterator<T> self;
#ifndef __CINT__
      typedef T                         value_type;
      typedef value_type*               pointer;
      typedef value_type&               reference;
      typedef const value_type&         const_reference;
      typedef forward_iterator_tag      iterator_category;
#endif
      typedef ptrdiff_t                 difference_type;

      /** @name --- Constructors --- */
      ///
      ConstGObjArrayIterator() {}
      ///
      ConstGObjArrayIterator( const ConstGObjArrayIterator<T>& rhs )
	: fArray( rhs.fArray ), fCursor( rhs.fCursor ) {}
      ///
      ConstGObjArrayIterator( const GObjArrayIterator<T>& rhs )
	: fArray( rhs.fArray ), fCursor( rhs.fCursor ) {}
      /// this would better be protected.
      ConstGObjArrayIterator( const TObjArray& array, Int_t idx )
	: fArray( const_cast<TObjArray*>(&array) ), fCursor( idx ) {}


      /** @name --- Operators --- */
      ///
      const ConstGObjArrayIterator<T>& operator=
      ( const ConstGObjArrayIterator<T>& rhs )
	{ fArray = rhs.fArray; fCursor = rhs.fCursor; return *this; }
      ///
      const ConstGObjArrayIterator<T>& operator=
      ( const GObjArrayIterator<T>& rhs )
	{ fArray = rhs.fArray; fCursor = rhs.fCursor; return *this; }
      ///
      T operator*()
	{ return static_cast<T>(fArray->UncheckedAt( fCursor )); }
      ///
      const T operator*() const
	{ return static_cast<T>(fArray->UncheckedAt( fCursor )); }
      ///
      ConstGObjArrayIterator<T>& operator++() {
	increment();
	return *this;
      }
      ///
      ConstGObjArrayIterator<T> operator++(int) {
	ConstGObjArrayIterator<T> tmp = *this;
	increment();
	return tmp;
      }
      ///
      ConstGObjArrayIterator<T>& operator--() {
	decrement();
	return *this;
      }
      ///
      ConstGObjArrayIterator<T> operator--(int) {
	ConstGObjArrayIterator<T> tmp = *this;
	decrement();
	return tmp;
      }
      ///
      ConstGObjArrayIterator<T> operator+(int i) const {
	ConstGObjArrayIterator<T> tmp = *this;
	while( i-->0 ) ++tmp;
	return tmp;
      }
      ///
      int operator!=( const ConstGObjArrayIterator<T>& rhs ) const
	{ return fCursor!=rhs.fCursor || fArray!=rhs.fArray; }
      ///
      int operator==( const ConstGObjArrayIterator<T>& rhs ) const
	{ return fCursor==rhs.fCursor && fArray==rhs.fArray; }
      ///
      inline bool operator<( const GObjArrayIterator<T>& rhs ) const
	{ return fCursor<rhs.fCursor && fArray==rhs.fArray; }
      
      ///
      void print( std::ostream& os ) const
	{ os << "[" << fArray << "," << fCursor << "]"; }

    protected:
      ///
      inline void increment() {
	while( fCursor!=fArray->GetLast()+1
	       && fArray->UncheckedAt(++fCursor) == 0 )
	  { }
      }

      ///
      inline void decrement() {
	while( fCursor > 0
	       && fArray->UncheckedAt(--fCursor) == 0 )
	  { }
      }
      
      TObjArray* fArray;
      Int_t fCursor;

      ClassDefT(ConstGObjArrayIterator<T>,1)
  }; // end of class ConstGObjArrayIterator


  //==============================================================================
  // operator <<
  //==============================================================================
  template<class T>
    inline std::ostream& operator<< (std::ostream& os,
				     const ConstGObjArrayIterator<T>& it ) {
    it.print(os);
    return os;
  }
  

#ifdef NAMESPACE
} // namespace CLUE

MyClassDefT2(CLUE,GObjArray,T)
  ; // for doc++
MyClassDefT2(CLUE,GObjArrayIterator,T)
  ; // for doc++
MyClassDefT2(CLUE,ConstGObjArrayIterator,T)
  ; // for doc++
#else
ClassDefT2(GObjArray,T)
  ; // for doc++
ClassDefT2(GObjArrayIterator,T)
  ; // for doc++
ClassDefT2(ConstGObjArrayIterator,T)
  ; // for doc++
#endif


  // template<class T> TBuffer& operator>>(TBuffer&, ConstGObjArrayIterator<T>* &);
#else  // USE_ROOT
#include <vector>

#define GObjArray              std::vector
#endif // USE_ROOT

#ifndef __CINT__
#include "gate/ObjArray.icc"
#endif
#endif
