#ifndef __GC_TRAIT_HH
#define __GC_TRAIT_HH
// *****************************************************************************
//
// source:
//
// type:      source code
//
// created:   04. Jan 2001
//
// author:    Thorsten Glebe
//            HERA-B Collaboration
//            Max-Planck-Institut fuer Kernphysik
//            Saupfercheckweg 1
//            69117 Heidelberg
//            Germany
//            E-mail: T.Glebe@mpi-hd.mpg.de
//
// Description: trait to configure GCombiner
//
// changes:
// 22 Jan 2001 (TG) moved GC_trait from GCombiner.hh to here
//
// *****************************************************************************

template <class Iterator> class IteratorPair;

/** Combiner trait.
    This trait class carries the definitions of containers and corresponding
    iterators which are used by the \texttt{GCombiner} and
    \texttt{GCombinerIterator} classes. For each kind of container \texttt{T}
    the internally used containers can be reconfigured by specializing this
    trait structure.
    The \texttt{GCombiner} class makes the following assumptions about the
    containers which should be combined:
    \begin{enumerate}
    \item all containers \texttt{T} must be of same type (e.g. vector<int>)
    \item container has a \texttt{T::iterator} typedef (e.g. vector<int>::iterator)
    \item container has a \texttt{begin()} method (e.g. vector<int>::begin())
    \item container has an \texttt{end()} method (e.g. vector<int>::end())
    \item \texttt{T::iterator} has an \texttt{operator*()} defined
    \item \texttt{T::iterator} has an \texttt{operator==()} defined
    \item \texttt{T::iterator} has an \texttt{operator!=()} defined
    \item \texttt{T::iterator} has an \texttt{operator++()} defined
    \end{enumerate}
    The above assumptions are fulfilled for all \texttt{STL} compatible containers.
    @memo Combiner
    @author T. Glebe
*/ 
//============================================================================
// The GC-Trait
//============================================================================
template <class T>
struct GC_trait {
  ///
  typedef T                                            Gvalue_type;
  ///
  typedef T*                                           Gpointer;
  ///
  typedef T* const                                     Gconst_pointer;
  ///
  typedef T&                                           Greference;
  ///
  typedef const T&                                     Gconst_reference;
  ///
  typedef std::list<Gpointer>                          GCombiner_v;
  ///
  typedef std::list<Gpointer>::iterator                GCombiner_it;
  ///
  typedef std::list<Gpointer>::const_iterator          GCombiner_cit;
  ///
  typedef std::list<Gpointer>::reverse_iterator        GCombiner_rit;
  ///
  typedef std::list<Gpointer>::const_reverse_iterator  GCombiner_rcit;

  ///
  typedef typename T::iterator                         GIterator;
  ///
  typedef IteratorPair<GIterator>                      ItPair;
  ///
  typedef std::list<ItPair>                            GIterator_v;
  ///
  typedef std::list<ItPair>::iterator                  GIterator_it;
  ///
  typedef std::list<ItPair>::const_iterator            GIterator_cit;
  ///
  typedef std::list<ItPair>::reverse_iterator          GIterator_rit;
  ///
  typedef std::list<ItPair>::const_reverse_iterator    GIterator_rcit;

  ///
  typedef std::vector<GIterator>                         GCombination;
  ///
  typedef std::vector<GIterator>::iterator               GCombination_it;
  ///
  typedef std::vector<GIterator>::const_iterator         GCombination_cit;
  ///
  typedef std::vector<GIterator>::reverse_iterator       GCombination_rit;
  ///
  typedef std::vector<GIterator>::const_reverse_iterator GCombination_rcit;
};

#endif
