#ifndef __SCOMBINER_HH
#define __SCOMBINER_HH
// *****************************************************************************
//
// source:
//
// type:      source code
//
// created:   04. Jan 2001
//
// author:    Thorsten Glebe
//            HERA-B Collaboration
//            Max-Planck-Institut fuer Kernphysik
//            Saupfercheckweg 1
//            69117 Heidelberg
//            Germany
//            E-mail: T.Glebe@mpi-hd.mpg.de
//
// Description: template for a single container combiner
//
// changes:
// 22 Jan 2001 (TG) creation
// 23 Jan 2001 (TG) operator<< problem with second template argument, therefore
//                  hardcoded operator<< functions for n=2,3,4.
// 07 Feb 2001 (TG) added #include GFaculty; added ncomb() member function
// 09 Jul 2001 (TG) SCominerIterator::iterator, ::const_iterator added
// 09 Jul 2001 (TG) non-const begin(), end() added
// 25 Jul 2001 (TG) added valid()
// 28 Aug 2001 (TG) bug fix in template declarations, commented some typedefs
//                  for CINT to make it parsed by ROOT versions >= 3.00
//
// *****************************************************************************
#include <iosfwd>
#include <vector>
#include <assert.h>
#include "GC_trait.hh"

#ifndef __CINT__
#include "gcombiner/GFaculty.hh"
#endif

// the declarations are wrong, but they make
// that CINT instantiates correctly! (?)
template <class T, unsigned int X> class SCombiner;
template <class T, unsigned int X> class SCombinerIterator;

/**
   Combiner for a single container.
   SCombiner is a combiner which combines $n$ objects from one
   container. The first template argument must be the type of a STL
   compatible container. $n$ specifies the number of objects to
   combiner, it must be given as the second template
   argument. With SCombiner, each combination will occur only once.
   In a combination, the same object will not appear more than once.

   It is assumed that the container \texttt{T} has the following
   properties:
   \begin{enumerate}
   \item a \texttt{T::begin()} and \texttt{T::end()} method
   \item a \texttt{T::iterator} typedef
   \item a \texttt{T::size()} method
   \end{enumerate}
   @memo Combiner
   @author T. Glebe
*/
//============================================================================
// SCombiner
//============================================================================
template <class T, unsigned int X>
class SCombiner {

#ifndef __CINT__
  typedef typename GC_trait<T>::Greference         Greference;
  typedef typename GC_trait<T>::GIterator          GIterator;
#else
  typedef GC_trait<T>::Greference                  Greference;
  typedef GC_trait<T>::GIterator                   GIterator;
#endif

public:
  /** @name --- Typedef's --- */
  /// not available in CINT
  typedef SCombinerIterator<T,X>             iterator;

  /** @name --- Constructors --- */
  ///
  SCombiner() : container(0) {}
  ///
  SCombiner(Greference rhs): 
    container(&rhs) {}
  
  /** @name --- Iterator methods --- */
  ///
  inline     iterator begin() const  { return iterator(*container, false); }
  ///
  inline     iterator end()   const  { return iterator(*container, true); }

  /** @name --- Access methods --- */
  /// size of container
  inline unsigned int size() const { return container->size(); }
  /// set pointer to container
  inline void set_container(Greference rhs) { container = &rhs; }

  /** @name --- Expert functions --- */
  ///
  std::ostream& print(std::ostream& os) const {
    const GIterator end = container->end();
    for(GIterator it = container->begin(); it != end; ++it)
      os << *it << " ";
    return os;
  }

  /// no of combinations $\frac{n!}{(n-k)! * k!}$
  inline unsigned long int ncomb() const {
    return noverk(size(),size()-X) / Faculty<X>::result;
  }

  /// check if combiner is valid: true if size() >= X
  inline bool valid() const { return container->size() >= X; }

private:
  T* container;
}; // end of SCombiner

//==============================================================================
// operator<<: the second template argument is a problem, therefore hardcoded!
//==============================================================================
template <class T>
inline std::ostream& operator<<(std::ostream& s, const SCombiner<T,2>& a) {
  return a.print(s);
}
template <class T>
inline std::ostream& operator<<(std::ostream& s, const SCombiner<T,3>& a) {
  return a.print(s);
}
template <class T>
inline std::ostream& operator<<(std::ostream& s, const SCombiner<T,4>& a) {
  return a.print(s);
}


/**
   Iterator for SCombiner.
   The \texttt{SCombinerIterator} iterates over all combinations which are
   possible by the given container object by excluding that the same
   object occurs more than once in a combination. The \texttt{SCombinerIterator}
   represents a vector of iterators pointing to the objects of the
   current combination. Therefore this iterator behaves very different
   from a STL iterator and is not STL conform! The
   \texttt{SCombinerIterator} itself has a \texttt{begin()} 
   and \texttt{end()} method to loop over the elements of the current
   combination. An \texttt{operator*()} for receiving the vector of container
   iterators and an \texttt{operator[]()} method for direct access to the
   container iterators is provided.
   
   @memo Iterates the combinations.
   @author T. Glebe
 */
//============================================================================
// SCombinerIterator
//============================================================================
template <class T, unsigned int X>
class SCombinerIterator {

  typedef SCombinerIterator<T,X>                   self;
#ifndef __CINT__
  typedef typename GC_trait<T>::Greference         Greference;
  typedef typename GC_trait<T>::GCombination       GCombination;
  typedef typename GC_trait<T>::GCombination_it    GCombination_it;
  typedef typename GC_trait<T>::GCombination_cit   GCombination_cit;
  typedef typename GC_trait<T>::GIterator          GIterator;
#else
  typedef GC_trait<T>::Greference         Greference;
  typedef GC_trait<T>::GCombination       GCombination;
  typedef GC_trait<T>::GCombination_it    GCombination_it;
  typedef GC_trait<T>::GCombination_cit   GCombination_cit;
  typedef GC_trait<T>::GIterator          GIterator;
#endif

 public:
  ///
  typedef GCombination_it  iterator;
  ///
  typedef GCombination_cit const_iterator;

  /** @name --- Constructors --- */
  ///
  //  SCombinerIterator() {}
  ///
  SCombinerIterator(Greference rhs, bool isend = false) :
    comb(GCombination(X)),
    cend(GCombination(X)) {

    assert(X > 0 && rhs.size() >= X);
    
    GIterator bit = rhs.begin();
    GIterator eit = rhs.end();
    unsigned int i=X-1;
    do {
      //      cout << "i: " << i << endl;
      cend[i] = eit;

      if(isend) {
	comb[i] = eit;
      } else {
	comb[X-i-1] = bit; ++bit; 
      }
      
      --eit;
    } while(i-->0);
  }

  ///
  virtual ~SCombinerIterator() {}

  /** @name --- Access methods --- */
  ///
  inline GCombination_it begin() { return comb.begin(); }
  ///
  inline GCombination_it end()   { return comb.end(); }
  ///
  inline GCombination_cit begin() const { return comb.begin(); }
  ///
  inline GCombination_cit end()   const { return comb.end(); }
  ///
  inline int size() const { return X; }

  /** @name --- Operators --- */
  ///
  inline self&  operator++() { iterate(); return *this; }
  ///
  inline bool operator==(const self& rhs) const 
    { return comb == rhs.comb; }
  ///
  inline bool operator!=(const self& rhs) const
    { return comb != rhs.comb; }
  ///
  inline const GCombination&  operator*() const { return comb; }
  ///
  inline const GIterator& operator[](unsigned int i) const
    { return comb[i]; }

  /** @name --- Expert functions --- */
  ///
  std::ostream& print(std::ostream& os) const {
    const GCombination_cit end = comb.end();
    for(GCombination_cit it = comb.begin(); it != end; ++it) {
      os << **it << " ";
    }
    return os;
  }

protected:
  /// algorithm for an iteration step
  virtual void iterate() {

    unsigned int i=X;
    do {
      --i;
      ++comb[i]; // save because of assert()
    } while(comb[i]==cend[i] && i>0);

    for(unsigned int j = i+1; j<X; ++j) {
      comb[j] = comb[j-1]; ++comb[j];
    }
  }

private:
  GCombination comb; // vector of iterators
  GCombination cend; // vector of end-iterators
}; // end of SCombinerIterator


//==============================================================================
// operator<<: the second template argument is a problem, therefore hardcoded
//==============================================================================
template <class T>
inline std::ostream& operator<<(std::ostream& s, const SCombinerIterator<T,2>& a) {
  return a.print(s);
}
template <class T>
inline std::ostream& operator<<(std::ostream& s, const SCombinerIterator<T,3>& a) {
  return a.print(s);
}
template <class T>
inline std::ostream& operator<<(std::ostream& s, const SCombinerIterator<T,4>& a) {
  return a.print(s);
}

#endif
