#ifndef __SDCOMBINER_HH
#define __SDCOMBINER_HH
// *****************************************************************************
//
// source:
//
// type:      source code
//
// created:   05. Jun 2001
//
// author:    Thorsten Glebe
//            HERA-B Collaboration
//            Max-Planck-Institut fuer Kernphysik
//            Saupfercheckweg 1
//            69117 Heidelberg
//            Germany
//            E-mail: T.Glebe@mpi-hd.mpg.de
//
// Description: template for a single container combiner, but with dynamic
//              combinatoric depth
//
// changes:
// 05 Jun 2001 (TG) creation from SCombiner class
// 02 Jul 2001 (TG) void* - CTOR added
// 09 Jul 2001 (TG) GCominerIterator::iterator, ::const_iterator added
// 09 Jul 2001 (TG) non-const begin(), end() added
//
// *****************************************************************************
#include <iosfwd>
#include <vector>
#include <assert.h>
#include "GC_trait.hh"

#ifndef __CINT__
#include "gcombiner/GFaculty.hh"
#endif

template <class T> class SDCombinerIterator;

/**
   Combiner for a single container.
   SDCombiner is a combiner which combines $n$ objects from one
   container. The template argument must be the type of a STL
   compatible container. $n$ specifies the number of objects to
   combiner, it must be given as a constructor argument.
   With SCombiner, each combination will occur only once.
   Within a combination, the same object will not appear more than once.

   It is assumed that the container \texttt{T} has the following
   properties:
   \begin{enumerate}
   \item a \texttt{T::begin()} and \texttt{T::end()} method
   \item a \texttt{T::iterator} typedef
   \item a \texttt{T::size()} method
   \end{enumerate}
   @memo Combiner
   @author T. Glebe
*/
//============================================================================
// SDCombiner
//============================================================================
template <class T>
class SDCombiner {

#ifndef __CINT__
  typedef typename GC_trait<T>::Greference         Greference;
  typedef typename GC_trait<T>::GIterator          GIterator;
#else
  typedef GC_trait<T>::Greference                  Greference;
  typedef GC_trait<T>::GIterator                   GIterator;
#endif

public:
  /** @name --- Typedef's --- */
  ///
  typedef SDCombinerIterator<T>             iterator;

  /** @name --- Constructors --- */
  ///
  SDCombiner() : container(0), X(0) {}
  ///
  SDCombiner(Greference rhs, unsigned int N): 
    container(&rhs), X(N) {}
  
  /** @name --- Iterator methods --- */
  ///
  inline     iterator begin() const  { return iterator(*container, X, false); }
  ///
  inline     iterator end()   const  { return iterator(*container, X, true); }

  /** @name --- Access methods --- */
  /// size of container
  inline unsigned int size() const { return container->size(); }
  /// set pointer to container
  inline void set_container(Greference rhs) { container = &rhs; }
  /// return pointer to Container
  inline T* pointer() const { return container; }
  /// combinatorics depth
  inline unsigned int N() const { return X; }
  /// set combinatorics depth
  inline void set_N(const unsigned int n) { X = n; }
  /// increment combinatorics depth
  inline void incr_N() { ++X; }

  /** @name --- Expert functions --- */
  ///
  std::ostream& print(std::ostream& os) const {
    os << "Combining " << X << "-tuple, elements: ";
    const GIterator end = container->end();
    for(GIterator it = container->begin(); it != end; ++it)
      os << *it << " ";
    return os;
  }

  /// no of combinations $\frac{n!}{(n-k)! * k!}$
  inline unsigned long int ncomb() const {
    return noverk(size(),size()-X) / faculty(X);
  }

private:
  T* container;
  unsigned int X; // combinatorics depth
}; // end of SCombiner

//==============================================================================
// operator<<: the second template argument is a problem, therefore hardcoded!
//==============================================================================
template <class T>
inline std::ostream& operator<<(std::ostream& s, const SDCombiner<T>& a) {
  return a.print(s);
}


/**
   Iterator for SDCombiner.
   The \texttt{SDCombinerIterator} iterates over all combinations which are
   possible by the given container object by excluding that the same
   object occurs more than once in a combination. The \texttt{SCombinerIterator}
   represents a vector of iterators pointing to the objects of the
   current combination. Therefore this iterator behaves very different
   from a STL iterator and is not STL conform! The
   \texttt{SCombinerIterator} itself has a \texttt{begin()} 
   and \texttt{end()} method to loop over the elements of the current
   combination. An \texttt{operator*()} for receiving the vector of container
   iterators and an \texttt{operator[]()} method for direct access to the
   container iterators is provided.
   
   @memo Iterates the combinations.
   @author T. Glebe
 */
//============================================================================
// SDCombinerIterator
//============================================================================
template <class T>
class SDCombinerIterator {

  typedef SDCombinerIterator<T> self;
#ifndef __CINT__
  typedef typename GC_trait<T>::Greference         Greference;
  typedef typename GC_trait<T>::GCombination       GCombination;
  typedef typename GC_trait<T>::GCombination_it    GCombination_it;
  typedef typename GC_trait<T>::GCombination_cit   GCombination_cit;
  typedef typename GC_trait<T>::GIterator          GIterator;
#else
  typedef GC_trait<T>::Greference         Greference;
  typedef GC_trait<T>::GCombination       GCombination;
  typedef GC_trait<T>::GCombination_it    GCombination_it;
  typedef GC_trait<T>::GCombination_cit   GCombination_cit;
  typedef GC_trait<T>::GIterator          GIterator;
#endif

 public:
  ///
  typedef GCombination_it  iterator;
  ///
  typedef GCombination_cit const_iterator;

  /** @name --- Constructors --- */
  ///
  //  SCombinerIterator() {}
  ///
  SDCombinerIterator(Greference rhs, unsigned int X, bool isend = false) :
    comb(GCombination(X)),
    cend(GCombination(X)) {

    //<< " rhs.size: " << rhs.size() << endl;
    assert(X > 0 && rhs.size() >= X);
    
    GIterator bit = rhs.begin();
    GIterator eit = rhs.end();
    unsigned int i=X-1;
    do {
      //      cout << "i: " << i << endl;
      cend[i] = eit;

      if(isend) {
	comb[i] = eit;
      } else {
	comb[X-i-1] = bit; ++bit; 
      }
      
      --eit;
    } while(i-->0);
  }

  ///
  virtual ~SDCombinerIterator() {}

  /** @name --- Access methods --- */
  ///
  inline GCombination_it begin() { return comb.begin(); }
  ///
  inline GCombination_it end()   { return comb.end(); }
  ///
  inline GCombination_cit begin() const { return comb.begin(); }
  ///
  inline GCombination_cit end()   const { return comb.end(); }
  ///
  inline int size() const { return comb.size(); }

  /** @name --- Operators --- */
  ///
  inline self&  operator++() { iterate(); return *this; }
  ///
  inline const GCombination&  operator*() const { return comb; }
  ///
  inline const GIterator& operator[](unsigned int i) const
    { return comb[i]; }
  ///
  inline bool operator==(const SDCombinerIterator<T>& rhs) const 
    { return comb == rhs.comb; }
  ///
  inline bool operator!=(const SDCombinerIterator<T>& rhs) const
    { return comb != rhs.comb; }

  /** @name --- Expert functions --- */
  ///
  std::ostream& print(std::ostream& os) const {
    const GCombination_cit end = comb.end();
    for(GCombination_cit it = comb.begin(); it != end; ++it) {
      os << **it << " ";
    }
    return os;
  }

protected:
  /// algorithm for an iteration step
  virtual void iterate() {

    const unsigned int X = comb.size();
    unsigned int i=X;
    do {
      --i;
      ++comb[i]; // save because of assert()
    } while(comb[i]==cend[i] && i>0);

    for(unsigned int j = i+1; j<X; ++j) {
      comb[j] = comb[j-1]; ++comb[j];
    }
  }

private:
  GCombination comb; // vector of iterators
  GCombination cend; // vector of end-iterators
}; // end of SCombinerIterator


//==============================================================================
// operator<<
//==============================================================================
template <class T>
inline std::ostream& operator<<(std::ostream& s, const SDCombinerIterator<T>& a) {
  return a.print(s);
}

#endif
