#ifndef __RLORENTZ_VECTOR_H__
#define __RLORENTZ_VECTOR_H__
// ********************************************************************
//
// source:
//
// type:      source code
//
// created:   15. Jun 2000
//
// author:    Thorsten Glebe
//            HERA-B Collaboration
//            Max-Planck-Institut fuer Kernphysik
//            Saupfercheckweg 1
//            69117 Heidelberg
//            Germany
//            E-mail: T.Glebe@mpi-hd.mpg.de
//
// Description: LorentzVector interface
//
// changes:
// 15 Jun 2000 (TG) creation
// 13 Nov 2000 (TG) added doc++ comments
// 20 Nov 2000 (TG) added namespace CLUE
//
// ********************************************************************

#ifdef USE_ROOT  //--------> ROOT
#include "geometry/ThreeVector.hh"
#include <TLorentzVector.h>

#ifdef NAMESPACE
namespace CLUE {
#endif

  /**
     LorentzVector - ROOT based implementation for usage with ROOT.
     For documentation of TLorentzVector member functions see the ROOT reference
     manual.
     @memo ROOT based implementation
     @author T. Glebe
  */
  //==============================================================================
  // LorentzVector
  //==============================================================================
  class LorentzVector: public TLorentzVector {
  public:
    /** @name --- Constructors --- */
    ///
    LorentzVector() {}
    ///
    LorentzVector( const LorentzVector& rhs): TLorentzVector(rhs) {}
    ///
    LorentzVector( const TLorentzVector& rhs): TLorentzVector(rhs) {}
    ///
    LorentzVector( const ThreeVector& rhs, double E): TLorentzVector(rhs, E) {}
    ///
    LorentzVector( const double px, const double py, const double pz, const
		   double E ): TLorentzVector(px,py,pz,E) {}
    ///
    const LorentzVector& operator=( const LorentzVector& rhs ) {
      *static_cast<TLorentzVector*>(this) = rhs;
      return *this;
    }
    
    /** @name --- Access methods --- */
    /// get spatial component
    inline ThreeVector vect() const { return Vect(); }
    /// get 4th component (energy)
    inline double e() const { return Energy(); }
    /// Invariant mass. If mag2() is negative then -sqrt(-mag2()) is returned.
    inline double mag() const { return Mag(); }
    /// Invariant mass squared.
    inline double mag2() const { return Mag2(); }
    /// Returns the rapidity, i.e. $0.5\cdot\ln((E+p_z)/(E-p_z))$
    inline double rapidity() const { return Rapidity(); }
    /// Returns the pseudo-rapidity, i.e. $-\ln(\tan(\theta/2))$
    inline double pseudoRapidity() const { return PseudoRapidity(); }

    /** @name --- Operators --- */
    ///
    inline LorentzVector operator+ (const LorentzVector &rhs) const {
      return LorentzVector(vect() + rhs.vect(), e() + rhs.e()); 
    }
    ///
    inline LorentzVector operator- (const LorentzVector &rhs) const {
      return LorentzVector(vect() - rhs.vect(), e() - rhs.e()); 
    }

#ifdef USE_ROOT
    ClassDef(LorentzVector,1)
#endif
  }; // class LorentzVector

#ifdef NAMESPACE
} // namespace CLUE
#endif

#else         // USE_ROOT
#include <CLHEP/Vector/LorentzVector.h>
// typedef HepLorentzVector LorentzVector;

#ifdef NAMESPACE
namespace CLUE {
#endif
  /**
     LorentzVector - CLHEP based implementation for usage without ROOT.
     For documentation of HepLorentzVector member functions see CLHEP documentation
     @memo CLHEP implementation
     @author T. Glebe
  */
  //==============================================================================
  // LorentzVector
  //==============================================================================
  class LorentzVector: public HepLorentzVector {
  public:
    /** @name --- Constructors --- */
    ///
    LorentzVector() {}
    ///
    LorentzVector( const LorentzVector& rhs): HepLorentzVector(rhs) {}
    ///
    LorentzVector( const HepLorentzVector& rhs): HepLorentzVector(rhs) {}
    ///
    LorentzVector( const Hep3Vector& rhs, double E): HepLorentzVector(rhs, E) {}
    ///
    LorentzVector( const double px, const double py, const double pz, const
		   double E ): HepLorentzVector(px,py,pz,E) {}
    ///
    const LorentzVector& operator=( const LorentzVector& rhs ) {
      *static_cast<HepLorentzVector*>(this) = rhs;
      return *this;
    }
  }; // class LorentzVector

#ifdef NAMESPACE
} // namespace CLUE
#endif
#endif // USE_ROOT
#endif
