#ifndef __RTHREE_VECTOR_H__
#define __RTHREE_VECTOR_H__
// ********************************************************************
//
// source:
//
// type:      source code
//
// created:   15. Jun 2000
//
// author:    Thorsten Glebe
//            HERA-B Collaboration
//            Max-Planck-Institut fuer Kernphysik
//            Saupfercheckweg 1
//            69117 Heidelberg
//            Germany
//            E-mail: T.Glebe@mpi-hd.mpg.de
//
// Description: ThreeVector interface
//
// changes:
// 15 Jun 2000 (TG) creation
// 13 Nov 2000 (TG) added doc++ comments
// 20 Nov 2000 (TG) added namespace CLUE
//
// ********************************************************************

#include <iostream.h>
#include <iomanip>

#ifdef USE_ROOT
#include <TVector3.h>
// class TPolyMarker3D;

// typedef TVector3 ThreeVector;

#ifdef NAMESPACE
namespace CLUE {
#endif

  /**
     ThreeVector - ROOT based implementation for usage with ROOT.
     For documentation of TVector3 member functions see the ROOT reference
     manual.
     @memo ROOT based implementation
     @author T. Glebe
  */
  //==============================================================================
  // ThreeVector
  //==============================================================================
  class ThreeVector: public TVector3 {
  public:
    /** @name --- Constructors --- */
    ///
    ThreeVector() {}
    ///
    ThreeVector( const ThreeVector& rhs ): TVector3( rhs ) {}
    ///
    ThreeVector( const TVector3& rhs ): TVector3( rhs ) {}
    ///
    ThreeVector( const double x, const double y, const double z ): TVector3( x, y, z ) {}
    ///
    const ThreeVector& operator=( const ThreeVector& rhs ) {
      *static_cast<TVector3*>(this) = rhs;
      return *this;
    }

    /** @name --- Access methods --- */
    /// X() can be used, too
    inline double x() const { return X(); }
    /// Y() can be used, too
    inline double y() const { return Y(); }
    /// Z() can be used, too
    inline double z() const { return Z(); }
    
    /// set x-component in cartesian coordinates
    inline void set_x(const double x) { SetX(x); }
    /// set y-component in cartesian coordinates
    inline void set_y(const double y) { SetY(y); }
    /// set z-component in cartesian coordinates
    inline void set_z(const double z) { SetZ(z); }

    /// Unit vector parallel to this.  
    inline ThreeVector unit() const { return Unit(); }
    
    // The magnitude squared (rho^2 in spherical coordinate system).
    inline double mag2() const { return Mag2(); }
    
#ifdef USE_ROOT
    /// only available in ROOT
    virtual void Paint(Option_t *option="");
    
    ClassDef(ThreeVector,1)
#endif
  }; // end of class ThreeVector
#ifdef NAMESPACE
} // namespace CLUE
#endif
#else // USE_ROOT
#include <CLHEP/Vector/ThreeVector.h>
#include "arte/TObject.hh"

#ifdef NAMESPACE
namespace CLUE {
#endif

  /**
     ThreeVector - CLHEP based implementation for usage outside ROOT.
     For documentation of Hep3Vector member functions see CLHEP documentation.
     @memo CLHEP implementation
     @author C. Borgmeier
  */
  //==============================================================================
  // ThreeVector
  //==============================================================================
  class ThreeVector: public TObject, public Hep3Vector {
  public:
    /** @name --- Constructors --- */
    ///
    ThreeVector() {}
    ///
    ThreeVector( const ThreeVector& rhs ): Hep3Vector( rhs ) {}
    ///
    ThreeVector( const Hep3Vector& rhs ): Hep3Vector( rhs ) {}
    ///
    ThreeVector( const double x, const double y, const double z ): Hep3Vector( x, y, z ) {}
    
    const ThreeVector& operator=( const ThreeVector& rhs ) {
      *static_cast<Hep3Vector*>(this) = rhs;
      return *this;
    }

    /** @name --- Access methods --- */
    /// x() can be used, too
    inline double X() const { return x(); }
    /// y() can be used, too
    inline double Y() const { return y(); }
    /// z() can be used, too
    inline double Z() const { return z(); }
    
    /// set x-component in cartesian coordinates
    inline void set_x(const double x) { setX(x); }
    /// set y-component in cartesian coordinates
    inline void set_y(const double y) { setY(y); }
    /// set z-component in cartesian coordinates
    inline void set_z(const double z) { setZ(z); }
  }; // end of class ThreeVector
  
#define Cross cross
#define Dot dot
#define Mag2 mag2
#define Mag mag
#ifdef NAMESPACE
} // namspace CLUE
#endif
#endif // USE_ROOT

#ifdef NAMESPACE
namespace CLUE {
#endif

  //==============================================================================
  // operator <<
  //==============================================================================
  inline ostream& operator<< (ostream& os, const ThreeVector& v) {
    os.setf(ios::right,ios::adjustfield);
    os.setf(ios::fixed,ios::floatfield);
    return os << '(' << setw(7) << setprecision(3) << v.x() 
	      << ',' << setw(7) << setprecision(3) << v.y()
	      << ',' << setw(7) << setprecision(3) << v.z()
	      << ')';
  }
#ifdef NAMESPACE
} // namespace CLUE
#endif
#endif


