#ifndef GROVER_WITHOUT_ARTE
#include "arte/arte.hh"
#include "farm/farm.hh"
#endif

#include "grover/higz.h"
#include "grover/grover.hh"

//
// Part of the Grover package
// Please see the header file grover.hh
// for authors etc.
//
//
// Modifications:
// When?     Who?     What?
// YY-MM-DD  Name     Modification
// ------------------------------------------------------------------------
// 00-10-01  M.Sang   added GroverEvent(void) as default
// 00-10-26  M.Sang   declare global variables for Arte implementation
// 00-10-26  M.Sang   Major revision to structure and safety checks
// 00-11-01  M.Sang   RTAR writing implemented
// 00-11-21  M.Sang   clear() RTAR before writing
// 00-12-06  M.Sang   More robust wire-following mechanism
// 00-12-12  M.Sang   Checks for 'vanished wires' (e.g. wires are retracted very
//                    quickly) and prompts GroverPrimaries to re-read the wires
// 01-02-14  M.Sang   Default track selection has changed. We used to use the same
//                    method as Rover v1.07. Now we use the method of Schmelling
//                    and Conde in anl_TrSelect.C. The old selection can still be
//                    used by putting TSELECTION 5 in your kumac.
// 01-03-13  Max      Added preprocessor directives
// 01-04-24  Michael  replaced one oldrunn==evhd->run with getnewwires
//                    changing filling of ggWires and writing of RTAR
// 01-04-27  Michael  bug fix if RTAR not booked 
// 01-05-07  Michael  get wires in the same way for real and mc events 
// 01-06-15  Michael  cloneremoved track selection using Rsegc::Bit20
//                    available with /GROVER/STEERING/TSELECTION   7
// 01-07-02  Michael  cloneremoved track selection rejecting VDS-only tracks
//                    available with /GROVER/STEERING/TSELECTION   8
// 01-07-05  Michael  Change the WARNING in case of  WIREFOLLOEING.
//                    It may improve the target positions for the first 
//                    events in a run but shoul NOT BE USED ON SELECTED EVENTS
// 01-07-16  Michael  Bug fix, now the rrel->flags are set to zero
//                    if RVER is replaced.
// 01-07-31  Michael  Changes due to bit20 -> Clonerm
//
// declare global Grover variables
GroverParticleList ggParticles;
GroverWireList     ggWires;
GroverVertexList   ggPrimaries;
GroverVertexList   ggSecondaries;
GroverVertexList   ggMCPrimaries;
GroverVertexList   ggMCSecondaries;
GroverParticleList ggSpecialParticles;
GroverCuts         ggCuts;

int GroverEvent(void) {

  /////////////////////////////////////////////////////////////////
  //
  // Routine to steer Grover in the automatic (official) 
  // reconstruction. Can be called by the user, to redo primaries,
  // and call secondary finders with some consistency checks. Operates
  // on a set of global variables ggXX.. defined above. ggCuts can
  // be modified with KUIP commands for run-time steering. See 
  // the manual for more info (or read the code...;-)
  //
  // Always under development!
  //
  // M.Sang 05 June 00
  //
  /////////////////////////////////////////////////////////////////

  int error = SUCCESS;

#ifndef GROVER_WITHOUT_ARTE

  // maximum number of vertex-less events on an 'active' wire before
  // re-checking for retracted wires
  const int MAXEMPTIES = 50; // probably 10 would be fine!

  bool verbose = (ggCuts.Verbose == 1);
  const bool wire_following = (ggCuts.GPWireFollowing==1);


  // wire following variables
  static int nempties[9] = {0};
  static bool active[9]  = {false};
  bool seenthistime[9]   = {false};

  
  // Counter for total number of calls 
  static int total_ncalls = 0;
  ++total_ncalls;
  bool firstcall = (total_ncalls == 1);
  if (firstcall) cout 
    << " ******************************************************" << endl
    << " *                                                    *" << endl
    << " *                 WELCOME TO GROVER                  *" << endl
    << " *                                                    *" << endl
    << " *               This is version "
    << GroverVersion
    <<                                      "                 *" << endl
    << " *                                                    *" << endl
    << " *         For information and a manual see:          *" << endl
    << " * http://wwwhera-b.mppmu.mpg.de/analysis/grover.html *" << endl
    << " *                                                    *" << endl
    << " ******************************************************" << endl;

  // Counter for number of calls in THIS event
  ArtePointer<EVHD> evhd = *(ArteTable<EVHD>::begin());
  static int oldrun, oldevt;
  static int ncalls_thisevt;
  bool duplicall = false;  
  bool getnewwires = false;

  if (wire_following && total_ncalls < 25) {
    cout << "GroverEvent.C:" << endl;
    cout << "          WARNING!!! WIREFOLLOWING SHOULD NOT BE DONE ON " << endl
	 << "          SELECTED EVENTS" << endl;
  }


  ncalls_thisevt = 1;
  if (firstcall) {
    oldrun = evhd->run; oldevt = evhd->evt; 
    duplicall = false;
    getnewwires = true;
  }
  else {
    if (evhd->run == oldrun && evhd->evt==oldevt) {
      duplicall = true; // multiple calls!
      ++ncalls_thisevt;
      getnewwires = false;
    }
    else { // first call on this evt
      duplicall = false;
      if (evhd->run != oldrun) {
	getnewwires = true;
	for (int k=0; k<9; ++k) active[k] = false;
      }
    }
    oldrun = evhd->run; oldevt = evhd->evt; // for next time
  }
  if (verbose) cout << "*** GroverEvent: duplicall = " 
		    << (duplicall ? "TRUE" : "FALSE") 
		    << endl;





  // also check for disappearing wires
  if (verbose) {
    cout << "*** GroverEvent: active/quiet wires =";
    for (int i=1; i<9; ++i) cout << " " << active[i] 
				 << "(" << nempties[i] << ")";
    cout << endl;
  }
  bool disappearingwire = false;
  for (int ii=1; ii<9; ++ii) 
    if (active[ii] && (nempties[ii] > MAXEMPTIES)) disappearingwire = true;
  if (disappearingwire) getnewwires = true;
  if (disappearingwire && wire_following) 
    cout << "*** GroverEvent: Disappearing wire?? forcing re-read of GTAR/RTAR" 
	 << endl;




  // Bail out if no GTAR table
  if (!ArteTable<GTAR>::booked()) {
    if (total_ncalls < 25) {
      cout << "*** GroverEvent: no GTAR table - failed!"<<endl;
      ARTE_exit(GRR_SERROR,
		 "GroverEvent(): FATAL - no GTAR table booked!",0);
    }
    return FAILURE; // Cannot continue
  }




  // Code steering
  // 1) Fill track banks?
  bool fill_tracks      = false;
  if ( (!duplicall)                           || 
       (ggParticles.empty())                  || 
       (duplicall && ggCuts.GERefillTracks==1)  ) {
    ggParticles.clear();
    fill_tracks = true;
  }

  // 2) Find primaries?
  bool find_primaries   = false;
  if ( (!duplicall && ggCuts.GEFindPrimaries==1) ||
       ( duplicall && ggCuts.GERedoPrimaries==1)   ) {
    ggPrimaries.clear();
    find_primaries = true;
  }

  // 3) Find Secondaries?
  bool find_secondaries = false;
  if ( (!duplicall && ggCuts.GEFindSecondaries==1) ||
       ( duplicall && ggCuts.GERedoSecondaries==1)   ) {
    // compatibility check
    if (fill_tracks && !find_primaries) {
      cout << "*** GroverEvent: WARNING - secondary vertexing impossible, "
	      "request ignored!"<< endl;
      find_secondaries = false;
      if (ggCuts.GERedoSecondaries==1) 
	ggSecondaries.clear(); // hmmm... not sure about this
    }
    else {
      ggSecondaries.clear();
      find_secondaries = true;
    }
  }

  // 4) Use wire following mechanism or read positions from GTAR & RTAR?
  if (wire_following && 
      //     (firstcall||evhd->run!=oldrun) && 
      getnewwires &&
      !ArteTable<GTAR>::booked()) {
    if (total_ncalls < 25) {
      cout << "*** GroverEvent: wire following requested but no GTAR - exit!"<<endl;
      ARTE_exit(GRR_SERROR,
		 "GroverEvent(): FATAL - wire following requested but no GTAR!",0);
    }
    return FAILURE; // Cannot continue
  }

  // 5) Do some evaluation - efficiency estimation, etc?
  bool verona = (ggCuts.GPEActivate == 1);

  // 6) Which package?
  bool usingGrover = false;
  bool usingVT     = false;
  /*
  if (ggCuts.VertexingPackage == 1) {
    usingGrover = true;  usingVT = false;
  }
  else if (ggCuts.VertexingPackage == 2) {
    usingGrover = true; usingVT = true; // Grover needed for wire alignment
  }
  else {
    cout << "*** GroverEvent: ERROR - unknown vertexing package ID = "
	 << ggCuts.VertexingPackage << endl;
    return FAILURE;
  }
  */
  usingGrover = true; usingVT = false; // no option in this version!

  // 7) Are we using RSEGs? Do we want them to be refitted with (optionally
  //    smeared) MC momentum?
  static bool rseg_changed = false;
  bool refitting_rsegs = false;
  if (!duplicall) rseg_changed = false;
  if (ggCuts.GERefitRSEGTable==1 && 
      !rseg_changed) { 
    if (ArteTable<MCIN>::booked()) refitting_rsegs = true;
    else {
      if (total_ncalls < 25) 
	cout << "*** GroverEvent: WARNING - real data!"
		" cannot refit VDS RSEGs using MC momentum" << endl;
    }
  }

  // 8) Are we writing out the RTAR table?
  bool write_rtar = (ggCuts.GEWriteRTAR==1);






  // Clear lists if first call for this event
  if (!duplicall) {
    ggParticles.clear(); ggPrimaries.clear(); ggSecondaries.clear();
    ggMCPrimaries.clear(); ggMCSecondaries.clear();
    ggSpecialParticles.clear();
    //    ggWires.clear(); 
  }






  // Get wires 
  if (wire_following) {
    if (!duplicall) {
      if (getnewwires) {
	// new balls, please
	if (verbose) cout << "*** GroverEvent: first call/ new run/ quiet wire: "
			  << "setting NewWires=1" << endl;
	ggCuts.GPNewWires = 1;
	ggWires.clear();
	GroverFillFromGTAR(&ggWires);
      }
      else {
	ggCuts.GPNewWires = 0; // continue as normal
      }
    }
    else {
      ggCuts.GPNewWires = 0; // don't adjust wires more than once per event!
      if (verbose) cout << "*** GroverEvent: duplicate call for this event..." 
			<< endl;
    }
  }  
  else { // no wire following - take directly from GTAR or RTAR tables
    if (!duplicall) {
      if (verbose) cout << "*** GroverEvent: no wire following "
			<< endl;
      ggCuts.GPNewWires=1;
      ggWires.clear();
      if (ArteTable<RTAR>::booked() && ArteTable<GTAR>::booked() ) { // if RTAR and GTAR
	GroverFillFromGTAR(&ggWires);
	//  replace GTAR wires with RTAR
	if (verbose) cout << "*** GroverEvent: GTAR and overwriting with RTAR!"
			  << endl;
	for (ArteTable<RTAR>::iterator irtar = ArteTable<RTAR>::begin();
	     irtar != ArteTable<RTAR>::end(); ++irtar) {
	  for (GroverWireList::iterator iwire = ggWires.begin();
	       iwire != ggWires.end(); ++iwire) {
	    if (iwire->Number() == (*irtar)->wnr) {
	      iwire->SetX((*irtar)->x);
	      iwire->SetY((*irtar)->y);
	      iwire->SetZ((*irtar)->z);
	    }
	  }
	}
      }
      else                                          // no GTAR or no RTAR 
	if (ArteTable<RTAR>::booked() ) {           // if RTAR
	  if (verbose) cout << "*** GroverEvent:     no GTAR; using RTAR "
			    << endl;
	  GroverFillFromRTAR(&ggWires);
	}
	else  
	  if (ArteTable<GTAR>::booked() ) {         // no RTAR but GTAR
	    if (verbose) cout << "*** GroverEvent:     no RTAR; using GTAR "
			      << endl;
	    GroverFillFromGTAR(&ggWires);
	  }
	  else                                      // no RTAR and no GTAR
	    {
	      ARTE_exit(GRR_SERROR,
			"GroverEvent(): FATAL - no GTAR and no RTAR",0);
	    }
    }
    //                   if duplicate call for the same event
    else {
      ggCuts.GPNewWires = 0;
      if (verbose) cout << "*** GroverEvent: duplicate call for this event..."
			<< endl;
    }
  }
  



  // check and transform
  bool wires_screwed = false;
  // Dirty hack for runs with bad GTAR
  if (wire_following) FakeWires(&ggWires, evhd->run);
  wires_screwed = !ReasonableWires(&ggWires);
  if (ggWires.empty() || (wires_screwed && wire_following)) {
    if (verbose) cout << "*** GroverEvent: " << endl
		      << "ggWires.empty(): " << ggWires.empty() 
		      << " wires_screwed: " << wires_screwed 
		      << " wire_following: " << wire_following << endl;
    if (total_ncalls < 25) {
      cout << "*****************************************************************" << endl;
      cout << "*** GroverEvent: FATAL ERROR - BAD WIRES. Diagnostic:" << endl;
      cout << "GTAR_wire:   ";
      for (ArteTable<GTAR>::iterator ig = ArteTable<GTAR>::begin();
	   ig != ArteTable<GTAR>::end(); ++ig) cout << (*ig)->wnr << "           ";
      cout << endl 
	   << "GTAR_x:    ";
      for (ArteTable<GTAR>::iterator ig = ArteTable<GTAR>::begin();
	   ig != ArteTable<GTAR>::end(); ++ig) cout << " " << (*ig)->x << " ";
      cout << endl 
	   << "GTAR_y:    ";
      for (ArteTable<GTAR>::iterator ig = ArteTable<GTAR>::begin();
	   ig != ArteTable<GTAR>::end(); ++ig) cout << " " << (*ig)->y << " ";
      cout << endl;
      cout << "*****************************************************************" << endl;
      ARTE_exit(GRR_SERROR,
		"GroverEvent(): FATAL - target wire info is unusable!",0);
    }
    return FAILURE;
  }
  MoveToLocal(&ggWires);




  if (fill_tracks) {
    if (refitting_rsegs) {
      if (verbose) 
	cout << "*** GroverEvent: Refitting VDS RSEGs with MC momenta" << endl;
      int err = GroverRefitRSEG(&ggCuts);
      rseg_changed = true;
      if (err != 0 && total_ncalls < 25) 
	cout << "*** GroverEvent: WARNING - error in rseg refitting" <<endl;
    }

    // Fill particle list
    switch(ggCuts.GETrackChoice) {
    case 0: { 
      // It is expected that ggParticles is filled outside 
      // GroverEvent 
      static int zeroParticles = 0;
      if (ggParticles.empty()) {
	zeroParticles++;
      }
      if (zeroParticles>=50) {
	ARTE_error(GRR_SINFO,
		   "GroverEvent: INFO: 50 events found in a row with no events selected by track selection 0!",0);  	
	zeroParticles=0;
      }
      break;
    }
      case 1: { // default (changes with time)
      GroverFillSmarter(&ggParticles);
      break;
    }
    case 2: { // CATS VDS segments
      GroverFillFromRSEG(&ggParticles);
      break;
    }
    case 3: { // ALL RTRAs (no clone removal)
      GroverFillFromRTRA(&ggParticles);
      break;
    }
    case 4: { // Michael Dressel's selection
      GroverFillLikeMD(&ggParticles);
      break;
    }
    case 5: { // Like Rover 1.07 (RSEGs, then RTRA replacement)
      GroverFillLikeRover(&ggParticles);
      break;
    }
    case 6: { // Schmelling/Conde (anl_TrSelect.C)
      GroverFillSmarter(&ggParticles);
      break;
    }
    case 7: { // use Rsegc::clonerm "Cloneremove" 
      GroverFillClonerm(&ggParticles);
      break;
    }
    case 8: { // use Rsegc::clonerm "Cloneremove" but no VDS-only tracks 
      GroverFillClonermNoVDSonly(&ggParticles);
      break;
    }
    case 9: { // like FillSmarter but no VDSonly tracks
      GroverFillSmartNoVDSonly(&ggParticles);
      break;
    }
    default: {
      GroverFillSmarter(&ggParticles);
      if (total_ncalls<25) 
	cout << "*** GroverEvent: WARNING - unknown track selection"
	        " requested -> using default!" << endl;
      break;
    }
    }
    if (ggParticles.empty()) {
      if (verbose) {
	cout << "*** GroverEvent: WARNING - no ";
	switch(ggCuts.GETrackChoice) {
	case 0: { // comments are allready printed above
	  break;
	}
	case 1: {
	  cout << "particles found using standard selection" << endl;
	  break;
	}
	case 2: {
	  cout << "VDS segments in RSEG" << endl;
	  break;
	}
	case 3: {
	  cout << "tracks in RTRA" << endl;
	  break;
	}
	case 4: 
	case 5:
	case 6:
	case 7:
	case 8:
	case 9:	{    
	  cout << "particles found using special selection" << endl;
	  break;
	}
	default: {
	  cout << "particles found using standard selection" << endl;
	  break;
	}
	}
	error -= 1;
      }
    }
    MoveToLocal(&ggParticles);
  }






  // Call primary finder(s)
  int flag = 0;
  if (find_primaries) {
    if (firstcall || verbose) 
      cout << "*** GroverEvent: finding primaries" <<endl;
    flag = GroverPrimaries(&ggParticles, &ggWires, &ggCuts, &ggPrimaries);
    if (flag != 0) {
      if (total_ncalls<25 || verbose) 
	cout << "*** GroverEvent: WARNING - GroverPrimaries error = " 
	     << flag << endl;
      if (total_ncalls < 25) ARTE_error(GRR_SERROR,
        "GroverEvent(): ERROR - GroverPrimaries() failed!",0);
      error -= 4;;
    }
    if (usingVT) {
      //rPrimaries.clear();
      cout << "*** GroverEvent: ERROR - VT disabled" << endl;
      //flag = GroverVTPrimaries(&Particles, &Wires, &cuts, &rPrimaries);
      if (flag != 0) {
	if (verbose)
	  cout << "*** GroverEvent: WARNING - GroverVTPrimaries error = " 
	       << flag << endl;
	error -= 8;;
      }
    }
  }




  // update counters for quiet wires
  if (!duplicall) {
    int wnr = 0;
    for (GroverVertexList::iterator iver = ggPrimaries.begin();
	 iver != ggPrimaries.end(); ++iver) {
      wnr = iver->Wire();
      if (wnr <=0 || wnr > 8) continue;
      active[wnr] = true;       // wire seen at least once
      seenthistime[wnr] = true; // wire seen in this event
    }
    // update counters for quiet wires
    for (int ii=1; ii<9; ii++) {
      if (seenthistime[ii]) nempties[ii] = 0;
      else {
	if (active[ii]) ++(nempties[ii]);      
      }
    }
  }





  // Get MC vertices if required
  GroverVertexList* mcverts = 0;
  if (verona || ggCuts.GroverDraw==1 || ggCuts.GroverDraw3D==1) {
    if (ArteTable<MCIN>::booked()) {
      ggMCPrimaries.clear();
      ggMCSecondaries.clear();
      flag = GroverFillFromMVER(&ggMCPrimaries, 1);
      if (flag != 0) {
	if (verbose) 
	  cout << "*** GroverEvent: WARNING - couldn't get "
	    "MC primary vertices" << endl;
	error -= 16;
      }
      else mcverts = &ggMCPrimaries;
    }
  }




  // Calculate primary efficiencies, etc...
  if (verona) {
    cout << "*** GroverEvent: WARNING - GroverPrimEval disabled in this version" << endl;
    //flag = GroverPrimEval(&ggPrimaries,mcverts,&ggCuts);
    flag = -1;
    if (flag != 0) {
      if (verbose) 
	cout << "*** GroverEvent: WARNING - GroverPrimEval failed" << endl;
      error -= 32;
    }
  }
  






  // Find Secondaries
  if (find_secondaries) {
    if (firstcall || verbose) 
      cout << "*** GroverEvent: finding secondaries" <<endl;
    flag = GroverSecondaries(&ggParticles, &ggWires, &ggCuts, 
			     &ggPrimaries, &ggSecondaries);
    if (flag != 0) {
      if (verbose) 
	cout << "*** GroverEvent: WARNING - GroverSecondaries failed" << endl;
      error -= 64;
    }
  }








  // Fill RVER
  if (ggCuts.GEWriteRVER==1) {
    // Clear RVER 
    ArteTable<RVER>::clear();
    // set the rrel->flag to zero
    for (ArteTable<RREL>::iterator itrrel = ArteTable<RREL>::begin();
	 itrrel != ArteTable<RREL>::end(); itrrel++) itrrel->flag = 0;
    
    flag = GroverAddToRVER(&ggPrimaries, 1);
    if (flag != 0) {
      if (verbose) 
	cout << "*** GroverEvent: WARNING - couldn't write "
	  "primary vertices into RVER" 
	     << endl;
      error -= 128;
    }
    flag = GroverAddToRVER(&ggSecondaries, 2);
    if (flag != 0) {
      if (verbose) 
	cout << "*** GroverEvent: WARNING - couldn't write "
	  "secondary vertices into RVER" 
	     << endl;
      error -= 256;
    }
  }



  // Write RTAR
  if (write_rtar) {
    ArteTable<RTAR>::clear();
    // fill RTAR
    for (GroverWireList::iterator iwire = ggWires.begin();
	 iwire != ggWires.end(); ++iwire) {
      iwire->LocalToGlobal();
      RTAR rtar;
      int wn = iwire->Number();
      if (wn<0||wn>8) {
	ARTE_exit(GRR_SERROR,
		  "GroverEvent(): FATAL - wire-number out of range",0);
      }
      rtar.x  = iwire->X();
      rtar.y  = iwire->Y();
      rtar.z  = iwire->Z(); 
      rtar.sx = iwire->SX(); // errors on mean positions
      rtar.sy = iwire->SY();
      rtar.sz = iwire->SZ();
      rtar.dx = iwire->SizeX();       // iwire dimensions
      rtar.dy = iwire->SizeY();
      rtar.dz = iwire->SizeZ();
      rtar.wnr=wn;
      ArteTable<RTAR>::push_back(rtar);
    }
  }


  
  if (verbose) cout 
    <<"*** GroverEvent: ggX prim,sec,par,mcprim,mcsec,wires,special "
    << ggPrimaries.size() << " "
    << ggSecondaries.size() << " "
    << ggParticles.size() << " "
    << ggMCPrimaries.size() << " "
    << ggMCSecondaries.size() << " "
    << ggWires.size() << " "
    << ggSpecialParticles.size() << " "
    << endl;

#endif

  return error;
}



/////////////////////////////////////////////
int GroverEvent(GroverCuts* usercuts)
{ 
  int r = 0;

#ifndef GROVER_WITHOUT_ARTE
  // hack to emulate old behaviour
  // M Sang 00-10-26
  // default
  GroverCuts temp = ggCuts;              // memberwise copy 
  if (usercuts != 0) ggCuts = *usercuts; // -"-
  r = GroverEvent();
  if (usercuts != 0) ggCuts = temp;      // -"-
#endif

  return r;
}
