#ifndef __PATTERNMATCHIF_HH
#define __PATTERNMATCHIF_HH
// ********************************************************************
//
// source:
//
// type:      source code
//
// created:   28. Feb 2001
//
// author:    Thorsten Glebe
//            HERA-B Collaboration
//            Max-Planck-Institut fuer Kernphysik
//            Saupfercheckweg 1
//            69117 Heidelberg
//            Germany
//            E-mail: T.Glebe@mpi-hd.mpg.de
//
// Description: PatternMatch object interface
//
// changes:
// 28 Feb 2001 (TG) creation
// 21 Aug 2001 (TG) added TrackIf inheritance, added getPattern(), mass(),
//                  massError(), massProb()
// 28 Aug 2001 (TG) added isMCmatch(), getRecoTrack(), getMC_Track()
// 29 Aug 2001 (TG) operator==() added
// 31 Aug 2001 (TG) added PatternMatch(i) from DecayPatternMatchIf
//
// ********************************************************************
#include <iosfwd>
#include "interfaces/TrackIf.hh"

class RecoTrack;
class MC_Track;
class SPattern;
class GTopp;

/** Abstract class which defines the minimal interface for a PatternMatch object.

    @author T. Glebe
*/
//==============================================================================
// PatternMatchIf
//==============================================================================
class PatternMatchIf : public TrackIf {
public:
  /** @name --- Access methods --- */
  /// get the TrackIf object
  virtual const TrackIf& getTrackIf() const = 0;
  /// get the RecoTrack object
  virtual const RecoTrack& getRecoTrack() const = 0;
  /// get the MC\_Track object
  virtual const MC_Track& getMC_Track() const = 0;
  /// read only pattern access
  virtual const SPattern& getPattern() const = 0;
  /// invariant mass of pattern
  virtual double mass() const = 0;
  /// error of inv. mass
  virtual double massError() const = 0;
  /// $\chi^2$ probability including mass-term
  virtual double massProb(double scale) const = 0;
  /// is this a match with a MC track?
  virtual bool isMCmatch() const = 0;
  /// return a PatternMatchIf pointer
  const PatternMatchIf* PatternMatch(unsigned int i) const;

  /** @name --- Expert methods --- */
  /// calculate mother track + cov. matrix
  virtual bool calc_mother() = 0;
  /// calculate mother track
  virtual bool calc_mother_tr() = 0;
  /// calculate mother covariance matrix
  virtual bool calc_mother_cov() = 0;
  /// used by operator<<()
  virtual std::ostream& print( std::ostream& ) const = 0;

  /** @name --- Operators --- */
  /// compare with GTopp
  virtual bool operator==(const GTopp& rhs) const = 0;
}; // end of class PatternMatchIf


//==============================================================================
// operator <<
//==============================================================================
inline std::ostream& operator<< (std::ostream& os, const PatternMatchIf& t) {
  return t.print(os);
}
#endif
