// ********************************************************************
//
// source:
//
// type:      source code
//
// created:   20. Apr 2001
//
// author:    Thorsten Glebe
//            HERA-B Collaboration
//            Max-Planck-Institut fuer Kernphysik
//            Saupfercheckweg 1
//            69117 Heidelberg
//            Germany
//            E-mail: T.Glebe@mpi-hd.mpg.de
//
// Description: SPattern implementation file
//
// changes:
// 20 Apr 2001 (TG) creation
// 27 Apr 2001 (TG) cut_ added, rejectTrack() added
// 30 Apr 2001 (TG) cut_ removed
// 18 Mai 2001 (TG) assert() removed in set_cc()
// 03 Jul 2001 (TG) improved property_ handling
// 03 Jul 2001 (TG) getDecayString() added
// 17 Aug 2001 (TG) operator==(MC_Track) added
// 22 Aug 2001 (TG) removed code for charge conjugation
// 23 Aug 2001 (TG) added setProperty()
// 24 Aug 2001 (TG) fillMC() added
// 27 Aug 2001 (TG) isReconstructed(MC_Track), fillRECOfromMC() added
// 28 Aug 2001 (TG) added isNotReconstructed(MC_Track), fillNORECOfromMC()
//
// ********************************************************************

#include <iostream>
#include <iomanip>
#include "pattern/SPattern.hh"
#include "pattern/SPatternMatch.hh"
#include "pattern/SProperty.hh"
#include "clue/GTopp.hh"
#include "clue/PdtNames.hh"
#include "clue/MC_Track.hh"
#include "clue/RecoTrack.hh"
#include "pattern/SPatternMatchContainer.hh"

//==============================================================================
// charge: declaration of global (and thus unique) SPMember object
//==============================================================================
template<> extern const SPMember<TrackIf,int> charge;

//==============================================================================
// Constructors
//==============================================================================
SPattern::SPattern() :
  topp_    (0),
  rm_      (0.),
  property_(SmartPointer<SProperty<TrackIf,bool> >()),
  cprop    (SmartPointer<SProperty<TrackIf,bool> >()),
  cut_     (0),
  spmc_    (0)
{ set_cprop(charge == topp_->chrg()/3.); }

SPattern::SPattern(const SPattern& rhs) :
  topp_    ( rhs.topp_ ),
  rm_      ( rhs.rm_ ),
  property_( rhs.property_ ),
  cprop    ( rhs.cprop ),
  cut_     ( rhs.cut_ ),
  spmc_    ( rhs.spmc_ )
{}

SPattern::SPattern(const ToppIf& tp, const SPattern& rhs) :
  topp_    ( &tp ),
  rm_      ( tp.valid() ? tp.mass() : 0 ),
  property_( rhs.property_ ),
  cprop    ( rhs.cprop ),
  cut_     ( rhs.cut_ ),
  spmc_    ( rhs.spmc_ )
{}

SPattern::SPattern(const ToppIf& gt) :
  topp_    ( &gt ),
  rm_      ( gt.valid() ? gt.mass() : 0 ),
  property_( SmartPointer<SProperty<TrackIf,bool> >() ),
  cprop    ( SmartPointer<SProperty<TrackIf,bool> >() ),
  cut_     ( 0 ),
  spmc_    ( 0 )
{ set_cprop(charge == topp_->chrg()/3.); }

SPattern::SPattern(const ToppIf& gt, const SProperty<TrackIf,bool>& rhs) :
  topp_    ( &gt ),
  rm_      ( gt.valid() ? gt.mass() : 0 ),
  property_( SmartPointer<SProperty<TrackIf,bool> >(rhs.propertyClone()) ),
  cprop    ( SmartPointer<SProperty<TrackIf,bool> >() ),
  cut_     ( 0 ),
  spmc_    ( 0 )
{ set_cprop(charge == topp_->chrg()/3. & *property_); }

SPattern::~SPattern() {}

SPattern& SPattern::operator=(const SPattern& rhs) {

  topp_     = rhs.topp_;
  rm_       = rhs.rm_;
  property_ = rhs.property_;
  cprop     = rhs.cprop;
  cut_      = rhs.cut_;
  spmc_     = rhs.spmc_;

  return *this;
}

//==============================================================================
// topp(): get topp object
//==============================================================================
const ToppIf& SPattern::topp() const { return *topp_; }

//==============================================================================
// evalTrack
//==============================================================================
bool SPattern::evalTrack(const TrackIf& rhs) const
{ return cprop.null() ? true : cprop->evaluate(rhs); }

//==============================================================================
// evalTrackProp
//==============================================================================
bool SPattern::evalTrackProp(const TrackIf& rhs) const
{ return property_.null() ? true : property_->evaluate(rhs); }

//==============================================================================
// setProperty
//==============================================================================
void SPattern::setProperty(const SProperty<TrackIf,bool>& rhs) {
  property_ = SmartPointer<SProperty<TrackIf,bool> >(rhs.propertyClone());
  set_cprop(charge == topp_->chrg()/3. & *property_);
}

//==============================================================================
// rm(): rest-masses
//==============================================================================
const double* SPattern::rm() const { return &rm_; }

//==============================================================================
// len(): length of rest-mass array
//==============================================================================
unsigned int SPattern::len() const { return 1; }

//==============================================================================
// operator==
//==============================================================================
bool SPattern::operator==(const MC_Track& rhs) const {
  return *topp_ == rhs.topp() && evalTrackProp(rhs);
}

//==============================================================================
// isReconstructed()
//==============================================================================
bool SPattern::isReconstructed(const MC_Track& rhs) const {
  return *topp_ == rhs.topp() && rhs.rtrack.size() > 0 && evalTrackProp(rhs);
}

//==============================================================================
// isNotReconstructed()
//==============================================================================
bool SPattern::isNotReconstructed(const MC_Track& rhs, bool& no_break) const {
  return ((no_break = (*topp_ == rhs.topp() && evalTrackProp(rhs))) && rhs.rtrack.size() == 0);
}

//==============================================================================
// getString
//==============================================================================
string SPattern::getString() const {
  return (!property_.null()) ? 
    getName() + string(" with ") + property_->getString() :
    getName();
}

//==============================================================================
// getName
//==============================================================================
string SPattern::getName() const {
  return (topp_==0) ? string("?") : string(topp_->name());
}

//==============================================================================
// getDecayString
//==============================================================================
string SPattern::getDecayString(unsigned int lvl) const {
  return getName();
}

//==============================================================================
// operator|
//==============================================================================
SPattern SPattern::operator|( const SPattern& rhs ) const {
  if( topp_ == rhs.topp_ ) 
    return SPattern(*topp_, *property_ | *rhs.property_);
  else
    return SPattern(Any, *property_ | *rhs.property_);
}

//==============================================================================
// set_cprop
//==============================================================================
void SPattern::set_cprop(const SProperty<TrackIf,bool>& rhs) {
  cprop = SmartPointer<SProperty<TrackIf,bool> >(rhs.propertyClone());
}

//==============================================================================
// printProperty
//==============================================================================
void SPattern::printProperty(std::ostream& os, unsigned int lvl) const {

  os.setf(ios::left,ios::adjustfield);
  os.setf(ios::fixed,ios::floatfield);

  const string tabb(lvl*3,' ');
  os << setprecision(2) << tabb << getName();

  if(!cprop.null()) {
    os << ", ";
    cprop->print(os);
  } else {
    if(!property_.null()) {
      os << ", ";
      property_->print(os);
    }
  }

  return;
}

//==============================================================================
// getContainer
//==============================================================================
const SPMP_l* SPattern::getContainer() const {
  return &spmc_->pmatches();
}

//==============================================================================
// print
//==============================================================================
std::ostream& SPattern::print(std::ostream& os) const {
  printProperty(os);
  return os;
}

//==============================================================================
// exist
//==============================================================================
template <class InputContainer>
bool SPattern::exist(const InputContainer&  src) const {
  typedef typename InputContainer::const_iterator Cit;

  const Cit end = src.end();
  for(Cit it = src.begin(); it != end; ++it) {
    const typename InputContainer::value_type& track = *it;
    if( evalTrack(*track) ) {
      return true;
    }
  }
  return false;
}

//==============================================================================
// fillMC
//==============================================================================
const PatternMatchIf* SPattern::fillMC(const MC_Track& mtra) const {
  const SPatternMatch& pm = spmc_->push_back(const_cast<MC_Track&>(mtra));
  return &pm;
}

//==============================================================================
// fillRECOfromMC: fill only with first reconstructed track
//==============================================================================
PatternMatchIf* SPattern::fillRECOfromMC(const MC_Track& mtra) const {
  SPatternMatch& pm = spmc_->push_back(**mtra.rtrack.begin());
  return &pm;
}

//==============================================================================
// fillNORECOfromMC
//==============================================================================
PatternMatchIf* SPattern::fillNORECOfromMC(const MC_Track& mtra) const {
  PatternMatchIf* pmptr = 0;
  if(mtra.rtrack.size() == 0) {
    pmptr = &spmc_->push_back(const_cast<MC_Track&>(mtra));
  } else {
    pmptr = &spmc_->push_back(**mtra.rtrack.begin());
//      SPatternMatch& pm = spmc_->push_back(**mtra.rtrack.begin());
//      pmptr = &pm;
  }
  return pmptr;
}
