// ********************************************************************
//
// source:
//
// type:      source code
//
// created:   20. Apr 2001
//
// author:    Thorsten Glebe
//            HERA-B Collaboration
//            Max-Planck-Institut fuer Kernphysik
//            Saupfercheckweg 1
//            69117 Heidelberg
//            Germany
//            E-mail: T.Glebe@mpi-hd.mpg.de
//
// Description: SPatternMatch implementation file
//
// changes:
// 20 Apr 2001 (TG) creation
// 27 Apr 2001 (TG) pt(), xfabs() added
// 02 Mai 2001 (TG) phi(), theta(), eta() added
// 06 Jun 2001 (TG) changed Vtx relation type from VertexIf to DecayPatternMatchIf
// 11 Jun 2001 (TG) added cov_tx(), cov_ty(), cov_p()
// 28 Jun 2001 (TG) removed templates
// 29 Jun 2001 (TG) added charge(), pz(), energy(), xf(), rap(), removed xfabs()
// 03 Jul 2001 (TG) added CTOR
// 05 Jul 2001 (TG) removed decay, added collect()
// 09 Jul 2001 (TG) isValid(), valid(), invalid() from TrackIf added
// 10 Jul 2001 (TG) bug fix: removed #ifndef ... statements
// 10 Jul 2001 (TG) operator==() added
// 22 Aug 2001 (TG) added #include SPatternMatch.hh
// 28 Aug 2001 (TG) ismc_ added, CTOR added, isMCmatch() added
// 29 Aug 2001 (TG) added operator==(GTopp)
// 31 Aug 2001 (TG) added PatternMatch(i)
//
// ********************************************************************
#include <iostream>
#include <iomanip>
#include "pattern/SPatternMatch.hh"
#include "interfaces/DecayPatternMatchIf.hh"
#include "smatrix/SVector.hh"
#include "clue/RecoSegment.hh"
#include "smatrix/Prob.hh"

//==============================================================================
// Constructors
//==============================================================================
SPatternMatch::SPatternMatch() : 
  vtx     (),
  track   (0),
  rs_     (0),
  pattern (SmartPointer<SPattern>()),
  user_   (0),
  N_      (0),
  ismc_   (false)
 {}

SPatternMatch::SPatternMatch(RecoSegment& rs, const SPattern& pat) :
  vtx     (),
  track   ( &rs ),
  rs_     ( &rs ),
  pattern ( SmartPointer<SPattern>(pat) ),
  user_   (0),
  N_      (0),
  ismc_   (false)
{
  createSingleRelation(rs,*this);
}

SPatternMatch::SPatternMatch(RecoSegment& rs, const SmartPointer<SPattern>& pat) :
  vtx      (),
  track    ( &rs ),
  rs_      ( &rs ),
  pattern  ( pat ),
  user_    (0),
  N_       (0),
  ismc_    (false)
{
  createSingleRelation(rs,*this);
}

SPatternMatch::SPatternMatch(MC_Track& mt, const SmartPointer<SPattern>& pat) :
  vtx      (),
  track    ( &mt ),
  rs_      ( 0 ),
  pattern  ( pat ),
  user_    (0),
  N_       (0),
  ismc_    (true)
{
}

SPatternMatch::SPatternMatch(TrackIf& tr) :
  vtx      (),
  track    ( &tr ),
  rs_      ( 0 ),
  pattern  ( SmartPointer<SPattern>() ),
  user_    (0),
  N_       (0),
  ismc_    (false)
{
}

SPatternMatch::SPatternMatch(TrackIf& tr, const SPattern& pat) :
  vtx      (),
  track    ( &tr ),
  rs_      ( 0 ),
  pattern  ( SmartPointer<SPattern>(pat) ),
  user_    (0),
  N_       (0),
  ismc_    (false)
{
}

SPatternMatch::SPatternMatch(TrackIf& tr, const SmartPointer<SPattern>& pat) :
  vtx      (),
  track    ( &tr ),
  rs_      ( 0 ),
  pattern  ( pat ),
  user_    (0),
  N_       (0),
  ismc_    (false)
{
}

SPatternMatch::SPatternMatch(const SPatternMatch& rhs) :
  vtx      ( rhs.vtx ),
  track    ( rhs.track ),
  rs_      ( rhs.rs_ ),
  pattern  ( rhs.pattern ),
  user_    ( (rhs.user_ != 0) ? new float[rhs.N_] : 0 ),
  N_       ( rhs.N_ ),
  ismc_    ( rhs.ismc_ )
{
  for(unsigned int i=0; i<N_; ++i) {
    user_[i] = rhs.user_[i];
  }
  if(rs_!=0) createSingleRelation(*rs_,*this);
}

SPatternMatch::~SPatternMatch()
{
  if(user_!=0) delete[] user_;
  if(rs_!=0) removeSingleRelation(*rs_,*this);
  vtx.unbook_all(*this);
}

//==============================================================================
// operator==
//==============================================================================
bool SPatternMatch::operator==( const TrackIf& rhs ) const {
  return &rhs == track;
}

//==============================================================================
// operator==
//==============================================================================
bool SPatternMatch::operator==( const GTopp& rhs ) const {
  return rhs == pattern->topp();
}

//==============================================================================
// PatternMatchIf access functions
//==============================================================================
const TrackIf&   SPatternMatch::getTrackIf() const {
  return *this; 
}
const RecoTrack& SPatternMatch::getRecoTrack() const 
{ return static_cast<const RecoTrack&>(*track); }
const MC_Track&  SPatternMatch::getMC_Track()   const 
{ return static_cast<const MC_Track&>(*track); }
const SPattern&  SPatternMatch::getPattern() const { return *pattern; }
double           SPatternMatch::mass()       const { return pattern->topp().mass(); }
double           SPatternMatch::massError()  const { return 0; }
bool             SPatternMatch::calc_mother()      { return true; }
bool             SPatternMatch::calc_mother_tr()   { return true; }
bool             SPatternMatch::calc_mother_cov()  { return true; }
bool             SPatternMatch::isMCmatch()  const { return ismc_; }

const PatternMatchIf* SPatternMatch::PatternMatch(unsigned int i) const
{ return 0; }

double SPatternMatch::massProb(double scale) const 
{ return Prob(track->chi2(), track->ndf()); }

//==============================================================================
// TrackIf forwarding functions
//==============================================================================
float SPatternMatch::x()        const { return track->x(); }
float SPatternMatch::y()        const { return track->y(); }
float SPatternMatch::x(float z) const { return track->x(z); }
float SPatternMatch::y(float z) const { return track->y(z); }
float SPatternMatch::z()        const { return track->z(); }
float SPatternMatch::tx()       const { return track->tx(); }
float SPatternMatch::ty()       const { return track->ty(); }
float SPatternMatch::p()        const { return track->p(); }
float SPatternMatch::pt()       const { return track->pt(); }
float SPatternMatch::pz()       const { return track->pz(); }
float SPatternMatch::chi2()     const { return track->chi2(); }
float SPatternMatch::phi()      const { return track->phi(); }
float SPatternMatch::theta()    const { return track->theta(); }
float SPatternMatch::eta()      const { return track->eta(); }
int   SPatternMatch::charge()   const { return track->charge(); }

unsigned short int SPatternMatch::ndf() const { return track->ndf(); }

float SPatternMatch::energy(double rm) const {
  return (rm==0.) ? track->energy(pattern->topp().mass()) : track->energy(rm);
}

float SPatternMatch::xf(double rm) const { 
  return (rm==0.) ? track->xf(pattern->topp().mass()) : track->xf(rm);
}

float SPatternMatch::rap(double rm) const { 
  return (rm==0.) ? track->rap(pattern->topp().mass()) : track->rap(rm);
}
  
bool SPatternMatch::isValid() const { return track->isValid(); }
void SPatternMatch::valid()         { track->valid(); }
void SPatternMatch::invalid()       { track->invalid(); }

SVector<double,3> SPatternMatch::xvec() const { return track->xvec(); }
SVector<double,3> SPatternMatch::tvec() const { return track->tvec(); }
SVector<double,3> SPatternMatch::evec() const { return track->evec(); }
SVector<double,3> SPatternMatch::pvec() const { return track->pvec(); }

const SMatrix<double,5>& SPatternMatch::COV()  const { return track->COV(); }
const SMatrix<double,5>& SPatternMatch::CINV() const { return track->CINV(); }

float SPatternMatch::cov_x(double dz=0.) const { return track->cov_x(dz); }
float SPatternMatch::cov_y(double dz=0.) const { return track->cov_y(dz); }

float SPatternMatch::cov_tx() const { return track->cov_tx(); }
float SPatternMatch::cov_ty() const { return track->cov_ty(); }
float SPatternMatch::cov_p()  const { return track->cov_p(); }

void SPatternMatch::propagate(const double z) { track->propagate(z); }

//==============================================================================
// collect
//==============================================================================
void SPatternMatch::collect(vector<TrackIf*>& c) const {
  c.push_back(track);
}


//==============================================================================
// print
//==============================================================================
std::ostream& SPatternMatch::print(std::ostream& os) const {
  if(track != 0) os << "Track: " << *track << endl;
  if(!pattern.null()) os << "Pattern: " << *pattern;
  return os;
}

