#ifndef __SDecayPattern_HH
#define __SDecayPattern_HH
// ********************************************************************
//
// source:
//
// type:      source code
//
// created:   20. Apr 2001
//
// author:    Thorsten Glebe
//            HERA-B Collaboration
//            Max-Planck-Institut fuer Kernphysik
//            Saupfercheckweg 1
//            69117 Heidelberg
//            Germany
//            E-mail: T.Glebe@mpi-hd.mpg.de
//
// Description: Class describing a decay pattern, base part.
//              This base part is only responsible for saving the
//              track patterns.
//
// changes:
// 20 Apr 2001 (TG) creation
// 23 Apr 2001 (TG) deleted all code dealing with CC, added select()
// 02 Mai 2001 (TG) flip_to_cc() added
// 28 Jun 2001 (TG) removed template parameters C and T
// 29 Jun 2001 (TG) added charge()
// 02 Jul 2001 (TG) added inheritance from SPattern
// 04 Jul 2001 (TG) added vprob_, hasVtxProperty, setVtxProperty(), evalVtxProperty()
// 17 Aug 2001 (TG) added operator==(MC_Track), removed operator==(ToppIf)
// 21 Aug 2001 (TG) removed hasVtxProp(), moved cut_, set_cut(), get_cut() to
//                  SPattern
// 22 Aug 2001 (TG) added operator&, added CTOR, changed rm(), added len()
// 23 Aug 2001 (TG) removed getBasicDecayString(), added sdpmc_,
//                  set_container(), getContainer()
// 24 Aug 2001 (TG) added getVtxProperty(), fillMC()
// 27 Aug 2001 (TG) added isReconstructed(MC_Track), fillRECOfromMC()
// 28 Aug 2001 (TG) added isNotReconstructed(MC_Track), fillmtramap(), fillNORECOfromMC()
//
// ********************************************************************

#include <iosfwd>
#include <set>
#include <multimap.h>
#include "pattern/P_trait.hh"
#include "smatrix/SVector.hh"
#include "pattern/SPattern.hh"

class VertexIf;
class ToppIf;
class PatternMatchIf;
class MC_Track;
template <unsigned int NTR> class SDecayPatternMatchContainer;

/** A class which handles selection pattern for decaying particles.
    This class handles the kind of information which is independent from the
    Vertex object, like the Topp table entry of the decay and the list of track
    patterns which are used to select the tracks which will then be combined
    and vertexed.

    @memo Generic class
    @author T. Glebe
*/
//==============================================================================
// SDecayPattern
//==============================================================================
template <unsigned int NTR>
class SDecayPattern : public SPattern {

  typedef typename DP_trait<NTR>::SDPM_l  SDPM_l;
  //  typedef multimap<const char*, const SPattern*> PMAP;
  typedef multimap<const char*, MC_Track*> MMAP;
  typedef multimap<const char*, SP_sp>           PMAP;
  typedef multimap<const char*, SP_sp>::iterator PMAP_it;


public:
  /** @name --- Constructors --- */
  ///
  SDecayPattern();
  ///
  SDecayPattern( const SDecayPattern<NTR>& rhs );
  ///
  SDecayPattern( const SDecayPattern<NTR-1>& lhs, const SP_sp& rhs);
#ifdef XXX
  ///
  template <unsigned int NTR2>
  SDecayPattern( const SmartPointer<SDecayPattern<NTR2> >& lhs, const SP_sp& rhs);
#endif
  ///
  SDecayPattern( const ToppIf& tp, const SDecayPattern<NTR>& rhs);
  ///
  SDecayPattern( const ToppIf& tp, 
		 const SP_sp& lhs,
		 const SP_sp& rhs);
  ///
  virtual ~SDecayPattern();
  ///
  SDecayPattern<NTR>& operator=( const SDecayPattern<NTR>& rhs );

  /** @name --- Operators --- */
  ///
  SDecayPattern<NTR+1> operator&( const SP_sp& rhs) const;
  ///
  bool operator==(const MC_Track& rhs) const;

  /** @name --- overloaded PatternIf access functions --- */
  /// return pointer to vector with rest-masses
  const double* rm() const;
  /// return length of vector (=NTR)
  unsigned int len() const;
  /// string with decay
  string getDecayString(unsigned int lvl=0) const;
  /// print propertyString
  void printProperty(std::ostream& os, unsigned int lvl=0) const;
  /// get pointer to list of pattern matches
  const SPMP_l* getContainer() const;
  /// fill container from MC structure
  const PatternMatchIf* fillMC(const MC_Track& rhs) const;
  /// same as operator==(MC\_Track), but also check whether MC\_Track was reconstructed
  bool isReconstructed(const MC_Track& rhs) const;
  /// same as operator==(MC\_Track), but also check if MC\_Track was not reconstructed
  bool isNotReconstructed(const MC_Track& rhs, bool& no_break) const;
  /// fill container with reconstructed tracks from MC event
  PatternMatchIf* fillRECOfromMC(const MC_Track& mt) const;
  /// fill container with reconstructed tracks or (if not reconstr.) MC tracks from MC event
  PatternMatchIf* fillNORECOfromMC(const MC_Track& mt) const;

  /** @name --- Access functions --- */
  ///
  static const unsigned int size = NTR;
  /// read-only pattern access
  const SP_sp& pattern(unsigned int i) const;
  /// read/write pattern access
  SP_sp& pattern(unsigned int i);
  /// set mass range cut around expected invariant mass
  void massrange(double mr);
  /// return mass range cut
  double massrange() const;
#ifdef XXX
  /// get particle name
  string getName() const;
#endif
  /// get the pattern string
  string getString() const;
  ///
  string getPatternString() const;
  /// check vertex property
  bool evalVtxProperty(const VertexIf& rhs) const;
  /// return rest-mass vector
  const SVector<double,NTR>& rmvec() const;
  
  /** @name --- Expert functions --- */
  /// set Vertex property object
  void setVtxProperty(const SProperty<VertexIf,bool>& rhs);
  /// get Vertex property object
  SmartPointer<SProperty<VertexIf,bool> >& getVtxProperty() const;
  /// set pointer to container object
  void set_container(SDecayPatternMatchContainer<NTR>* rhs);
  /// get a pattern by TOPP id
  const SP_sp get(const ToppIf& topp) const;
  /// fill an array with a sorted list of list<PatternMatchIf>*
  void fill_SPMP_l(const SPMP_l** array);
  /// used by operator<<()
  virtual std::ostream& print(std::ostream& os) const;

protected:
  /// internal utility function
  void fillmtramap(const MC_Track& rhs) const;

  SP_sp          patlist[NTR];              // list of pattern pointer!
  double                dm;                 // mass range of decaying particle
  SVector<double,NTR>   rm_;                // vector of restmasses
  SmartPointer<SProperty<VertexIf,bool> > vprop_; // vertex properties
  PMAP    pattmap;                          // sorted map for Topp/SPattern 
  mutable MMAP    mtramap;                  // sorted map of Topp/MC_Track pointers
  SDecayPatternMatchContainer<NTR>* sdpmc_; // pointer to container object
}; // end of SDecayPattern

//==============================================================================
// operator<<
//==============================================================================
template <unsigned int NTR>
inline std::ostream& operator<<(std::ostream& os,
				const SDecayPattern<NTR>& rhs) {
  return rhs.print(os);
}

template <unsigned int NTR>
inline std::ostream& operator<<(std::ostream& os,
				const SmartPointer<SDecayPattern<NTR> >& rhs) {
  return rhs->print(os);
}

//==============================================================================
// operator&
//==============================================================================
template <unsigned int NTR>
inline SDecayPattern<2> operator&(const SmartPointer<SDecayPattern<NTR> >& lhs,
				  const SP_sp& rhs) {
  return SDecayPattern<2>(lhs,rhs);
}

//==============================================================================
// operator&
//==============================================================================
template <unsigned int NTR>
inline SDecayPattern<2> operator&(const SP_sp& lhs,
				  const SmartPointer<SDecayPattern<NTR> >& rhs) {
  return SDecayPattern<2>(lhs,rhs);
}
#include "pattern/SDecayPattern.icc"
#endif
