#ifndef __SDecayPatternMatch_HH
#define __SDecayPatternMatch_HH
// ********************************************************************
//
// source:
//
// type:      source code
//
// created:   23. Apr 2001
//
// author:    Thorsten Glebe
//            HERA-B Collaboration
//            Max-Planck-Institut fuer Kernphysik
//            Saupfercheckweg 1
//            69117 Heidelberg
//            Germany
//            E-mail: T.Glebe@mpi-hd.mpg.de
//
// Description: Class for handling matched vertices.
//
// changes:
// 23 Apr 2001 (TG) creation
// 24 Apr 2001 (TG) valid(), invalid(), TrackIf members added
// 25 Apr 2001 (TG) added prob()
// 27 Apr 2001 (TG) pt(), xfabs() added
// 02 Mai 2001 (TG) phi(), theta(), eta() added
// 02 Mai 2001 (TG) operator==() added
// 07 Mai 2001 (TG) vertex() added
// 18 Mai 2001 (TG) set_user(), user(), user_size(), user_, N_ added
// 06 Jun 2001 (TG) added inheritance from DecayPatternMatchIf, added massProb()
// 11 Jun 2001 (TG) added massError(), cov_tx(), cov_ty(), cov_p()
// 13 Jun 2001 (TG) added vtx_cov_x(), vtx_cov_y(), vtx_cov_z()
// 28 Jun 2001 (TG) removed template parameter C and T
// 29 Jun 2001 (TG) added charge(), pz(), energy(), xf(), rap(), removed xfabs()
// 05 Jul 2001 (TG) collect() added
// 06 Jul 2001 (TG) added scale arg to massProb()
// 09 Jul 2001 (TG) bug fix: moved SPatternMatch-Relation to DecayPatternMatchIf
// 09 Jul 2001 (TG) valid(), invalid() removed (now part of TrackIf), added ntracks
// 25 Jul 2001 (TG) added track()
// 17 Aug 2001 (TG) CTOR added
// 20 Aug 2001 (TG) moved user field + access functions to SPatternMatch
// 21 Aug 2001 (TG) inheritance from new DecayPatternMatchIf class (located in
//                  interfaces package), added getTrackIf, added PatternMatch
//                  relations (out), removed maxImpact()
// 22 Aug 2001 (TG) changed stored pattern to SPattern
// 24 Aug 2001 (TG) added mcmother_
// 27 Aug 2001 (TG) CTOR added
// 28 Aug 2001 (TG) ismc_ added, CTOR added, set_mother(), set_MC(),
//                  PatternMatch(i) added, getRecoTrack(), getMC_Track() added
// 29 Aug 2001 (TG) added operator==(GTopp)
// 05 Sep 2001 (TG) added user_, N_, set_user(), user_size(), user() members
// 06 Sep 2001 (TG) removed getTrackIf() (obsolete)
//
// ********************************************************************
#include <iosfwd>
#include "interfaces/DecayPatternMatchIf.hh"
#include "grelation/GRelationList.hh"

template <unsigned int NTR> class SDecayPattern;
template <unsigned int NTR> class SVertex;

/** Pattern Match class for Decays.
    SDecayPatternMatch represents a SVertex vertex object together with the
    DecayPattern which was used to fit and select the vertex object.

    @memo Generic class
    @author T. Glebe
*/
//==============================================================================
// SDecayPatternMatch
//==============================================================================
template <unsigned int NTR>
class SDecayPatternMatch : public DecayPatternMatchIf {
public:
  static const unsigned int size = NTR;

  /** @name --- Constructors --- */
  ///
  SDecayPatternMatch();
  /// create from MC
  SDecayPatternMatch(const MC_Track& t,
		     const SmartPointer<SPattern>& vpat);
  /// create from reconstructed track
  SDecayPatternMatch(const TrackIf& t,
		     const SmartPointer<SPattern>& vpat);
  /// create from Vertex
  SDecayPatternMatch(SmartPointer<SVertex<NTR> >& v,
		     const SmartPointer<SPattern>& vpat);
  /// take over ownership of vertex object
  SDecayPatternMatch(SVertex<NTR>* v,
		     const SmartPointer<SPattern>& vpat);
  ///
  virtual ~SDecayPatternMatch();
  
  /** @name --- TrackIf access methods --- */
  /// Track $x$ position
  float x()   const;
  /// Track $y$ position
  float y()   const;
  /// Track $x$ position at $z$
  float x(float z)   const;
  /// Track $y$ position at $z$
  float y(float z)   const;
  /// Track $z$ position
  float z()   const;
  /// Track slope $t_x$
  float tx()  const;
  /// Track slope $t_y$
  float ty()  const;
  /// Track momentum $p$
  float p()   const;
  /// transverse Track momentum $p_t$
  float pt()  const;
  /// Track momentum along $z$
  float pz()  const;
  /// azimuthal angle $\phi$ [deg]
  float phi() const;
  /// polar angle $\theta = \cos^{-1}(e_z)$ [deg]
  float theta() const;
  /// rapidity $\eta = -\log\tan(\theta/2.)$
  float eta() const;
  /// track charge
  int   charge() const;

  /// track energy $E=\sqrt{m^2 + p^2}$
  float energy(double rm = 0.) const;
  /// Feynman-$x$: $x_F = \frac{E+p_z}{(E+p_z)_{max}}$
  float xf(double rm = 0.) const;
  /// Rapidity $y = \frac{1}{2}\ln(\frac{E+p_z}{E-p_z})$
  float rap(double rm = 0.) const;
  
  /// is PatternMatch valid?
  virtual bool isValid() const;
  /// mark PatternMatch as valid
  virtual void valid();
  /// mark PatternMatch as invalid
  virtual void invalid();

#ifndef __CINT__
  /// $\vec{x} = (x,y,z)$
  SVector<double,3> xvec() const;
  /// $\vec{t} = (t_x,t_y,1)$
  SVector<double,3> tvec() const;
  /// $\vec{e} = (e_x,e_y,ez)$ unit vector along $\vec{p}$
  SVector<double,3> evec() const;
  /// $\vec{p} = (p_x,p_y,p_z)$
  SVector<double,3> pvec() const;
  /// cov. matrix
  const SMatrix<double,5>& COV() const;
  /// inverse cov. matrix
  const SMatrix<double,5>& CINV() const;
#endif

  /// get $\sigma_x^2|_{z+dz}$
  float cov_x(double dz=0.)        const;
  /// get $\sigma_y^2|_{z+dz}$
  float cov_y(double dz=0.)        const;
  /// get $\sigma_{t_x}$
  float cov_tx()            const;
  /// get $\sigma_{t_y}$
  float cov_ty()            const;
  /// get $\sigma_p^2$
  float cov_p()             const;

  /// collect pointers
  void collect(vector<TrackIf*>& c) const;
  /// propagate track parameters \& covariance matrix to $z$
  void propagate(const double z);

  /** @name --- PatternMatchIf access methods --- */
  /// get TrackIf part of reconstructed particle
  const TrackIf& getTrackIf() const;
  /// get the mother RecoTrack object (if available)
  const RecoTrack& getRecoTrack() const;
  /// get the mother MC\_Track object (if available)
  const MC_Track& getMC_Track() const;
  /// read only pattern access
  const SPattern& getPattern() const;
  /// invariant mass of vertex
  double mass() const;
  /// error on invariant mass
  double massError() const;
  /// calculate a probabilty including the mass
  double massProb(double scale) const;
  /// does this match contain MC tracks?
  bool isMCmatch() const;
  /// calculate mother track + cov. matrix
  bool calc_mother();
  /// calculate mother track
  bool calc_mother_tr();
  /// calculate mother covariance matrix
  bool calc_mother_cov();

  /** @name --- VertexIf Access methods --- */
  /// vertex $x$ position
  float vx()   const;
  /// vertex $y$ position
  float vy()   const;
  /// vertex $z$ position
  float vz()   const;
  /// vertex $\chi^2$
  float chi2() const;
  /// degrees of freedom of vertex fit
  unsigned short int ndf()  const;
  /// upper tail $\chi^2$ probability
  float prob() const;
  /// no of tracks in vertex
  unsigned short int ntracks() const;
  /// return a track pointer
  const TrackIf* track(unsigned int i) const;
  /// read/write track access
  const TrackIf*& track(unsigned int i);
  
  /// $\vec{v} = (x,y,z)$
  SVector<double,3> vpos() const;
  /// vertex cov. matrix
  const SMatrix<double,3>& VCOV() const;

  /// $\sigma_x^2$ of vertex
  float vtx_cov_x()   const;
  /// $\sigma_y^2$ of vertex
  float vtx_cov_y()   const;
  /// $\sigma_z^2$ of vertex
  float vtx_cov_z()   const;

  /// $\vec{x} = (x,y,z)$ (fast access)
  const SVector<double,3>& vposR() const;

  /** @name --- SVertex specific members --- */
  /// inverse vertex cov. matrix
  const SMatrix<double,3>& VCINV() const;
  /// mother track parameters
  const SVector<double,6>& mother() const;

  /** @name --- Operators --- */
  /// compare Track member with argument pointers
  bool operator==(const TrackIf& rhs) const;
  /// compare Topp pointer with decay pattern
  bool operator==(const ToppIf& rhs) const;
  /// compare with GTopp
  bool operator==(const GTopp& rhs) const;

  /** @name --- DecayPatternMatchIf access functions --- */
  /// return a PatternMatchIf pointer
  const PatternMatchIf* PatternMatch(unsigned int i) const;

  /** @name --- Expert functions --- */
  /// r/w access to SVertex object
  SmartPointer<SVertex<NTR> >& vertex();
  /// set mother track manually
  void set_mother(const TrackIf& rhs);
  /// set MC flag manually
  void set_MC(bool ismc);
  /// set a user field of given size, return pointer to it
  float* set_user(unsigned int size);
  /// get pointer to user field
  float* user() const;
  /// get size of user field
  unsigned int user_size() const;
  /// used by operator<<
  std::ostream& print(std::ostream& os) const;

  /** @name --- Relations --- */
  /// DecayPatternMatch $\rightarrow$ PatternMatch (relation to outgoing tracks)
  GRelationList<SPatternMatch> out; //! not persistent
  ///
  const GRelationList<SPatternMatch>& Out() const;
  ///
  GRelationList<SPatternMatch>& Out();
  ///
  GRelationList<SPatternMatch>& container(const SPatternMatch& p);

private:
  SmartPointer<SVertex<NTR> >     vtx_;      // vertex object
  mutable SmartPointer<SPattern>  vtxpat;    // decay pattern
  const TrackIf*                  mother_;   // a manually putted mother track (if available)
  bool                            ismc_;     // are MC tracks involved?
  float* user_;                              // pointer to a user array
  unsigned int N_;                           // size of user array
};


//==============================================================================
// operator <<: the second template argument is a problem, therefore hardcoded
//==============================================================================
template <unsigned int NTR>
inline std::ostream& operator<<(std::ostream& os,
                                const SDecayPatternMatch<NTR>& rhs) {
  return rhs.print(os);
}

#include "pattern/SDecayPatternMatch.icc"
#endif
