#ifndef __SDECAYPATTERNMATCHCONTAINER_HH
#define __SDECAYPATTERNMATCHCONTAINER_HH
// ********************************************************************
//
// source:
//
// type:      source code
//
// created:   30. Apr 2001
//
// author:    Thorsten Glebe
//            HERA-B Collaboration
//            Max-Planck-Institut fuer Kernphysik
//            Saupfercheckweg 1
//            69117 Heidelberg
//            Germany
//            E-mail: T.Glebe@mpi-hd.mpg.de
//
// Description: A DecayPatternMatch container class
//
// changes:
// 30 Apr 2001 (TG) creation
// 01 Mai 2001 (TG) select(), do_combine() added
// 03 Mai 2001 (TG) massrange() added
// 18 Mai 2001 (TG) operator&() added, CTOR, pcList(), pmList(), print(),
//                  operator<<() added
// 18 Mai 2001 (TG) added member Comb, prob_collision()
// 01 Jun 2001 (TG) check_coll, do_collision_free_combine(), check_collision() added
// 06 Jun 2001 (TG) removed check_coll_do_collision_free_combine(),
//                  check_collision() because they are obsolete due to a
//                  superior Combiner class (GSCombiner)
// 12 Jun 2001 (TG) small bug fix in operator<<(SmartPointer...)
// 13 Jun 2001 (TG) added massProb_collision(), massProb_bg_collision
// 25 Jun 2001 (TG) moved typedefs to P_trait and DP_trait
// 28 Jun 2001 (TG) removed template parameter C and T
// 02 Jul 2001 (TG) changed ptrlist added, type of pmlist to SPMP_l, pmatches() added
// 03 Jul 2001 (TG) added calc_mother(), calc_mother_tr(), calc_mother_cov()
// 06 Jul 2001 (TG) check_combination() added, added bool arg to combine(), do_combine()
//                  added scale arg to massProb_collision
// 25 Jul 2001 (TG) updated comments
// 22 Aug 2001 (TG) removed pattern, changed return type of dpattern()
// 23 Aug 2001 (TG) changed dpattern_ type to SDecayPattern<NTR>*, added push_back()
// 27 Aug 2001 (TG) added selectRECOfromMCtruth(), added push_back(SVertex)
// 28 Aug 2001 (TG) added selectNORECOfromMCtruth(), push_back(MC_Track),
//                  push_back(SDecayPatternMatch)
//
// ********************************************************************

#include <iosfwd>
#include "pattern/P_trait.hh"

class ToppIf;
class TrackIf;
class VertexIf;
class CutIf;
class MC_Track;
template <unsigned int NTR> class SDecayPattern;
template <unsigned int NTR> class SDecayPatternMatch;
template <unsigned int NTR> class SPatternExpression;
template <class T> class SmartPointer;
template <unsigned int NTR> class SVertex;

/** Class which represents a container for DecayPattern matches.

    @memo Generic class
    @author T. Glebe
*/
//==============================================================================
// SDecayPatternMatchContainer
//==============================================================================
template <unsigned int NTR>
class SDecayPatternMatchContainer {

  typedef typename DP_trait<NTR>::SDPM_l         SDPM_l;
  typedef typename DP_trait<NTR>::SDPM_it        SDPM_it;
  typedef typename DP_trait<NTR>::SDPM_cit       SDPM_cit;
  typedef typename DP_trait<NTR>::SDPMP_l        SDPMP_l;
  typedef typename DP_trait<NTR>::SDPMP_cit      SDPMP_cit;
  //  typedef typename DP_trait<NTR>::SDP_sp         SDP_sp;
  typedef SDecayPattern<NTR>                     SDP;

public:
  /** @name --- Constructors --- */
  ///
  SDecayPatternMatchContainer();
  ///
  SDecayPatternMatchContainer(const ToppIf& topp,
			      const SPatternExpression<NTR>& rhs);
  ///
  virtual ~SDecayPatternMatchContainer();

  /** @name --- Access methods --- */
  /// get pattern pointer
  const SP_sp& pattern() const;
#ifdef XXX
  /// get decay pattern pointer
  const SmartPointer<SDecayPattern<NTR> >& dpattern() const;
#endif
  /// set a Track Property
  void with(const SProperty<TrackIf,bool>& rhs);
  /// set a Vertex Property
  void with(const SProperty<VertexIf,bool>& rhs);
  /// read only access to matches
  const typename DP_trait<NTR>::SDPM_l& matches() const;
  /// read/write access to matches
  typename DP_trait<NTR>::SDPM_l& matches();
  /// read only access to matches
  const SPMP_l& pmatches() const;
  /// read/write access to matches
  SPMP_l& pmatches();
  /// set mass range cut around expected invariant mass
  void massrange(double mr);
  /// return mass range cut
  double massrange() const;
#ifdef XXX
  ///
  const SmartPointer<SPatternMatchContainer>* pcList() const;
#endif
  ///
  const SPMP_l* const * pmList() const;
  /// add a MC decay pattern match
  SDecayPatternMatch<NTR>& push_back(const MC_Track& mtra);
  /// add a reconstructed decay pattern match
  SDecayPatternMatch<NTR>& push_back(const TrackIf& track);
  /// add a SVertex decay pattern match
  SDecayPatternMatch<NTR>& push_back(SVertex<NTR>* vtx);
  /// add a decay pattern match
  SDecayPatternMatch<NTR>& push_back(const SDecayPatternMatch<NTR>& dpm);

  /** @name --- Expert methods --- */
  /// do combinatorics and fill decay pattern match container
  bool combine(bool check = false);
  ///
  bool check_combination(const GSC_it& it) const;
#ifdef XXX
  /** get tracks via SPatternMatchContainer::select(ArteTable<C>())
      and do combinatorics + vertexing */
  bool select();
#endif
  /** Select decay pattern from MC truth (MC\_Track table) and fill tracks
      into container. */
  bool selectMCtruth();
  /** Select decay pattern from MC truth (MC\_Track table) which have been
      reconstructed and fill (reconstructed) tracks into container. */
  bool selectRECOfromMCtruth();
  /** Select decay pattern from MC truth (MC\_Track table) which could not be
      fully reconstructed and fill tracks into container. In case of
      reconstructed tracks, fill RecoTrack, in case of not reconstructed
      tracks, fill MC\_Track. */
  bool selectNORECOfromMCtruth();
  /// set a cut object
  void set_cut(const CutIf& c);
  /// if DecayPatternMatches share same track, flag DPM with smaller prob as invalid
  bool prob_collision();
  /// if DecayPatternMatches share same track, flag DPM with smaller massProb as invalid
  bool massProb_collision(double scale = 0.1);
  /// massProb\_collision function for background studies
  bool massProb_bg_collision();
#ifdef XXX
  /// fill container with tracks NOT used in decaypatternmatches
  bool getComplement(const unsigned int nr, PMC_sp& container);
#endif
  /// calculate mother tracks with covariance matrix, ignore invalid pattern matches
  bool calc_mother();
  /// calculate mother tracks, ignore invalid pattern matches
  bool calc_mother_tr();
  /// calculate cov. matrix of mother tracks, ignore invalid pattern matches
  bool calc_mother_cov();
  ///
  std::ostream& print(std::ostream& os) const;

protected:
  /// routine which does the work
  virtual bool do_combine(bool check);

  SDPM_l        dpmlist;     // output: decay pattern matches
  SPMP_l        ptrlist;     // list of pattern match pointers (=mother particles)
  const SPMP_l* pmlist[NTR]; // input: pattern match pointer lists
  SDP*          dpattern_;   // the decay pattern
  SP_sp         pattern_;    // decay pattern type converted to SPattern
  GSC           Comb;        // Combiner object
};

//==============================================================================
// operator <<: the second template argument is a problem, therefore hardcoded
//==============================================================================
template <unsigned int NTR>
inline std::ostream& operator<<( std::ostream& os,
				 const SDecayPatternMatchContainer<NTR>& a ) {
  return a.print( os );
}

//==============================================================================
// operator <<: the second template argument is a problem, therefore hardcoded
//==============================================================================
template <unsigned int NTR>
inline std::ostream& operator<<( std::ostream& os, const
				 SmartPointer<SDecayPatternMatchContainer<NTR> >& a ) {
  return a.null() ? os << "(nil)" : a->print( os );
}

#include "pattern/SDecayPatternMatchContainer.icc"
#endif
