#ifndef __SPATTERN_HH
#define __SPATTERN_HH
// ********************************************************************
//
// source:
//
// type:      source code
//
// created:   20. Apr 2001
//
// author:    Thorsten Glebe
//            HERA-B Collaboration
//            Max-Planck-Institut fuer Kernphysik
//            Saupfercheckweg 1
//            69117 Heidelberg
//            Germany
//            E-mail: T.Glebe@mpi-hd.mpg.de
//
// Description: A track pattern class
//
// changes:
// 20 Apr 2001 (TG) creation
// 23 Apr 2001 (TG) bug fix in getString()
// 27 Apr 2001 (TG) cut_ added, rejectTrack() added
// 30 Apr 2001 (TG) cut_ removed, added evalTrackProp()
// 28 Jun 2001 (TG) changed from ClProperty to SProperty, removed template
//                  parameter T
// 03 Jul 2001 (TG) getDecayString() added
// 04 Jul 2001 (TG) setProperty(), hasProperty() added, removed set_property()
// 17 Aug 2001 (TG) operator==(MC_Track) added
// 21 Aug 2001 (TG) moved cut_, set_cut(), get_cut() from SDecayPattern to here
// 22 Aug 2001 (TG) removed code for charge conjugation, added CTOR, added
//                  PatternIf inheritance, added rm(), len(), rm_
// 23 Aug 2001 (TG) added spmc_, set_container(), getContainer()
// 24 Aug 2001 (TG) added fillMC()
// 27 Aug 2001 (TG) added isReconstructed(MC_Track), fillRECOfromMC()
// 28 Aug 2001 (TG) added isNotReconstructed(MC_Track), fillNORECOfromMC()
//
// ********************************************************************

#include <iosfwd>
#include <string>
#include "interfaces/PatternIf.hh"
#include "pattern/P_trait.hh"
#include "pattern/SmartPointer.hh"

class ToppIf;
class TrackIf;
class CutIf;
class PatternMatchIf;
class MC_Track;
class SPatternMatchContainer;
template <class C, class T> class SProperty;
template <unsigned int NTR> class SDecayPattern;

/** A generic template class for defining track selection pattern, based on
    SProperty expressions.

    @memo Generic class
    @author T. Glebe
 */
//==============================================================================
// SPattern
//==============================================================================
class SPattern : public PatternIf {
public:
  /** @name --- Constructors --- */
  ///
  SPattern();
  ///
  SPattern(const SPattern& rhs);
  /// copy rhs and replace ToppIf of rhs by tp
  SPattern(const ToppIf& tp, const SPattern& rhs);
  ///
  SPattern(const ToppIf& gt);
  ///
  SPattern(const ToppIf& gt, const SProperty<TrackIf,bool>& rhs);
  ///
  virtual ~SPattern();
  ///
  SPattern& operator=(const SPattern& rhs);

  /** @name --- PatternIf access functions --- */
  /// check if single Track passes property criterias
  bool evalTrack(const TrackIf& rhs) const;
  /// check track property without charge
  bool evalTrackProp(const TrackIf& rhs) const;
  /// get property object
  const SProperty<TrackIf,bool>& property() const;
  /// get topp object
  const ToppIf& topp() const;
  /// return pointer to topp mass
  const double* rm() const;
  /// always return 1
  unsigned int len() const;
  /// get decay string
  string getDecayString(unsigned int lvl=0) const;
  /// print property
  void printProperty(std::ostream& os, unsigned int lvl=0) const;
  /// get pointer to list of pattern matches
  const SPMP_l* getContainer() const;
  /// fill container from MC event
  const PatternMatchIf* fillMC(const MC_Track& mt) const;
  /// same as operator==(MC\_Track), but also check whether MC\_Track was reconstructed
  bool isReconstructed(const MC_Track& rhs) const;
  /// same as operator==(MC\_Track), but also check if MC\_Track was not reconstructed
  bool isNotReconstructed(const MC_Track& rhs, bool& no_break) const;
  /// fill container with reconstructed tracks from MC event
  PatternMatchIf* fillRECOfromMC(const MC_Track& mt) const;
  /// fill container with reconstructed tracks or (if not reconstr.) MC tracks from MC event
  PatternMatchIf* fillNORECOfromMC(const MC_Track& mt) const;

  /** @name --- Operators --- */
  /// OR of two patterns
  SPattern operator|( const SPattern& rhs ) const;
  /// AND of patterns defines a decay
  SDecayPattern<2> operator&(const SPattern& rhs) const;
  ///
  bool operator==(const SPattern& rhs) const;
  ///
  bool operator==(const ToppIf& rhs) const;
  ///
  bool operator!=(const ToppIf& rhs) const;
  ///
  virtual bool operator==(const MC_Track& rhs) const;

  /** @name --- User functions --- */
  /** select all tracks from InputContainer which passes property 
      criteria and put them as PatternMatches in OutputContainer */
#ifdef XXX
  template <class InputContainer, class OutputContainer>
  bool select(const InputContainer&  src, OutputContainer& trg) const;
#endif
  /// return true if a track which matches the pattern exists in InputContainer
  template <class InputContainer>
  bool exist(const InputContainer& src) const;

  /// set cprop pointer
  void set_cprop(const SProperty<TrackIf,bool>& rhs);
  /// Topp charge divided by 3
  int chrg() const;
  /// does property exist?
  bool hasProperty() const;
  /// get pattern Name
  string getName() const;
  /// get property string
  string PropertyString() const;
  /// get the pattern string
  virtual string getString() const;
  /// true if rhs is equal to pattern property
  bool check_property(const SProperty<TrackIf,bool>& rhs) const;
  /// set a cut object
  void set_cut(const CutIf& c);
  /// get the cut object
  const CutIf* get_cut() const;
  
  /** @name --- Expert functions --- */
  /// set property object
  void setProperty(const SProperty<TrackIf,bool>& rhs);
  /// set pointer to container object
  void set_container(SPatternMatchContainer& rhs);
  /// used by operator<<
  virtual std::ostream& print(std::ostream& os) const;
  /// merge two pattern, only merge property
  void merge_property(const SPattern& rhs);

protected:
  const ToppIf*          topp_;     // particle property
  double                 rm_;       // particle rest mass
  SmartPointer<SProperty<TrackIf,bool> > property_; // cuts defined by ClProperty
  SmartPointer<SProperty<TrackIf,bool> > cprop;     // cuts plus charge cut
  const CutIf*           cut_;      // object for cuts
  SPatternMatchContainer* spmc_;    // pointer to container object
}; // end of SPattern.hh


//==============================================================================
// operator <<: the second template argument is a problem, therefore hardcoded
//==============================================================================
inline std::ostream& operator<<( std::ostream& os, const SPattern& a ) {
  return a.print( os );
}

//==============================================================================
// operator&
//==============================================================================
SDecayPattern<2> operator&(const SP_sp& lhs, const SP_sp& rhs);

#include "pattern/SPattern.icc"
#endif
