#ifndef __SPATTERNMATCH_HH
#define __SPATTERNMATCH_HH
// ********************************************************************
//
// source:
//
// type:      source code
//
// created:   20. Apr 2001
//
// author:    Thorsten Glebe
//            HERA-B Collaboration
//            Max-Planck-Institut fuer Kernphysik
//            Saupfercheckweg 1
//            69117 Heidelberg
//            Germany
//            E-mail: T.Glebe@mpi-hd.mpg.de
//
// Description: A PatternMatch class for use without Vt
//
// changes:
// 20 Apr 2001 (TG) creation
// 27 Apr 2001 (TG) pt(), xfabs() added
// 02 Mai 2001 (TG) phi(), theta(), eta() added
// 06 Jun 2001 (TG) changed Vtx relation type from VertexIf to DecayPatternMatchIf
// 11 Jun 2001 (TG) added COV(), cov_tx(), cov_ty(), cov_p()
// 28 Jun 2001 (TG) removed templates
// 29 Jun 2001 (TG) added charge(), pz(), energy(), xf(), rap(), removed xfabs()
// 29 Jun 2001 (TG) added decay-relation
// 03 Jul 2001 (TG) added CTOR
// 05 Jul 2001 (TG) removed decay-relation, added collect()
// 09 Jul 2001 (TG) isValid(), valid(), invalid() from TrackIf added
// 10 Jul 2001 (TG) operator==() added
// 20 Aug 2001 (TG) added user field from SDecayPatternMatch
// 21 Aug 2001 (TG) changed inheritance from TrackIf to PatternMatchIf
// 28 Aug 2001 (TG) ismc_ added, CTOR added
// 29 Aug 2001 (TG) added operator==(GTopp)
// 31 Aug 2001 (TG) added PatternMatch(i)
//
// ********************************************************************
#include <iosfwd>
#include "pattern/SmartPointer.hh"
#include "interfaces/PatternMatchIf.hh"
#include "grelation/GRelationList.hh"

class SPattern;
class ToppIf;
class DecayPatternMatchIf;
class RecoSegment;
class MC_Track;

/** Class which represents a selected track object together with the cuts which
    were used to select it. The cuts are represented by a ClProperty
    expression object.

    @memo Generic class
    @author T. Glebe
*/
//==============================================================================
// SPatternMatch
//==============================================================================
class SPatternMatch : public PatternMatchIf {
public:
  /** @name --- Constructors --- */
  ///
  SPatternMatch();
  ///
  SPatternMatch(RecoSegment& rs, const SPattern& pat);
  ///
  SPatternMatch(RecoSegment& rs, const SmartPointer<SPattern>& pat);
  ///
  SPatternMatch(MC_Track& mt, const SmartPointer<SPattern>& pat);
  ///
  SPatternMatch(TrackIf& tr);
  ///
  SPatternMatch(TrackIf& tr, const SPattern& pat);
  ///
  SPatternMatch(TrackIf& tr, const SmartPointer<SPattern>& pat);
  ///
  SPatternMatch(const SPatternMatch& rhs);
  ///
  virtual ~SPatternMatch();

  /** @name --- TrackIf access methods --- */
  /// Track $x$ position
  float x()   const;
  /// Track $y$ position
  float y()   const;
  /// Track $x$ position at $z$
  float x(float z)   const;
  /// Track $y$ position at $z$
  float y(float z)   const;
  /// Track $z$ position
  float z()   const;
  /// Track slope $t_x$
  float tx()  const;
  /// Track slope $t_y$
  float ty()  const;
  /// Track momentum $p$
  float p()   const;
  /// transverse Track momentum $p_t$
  float pt()  const;
  /// Track momentum along $z$
  float pz()  const;
  /// Track $\chi^2$
  float chi2() const;
  /// degrees of freedom of track fit
  unsigned short int ndf()  const;
  /// azimuthal angle $\phi$ [deg]
  float phi() const;
  /// polar angle $\theta = \cos^{-1}(e_z)$ [deg]
  float theta() const;
  /// rapidity $\eta = -\log\tan(\theta/2.)$
  float eta() const;
  /// track charge
  int   charge() const;

  /// track energy $E=\sqrt{m^2 + p^2}$
  float energy(double rm = 0.) const;
  /// Feynman-$x$: $x_F = \frac{E+p_z}{(E+p_z)_{max}}$
  float xf(double rm = 0.) const;
  /// Rapidity $y = \frac{1}{2}\ln(\frac{E+p_z}{E-p_z})$
  float rap(double rm = 0.) const;
  
  /// is PatternMatch valid?
  bool isValid() const;
  /// mark PatternMatch as valid
  void valid();
  /// mark PatternMatch as invalid
  void invalid();

#ifndef __CINT__
  /// $\vec{x} = (x,y,z)$
  SVector<double,3> xvec() const;
  /// $\vec{t} = (t_x,t_y,1)$
  SVector<double,3> tvec() const;
  /// $\vec{e} = (e_x,e_y,ez)$ unit vector along $\vec{p}$
  SVector<double,3> evec() const;
  /// $\vec{p} = (p_x,p_y,p_z)$
  SVector<double,3> pvec() const;
  /// cov. matrix
  const SMatrix<double,5>& COV() const;
  /// inverse cov. matrix
  const SMatrix<double,5>& CINV() const;
#endif

  /// get $\sigma_x^2|_{z+dz}$
  float cov_x(double dz=0.)        const;
  /// get $\sigma_y^2|_{z+dz}$
  float cov_y(double dz=0.)        const;
  /// get $\sigma_{t_x}$
  float cov_tx()            const;
  /// get $\sigma_{t_y}$
  float cov_ty()            const;
  /// get $\sigma_p^2$
  float cov_p()             const;

  /// collect pointers
  void collect(vector<TrackIf*>& c) const;
  /// propagate track parameters \& covariance matrix to $z$
  void propagate(const double z);

  /** @name --- PatternMatchIf access methods --- */
  /// read only track access on interface-level
  virtual const TrackIf& getTrackIf() const;
  /// get the RecoTrack object
  virtual const RecoTrack& getRecoTrack() const;
  /// get the MC\_Track object
  virtual const MC_Track& getMC_Track() const;
#ifdef XXX
  /// read only track access to full track
  const RecoSegment& getRecoSegment() const;
#endif
  /// read only pattern access
  const SPattern& getPattern() const;
  /// assumed rest-mass of track (defined by pattern)
  double mass() const;
  /// dummy function: always returns 0
  double massError() const;
  /// dummy function: same as track prob() (as massError()==0)
  double massProb(double scale) const;
  /// is this a match with a MC track?
  bool isMCmatch() const;
  /// dummy function: always return 0
  const PatternMatchIf* PatternMatch(unsigned int i) const;
  /// dummy function: always return true
  bool calc_mother();
  /// dummy function: always return true
  bool calc_mother_tr();
  /// dummy function: always return true
  bool calc_mother_cov();

  /** @name --- Operators --- */
  /// compare Track member with argument pointers
  bool operator==(const TrackIf& rhs) const;
  /// compare Track members pointers
  bool operator==(const SPatternMatch& rhs) const;
  /// compare topp pointer with pattern
  bool operator==(const ToppIf& rhs) const;
  /// compare with GTopp
  bool operator==(const GTopp& rhs) const;

  /** @name --- Expert functions --- */
  /// set a user field of given size, return pointer to it
  float* set_user(unsigned int size);
  /// get pointer to user field
  float* user() const;
  /// get size of user field
  unsigned int user_size() const;
  /// used by operator<<
  std::ostream& print(std::ostream& os) const;

  /** @name --- Relations --- */
  /** PatternMatch $\rightarrow$ DecayPatternMatch: relation of outgoing
      particles to decay objects */
  GRelationList<DecayPatternMatchIf> vtx; //! not persistent
  ///
  const GRelationList<DecayPatternMatchIf>& Vtx() const;
  ///
  GRelationList<DecayPatternMatchIf>& Vtx();
  ///
  GRelationList<DecayPatternMatchIf>& container(const DecayPatternMatchIf& p);

private:
  TrackIf*                      track;    // track object
  RecoSegment*                  rs_;      // pointer to RecoSegment object
  mutable SmartPointer<SPattern>  pattern;  // pattern which selected track
  float* user_;                           // pointer to a user array
  unsigned int N_;                        // size of user array
  bool ismc_;                             // flag if track ptr is MC_Track
}; // end of SPatternMatch

//==============================================================================
// operator <<
//==============================================================================
std::ostream& operator<<( std::ostream& os, const SPatternMatch& a );

#include "pattern/SPatternMatch.icc"
#endif
