#ifndef GPROPERTY_HH
#define GPROPERTY_HH
// **********************************************************************
//
// source:
//
// type:      source code
//
// created:   16. Sep 1998
//
// author:    Thorsten Glebe
//            HERA-B Collaboration
//            Max-Planck-Institut fuer Kernphysik
//            Saupfercheckweg 1
//            69117 Heidelberg
//            Germany
//            E-mail: T.Glebe@mpi-hd.mpg.de
//
// Description: A generic pattern expression engine.
//              At the moment only the operators & and | are provided.
//
// changes:
// 16 Sep 1998 creation by C. Borgmeier
//
// **********************************************************************

#include <iosfwd>

template <class C, class T> class GExpression;

template <class T> T AndOp(const T& lhs, const T& rhs);
template <class T> T OrOp (const T& lhs, const T& rhs);


/** Expression template base class. Each class which wants to take part in a
    GExpression must inherit from GProperty.
    @memo class GProperty
    @author T. Glebe
*/
//==============================================================================
// GProperty
//==============================================================================
template <class C, class T>
class GProperty {
public:
  /** @name --- Constructors --- */
  ///
  GProperty() {}
  ///
  virtual ~GProperty() {}
  
  /** @name --- Operators --- */
  ///
  inline GExpression<C,T> operator&( const GProperty<C,T>& rhs ) const
    { return GExpression<C,T>( *this, rhs, &AndOp ); }
  ///
  inline GExpression<C,T> operator|( const GProperty<C,T>& rhs ) const
    { return GExpression<C,T>( *this, rhs, &OrOp ); }
  
  
  /** @name --- pure virtual functions --- */
  /** This member function determines the template parameters C and T.
      T: information passed up the tree
      C: information passed down the tree */
  virtual T evaluate( const C& ) const = 0;
  /// return pointer to copy of object on heap
  virtual GProperty<C,T>* propertyClone() const = 0;
  /// used by operator<<()
  virtual std::ostream& print( std::ostream& os ) const = 0;
  
}; // end of class GProperty


/** Expression template.
    GExpression implements a binary operation. Due to the inheritance from
    GProperty it is possible to build up a parse tree. The evaluate()
    member functions evaluates the parse tree.
    @author T. Glebe
*/
//==============================================================================
// GExpression
//==============================================================================
template <class C, class T>
class GExpression: public GProperty<C,T> {
public:
  /** @name --- Constructors --- */
  ///
  GExpression();
  ///
  GExpression( const GProperty<C,T>& lhs,
	       const GProperty<C,T>& rhs,
	       T (*f)( const T& r, const T& l));
  ///
  GExpression( const GExpression<C,T>& rhs );
  ///
  ~GExpression();
  ///
  const GExpression<C,T>& operator=( const GExpression<C,T>& );

  /** @name --- Expert functions --- */
  /// replacement for operator()
  T evaluate( const C& obj ) const;
  ///
  virtual GProperty<C,T>* propertyClone() const;
  /// used by operator<<()
  virtual std::ostream& print( std::ostream& os ) const;
  
private:
  
  GProperty<C,T>* lhs;
  GProperty<C,T>* rhs;
  T (*func)( const T&, const T& );

}; // GExpression


// Load the implementation file
#ifndef __CINT__
#include "GProperty.icc"
#endif
#endif // GPROPERTY_HH
